
//		ISO8601.cpp - adds support for more date formats


#define _UNICODE			1

#include <tchar.h>

//		Needed only for @RANDDATE and @RANDTIME :
//		#define _CRT_RAND_S
//		#include <stdlib.h>

#include <windows.h>

#ifndef QCAL_ONLY
#include <WinInet.h>
#endif	//	QCAL_ONLY

#include "PlugIn.h"
#include "TakeCmd.h"
#include "popcal\popcal.h"

#define VER_MAJOR			1
#define VER_MINOR			7
#define VER_BUILD			4
#define VER_PATCH			1

#ifdef _WIN64
#define PLATFORM			L"x64"
#else	//	_WIN64
#define PLATFORM			L"x86"
#endif	//	_WIN64


#define ERRHANDLE			GetStdHandle( STD_ERROR_HANDLE )

#define SHORT_BUF_LEN		1024
#define MAX_ARG_LEN			8192
#define MAX_FILENAME_LEN	16384


const int Epoch				= 1601;			//  first year in supported range
const int Ultima			= 9999;			//  last year in supported range



//		The supported date range of julians:
const long int MinimumJulian	= 1;
const long int MaximumJulian	= 3067671;

//		Used to convert between julians and MS-DOS day numbers:
const long int JulDOSEpoch		= 138427;

//      Used to convert between julians and Unix day numbers:
const long int JulUnixEpoch		= 134775;

//		The supported date range in MS-DOS day numbers:
const long int MinimumDayNum	= MinimumJulian - JulDOSEpoch;
const long int MaximumDayNum	= MaximumJulian - JulDOSEpoch;


const int MonthLength[13][2]	= { {0, 0}, {31, 31}, {28, 29}, {31, 31}, {30, 30}, {31, 31}, {30, 30}, {31, 31}, {31, 31}, {30, 30}, {31, 31}, {30, 30}, {31, 31} };
const int Kalends[13][2]		= { {0, 0}, {0, 0}, {31, 31}, {59, 60}, {90, 91}, {120, 121}, {151, 152}, {181, 182}, {212, 213}, {243, 244}, {273, 274}, {304, 305}, {334, 335} };


#define LOCAL_NAME_LEN							40

const wchar_t DowName[7][4]						= { L"Sun", L"Mon", L"Tue", L"Wed", L"Thu", L"Fri", L"Sat" };
const wchar_t DowNameLong[7][10]				= { L"Sunday", L"Monday", L"Tuesday", L"Wednesday", L"Thursday", L"Friday", L"Saturday" };
wchar_t LocalDowName[7][LOCAL_NAME_LEN]			= { L"Sun", L"Mon", L"Tue", L"Wed", L"Thu", L"Fri", L"Sat" };
wchar_t LocalDowNameLong[7][LOCAL_NAME_LEN]		= { L"Sunday", L"Monday", L"Tuesday", L"Wednesday", L"Thursday", L"Friday", L"Saturday" };

const wchar_t MonthName[13][10]					= { L"", L"January", L"February", L"March", L"April", L"May", L"June", L"July", L"August", L"September", L"October", L"November", L"December" };
wchar_t LocalMonthName[13][LOCAL_NAME_LEN]		= { L"", L"Jan", L"Feb", L"Mar", L"Apr", L"May", L"Jun", L"Jul", L"Aug", L"Sep", L"Oct", L"Nov", L"Dec" };
wchar_t LocalMonthNameLong[13][LOCAL_NAME_LEN]	= { L"", L"January", L"February", L"March", L"April", L"May", L"June", L"July", L"August", L"September", L"October", L"November", L"December" };
wchar_t LocalMonthNameCaps[13][LOCAL_NAME_LEN];

wchar_t CountryName[128]			= L"";

#define WEEKDAYS_MASK	0x3e
#define WEEKENDS_MASK	0x41

int LocalDateFormat					= 1;		// 1: m/d/y, 2: d/m/y, 3: y/m/d


#ifndef QCAL_only

//		Global variables not needed for QCAL_only plugin :

int LongDateFormat			= 1;
wchar_t LocalDateSep		= '/';		// local date separator character
wchar_t LocalTimeSep		= ':';     // local time separator character
wchar_t LocalDecSep			= '.';     // local decimal separator character
BOOL FourDigitsAlways		= FALSE;  // always display four-digit years
BOOL StrictYearInput		= FALSE;  // require four-digit input years

const long long int MaximumAge		= 2650467743999999999;

#else	//	QCAL_only

#define StrictYearInput		FALSE

#endif	//	QCAL_only

int LeftColumnDow			= 0;

wchar_t svSetDos[32]		= L"";
wchar_t fxSetDos[32]		= L"";


#ifndef QCAL_only

//		Strings used only by the full ISO8601 plugin :

wchar_t PluginFeatures[] = L"@agedate,@century,@date,@dateconv,@dateplus,@day,@days,@dectotime,@dow,@dowf,@dowi,@dowis,\
@doy,@easter,@filestamp,@holiday,@idow,@idowf,@inlieu,@inrange,@isdate,@isholiday,@isodowi,@isoweek,@isowyear,@istime,\
@julian,@leap,@localdatetime,@makeage,@makedate,@month,@monthlen,@monthname,@nextdate,@nextdow,@ord,@pickdate,@roman,\
@timediff,@timefmt,@timeparts,@timeplus,@timetodec,@useconds,@usweek,@utcdatetime,@uxtodatetime,@year,\
_isodowi,_isordate,_isowdate,_isoweek,_isowyear,_t4,_tomorrow,_useconds,_yesterday,\
iso8601help,longyears,qcal";



const wchar_t ISO8601HelpHelpText[] = L"Open this plugin\u2019s help file.\n\
\n\
ISO8601HELP /C /F /S /S:text /V topic\r\n\
\n\
\t/C\tselect the Contents tab\n\
\t/F\tselect the Favorites tab\n\
\t/S\tselect the Search tab\n\
\t/S:\tselect the Search tab and search for text\n\
\t/V\tshow detailed plugin version info\n\
\n\
Opens this plugin\u2019s help file to the specified topic.\n\
If no topic is given, opens the Overview page by default. \n";


const wchar_t LongYearsHelpText[] = L"Enable or disable two-digit years in local date formats.\n\
\n\
LONGYEARS [OFF | ON | STRICT]\r\n\n\
\tOFF:\toutput two-digit years in formats 0-3 (default)\n\
\tON:\toutput four-digit years in formats 0-3\n\
\tSTRICT:\talso require four-digit years in all input dates  \n";


#define PLUGIN_NAME		L"ISO8601"
#define UNLOAD_PLUGIN	L"QCal"


#else	//	QCAL_only


//		Strings used by the QCAL_only plugin :

const wchar_t PluginFeatures[] = L"@easter,@holiday,@inlieu,@nextdow,_t4,_tomorrow,_yesterday,qcal";


#define PLUGIN_NAME		L"QCal"
#define UNLOAD_PLUGIN	L"ISO8601"


const wchar_t QcalAdditionalHelpText[] = L"\nYou may specify any one of:  a full date, in any input date format\n\
supported by this plugin; a month (1 - 12, or JAN - DEC), and/or a\n\
four-digit year; or nothing, in which case QCAL will default to today\u2019s\n\
date.  If only a year is specified, then QCAL will display a calendar\n\
for the entire year; otherwise, it will show a one-month calendar by\n\
default.  If a month is specified without a year, then QCAL will assume\n\
the current year if the specified month is greater than or equal to the\n\
current month, or next year if the specified month is less than the\n\
current month.\n\
\n\
If /I is specified, QCAL will automatically put Mondays on the left,\n\
unless directed otherwise with /W:.  The week numbers shown are for\n\
the leftmost valid date in each row \u2014 this is tolerably obvious when\n\
Mondays are on the left, but less intuitive in other arrangements.\n\
\n\
QCAL\u2019s output may be piped or redirected.\n\n";


wchar_t F_EasterHelpText[] = L"Returns the date of Easter for the specified year.\n\n\
Syntax:  @EASTER[year, church, fmt, offset]\n\n\
  year:\tfour digits, 1601 - 4000; defaults to this year  \n\
  church:\t0 = Western, 1 = Eastern; defaults to 0\n\
  fmt:\tnot used in this plugin\n\
  offset:\toptional number of days to add or subtract\n\n\
The date will be returned in ISO 8601 format.";


wchar_t F_HolidayHelpText[] = L"Returns the date of a specific day-of-the-week within a month\n\
(e.g., the second Monday in October.)\n\n\
Syntax:  @HOLIDAY[index, dow, month, year, fmt, offset]\n\n\
  index:\t1 = first, 2 = second, 3 = third, 4 = fourth, 5 = final  \n\
  dow:\t1 = Monday, 2 = Tuesday, 3 = Wednesday, etc.\n\
  month:\t1 = January, 2 = February, 3 = March, etc.\n\
  year:\tfour digits, 1601 - 4000; defaults to this year\n\
  fmt:\tnot used in this plugin\n\
  offset:\toptional number of days to add or subtract\n\n\
The date will be returned in ISO 8601 format.";


wchar_t F_InlieuHelpText[] = L"Substitutes a weekday for a weekend date.\n\n\
Syntax:  @INLIEU[date, infmt, mode]\n\n\
  date:\tthe date to check; defaults to today\n\
  infmt:\tinput date format\n\
  mode:\t0 = nearest, 1 = previous Friday,\n\
\t2 = following Monday, 3 = Sunday (only)\n\
\tto Monday; defaults to 0\n\n\
If the input date is a weekday, no substitution will be  \n\
made.  The date will be returned in ISO 8601 format.";


wchar_t F_NextdowHelpText[] = L"Returns the next date matching a specified day\n\
(or days) of the week.\n\n\
Syntax:  @NEXTDOW[dow, date, index, infmt]\n\n\
  dow:\tthe day(s) of the week to find\n\
  date:\tthe starting date; defaults to today\n\
  index:\t1 = next after, 0 = next on or after,\n\
\t-1 = last before, etc.; defaults to 1\n\
  infmt:\tinput date format\n\n\
The date will be returned in ISO 8601 format.";


wchar_t V_T4HelpText[] = L"Returns the current local time as four digits, HHMM.\n";


wchar_t V_TomorrowHelpText[] = L"Returns tomorrow\u2019s date in ISO8601 format.\n";


wchar_t V_YesterdayHelpText[] = L"Returns yesterday\u2019s date in ISO8601 format.\n";


#endif	//	QCAL_only


//		Strings common to both plugins :

const wchar_t PluginRegKey[] = L"Software\\JPPlugins\\ISO8601";


const wchar_t QcalHelpText[] = L"Display a quick-and-dirty calendar.\n\
\n\
QCAL [date | [month] [year]] /3 /6 /C: /D /E /H /H: /I /L /N /P /W: /Y \r\n\
\n\
\t/3   three months   \t/I   ISO week numbers\n\
\t/6   six months\t\t/L   list holiday names\n\
\t/C:  country name  \t/N   don\u2019t highlight today\u2019s date  \n\
\t/D   disable color   \t/P   page output\n\
\t/E   English names  \t/W:  day on left (1 = Monday)\n\
\t/H   default holidays\t/Y   full year\n\
\t/H:  specify holidays\n\
\n\
The date may be entered in any format supported by this plugin,\n\
including ISO week-day and ordinal formats, day numbers, and\n\
T format.  To specify a date using its day number, prefix a + or - \n\
sign; otherwise it may be misinterpreted as a month or year.\n";



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables(shortbuf, 1);
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables(shortbuf, 1);
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



#include "ParseArgs.cpp"



SYSTEMTIME Now;


BOOL OutputToConsole	= FALSE;
int ScreenRows			= 0;
BOOL paged				= FALSE;
int LineCount			= 0;

BOOL EnableColor	= TRUE;
int NormalColor		= 0x07;
int HLColor			= 0x70;
int SundayColor		= 0x0c;
int HLSundayColor	= 0xc0;
int WeeksColor		= 0x03;
int HolidayColor	= 0x27;

int DefColors[16][5];


HINSTANCE MyDLLHandle = NULL;

OSVERSIONINFOEX OSVerInfo;


struct _HolidayDef {
	_HolidayDef *Next;
	long int Julian;
	BOOL Enabled;
	LPTSTR Name;
	wchar_t Text[1];
};

_HolidayDef *FirstHolidayDef		= NULL;


int QcalDoingDelayedExpansion = 0;



BOOL GetVersion2( OSVERSIONINFOEXW *VerInfo )
{
	if ( !VerInfo )
		return -1;


	typedef NTSTATUS ( WINAPI _RtlGetVersionProc ) ( OSVERSIONINFOEXW *VerInfo );
	HMODULE NtDllHandle						= GetModuleHandle( L"NtDll.dll" );
	_RtlGetVersionProc *RtlGetVersionProc	= NULL;


	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	if ( NtDllHandle )
		RtlGetVersionProc = (_RtlGetVersionProc*) GetProcAddress( NtDllHandle, "RtlGetVersion" );

	if ( RtlGetVersionProc ) {
		RtlGetVersionProc( VerInfo );
		if ( VerInfo->dwMajorVersion >= 6 )
			return TRUE;
	}

	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	return ( GetVersionEx( (OSVERSIONINFOW*) VerInfo ) != 0 );
}



int FreeLinkedList( void **Pointer )
{
	if ( !Pointer )
		return -1;

	size_t *This = (size_t*) *Pointer, Next = NULL;

	while ( This ) {
		// Printf( L"   * Freeing memory at 0x%08x\r\n", This );
		Next = *This;
		free( This );
		This = (size_t*) Next;
	}

	*Pointer = NULL;
	return 0;
}



int FreeHolidayDefs()
{
	FreeLinkedList( (void **) &FirstHolidayDef );
	return 0;
}



int QcalReadHolidaysFile();
int QcalDefaultHolidays();
int ExpandHolidays( int year, unsigned short int Holidays[][31] );



int GetNow( BOOL utcflag )
{
	if ( utcflag )
		GetSystemTime( &Now );
	else
		GetLocalTime( &Now );

	return 0;
}



BOOL InTheFuture( SYSTEMTIME Date )
{
	//  Returns true if the input date is after today's date.

	if ( Date.wYear < Now.wYear )
		return FALSE;
	else if ( Date.wYear > Now.wYear )
		return TRUE;

	if ( Date.wMonth < Now.wMonth )
		return FALSE;
	else if ( Date.wMonth > Now.wMonth )
		return TRUE;

	return ( Date.wDay > Now.wDay );
}



BOOL Leapness(int year)
{
	if ( ( year % 400 ) == 0 ) return TRUE;
	if ( ( year % 100 ) == 0 ) return FALSE;
	return ( ( year % 4 ) == 0 );
}



long int Julian( int year, int month, int day )
{
	//  A "julian" in this program is a day number, with julian 1
	//  being January 1, 1601.  It does not correspond to the
	//  traditional, big-J Julian day number.

	if ( ( year < Epoch ) || ( year > Ultima ) )
		return -1;

	if ( ( month < 1 ) || ( month > 12 ) )
		return -1;

	BOOL leap = Leapness( year );

	if ( ( day < 1 ) || ( day > MonthLength[month][leap] ) )
		return -1;

	long int j = ( ( year - Epoch ) * 365 );
	j += ( ( year - Epoch ) / 4 );
	j -= ( ( year - Epoch ) / 100 );
	j += ( ( year - Epoch ) / 400 );
	j += Kalends[month][leap];
	j += day;

	return j;
}



long int Julian( SYSTEMTIME Date )
{
	return Julian( Date.wYear, Date.wMonth, Date.wDay );
}



int doy( SYSTEMTIME Date )
{
	return Kalends[Date.wMonth][Leapness( Date.wYear )] + Date.wDay;
}



int ISOWeekOffset(int year)
{
	//  Returns a value indicating where ISO week 01 begins for
	//  the given year, relative to January 1.  Possible return
	//  values range from -3 (three days prior, i.e. December 29)
	//  through +3 (three days later, i.e. January 4).  A return
	//  value of 0 would mean that January 1st is a Monday.

	if ( ( year < Epoch ) || ( year > Ultima ) )
		return -999;

	int w = ( Julian( year, 1, 4 ) % 7 );
	int offset = 4 - ( ( w==0 ) ? 7 : w );

	return offset;
}



unsigned int ISOWeeksIn( int year )
{
	//  Years starting on a Thursday, and leap years starting on
	//  a Wednesday, have 53 ISO weeks.  All others have 52.

	BOOL leap = Leapness( year );
	int w = Julian( year, 1, 1 ) % 7;

	if ( ( w == 4 ) || ( leap && ( w == 3 ) ) )
		return 53;

	return 52;
}



#ifndef QCAL_only


int MakeISOWeekDate( SYSTEMTIME Date, unsigned int &isoyear, unsigned int &isoweeknum, unsigned int &isoweekday )
{
	isoyear		= Date.wYear;
	int ord		= doy( Date ) - 1;
	int offset	= ISOWeekOffset( isoyear );
	isoweeknum	= ( ( ord - offset ) / 7 ) + 1;
	isoweekday	= ( ( ord - offset + 7 ) % 7 ) + 1;


	if ( isoweeknum > ISOWeeksIn( isoyear ) ) {
		isoyear++;
		isoweeknum = 1;
	}
	else if ( ord < offset ) {
		isoyear--;
		isoweeknum = ISOWeeksIn( isoyear );
	}

	return 0;
}



int OutputFormattedDate( SYSTEMTIME Date, int fmt, LPTSTR outString )
{
	if ( !outString )
		return -1;


	BOOL leap = Leapness( Date.wYear );

	BOOL FourDigitYears = ( ( fmt & 0x100 ) > 0 ) || FourDigitsAlways;
	fmt = fmt & 0xff;
	unsigned int isoyear = 0, isoweeknum = 0, isoweekday = 0;
	long int j = 0, tj = 0;

	wchar_t FormatPic[SHORT_BUF_LEN]		= L"";
	LPTSTR PicPtr							= NULL;
	int wfmt = DATE_LONGDATE;


	if ( fmt == 0 )
		fmt = LocalDateFormat;

	if ( QcalDoingDelayedExpansion )
		fmt = 4;


	switch ( fmt ) {

	case 1:
		if ( FourDigitYears )
			wsprintf( outString, L"%02u%c%02u%c%04u", Date.wMonth, LocalDateSep, Date.wDay, LocalDateSep, Date.wYear );
		else
			wsprintf( outString, L"%02u%c%02u%c%02u", Date.wMonth, LocalDateSep, Date.wDay, LocalDateSep, Date.wYear % 100 );
		break;

	case 2:
		if ( FourDigitYears )
			wsprintf( outString, L"%02u%c%02u%c%04u", Date.wDay, LocalDateSep, Date.wMonth, LocalDateSep, Date.wYear );
		else
			wsprintf( outString, L"%02u%c%02u%c%02u", Date.wDay, LocalDateSep, Date.wMonth, LocalDateSep, Date.wYear % 100 );
		break;

	case 3:
		if ( FourDigitYears )
			wsprintf( outString, L"%04u%c%02u%c%02u", Date.wYear, LocalDateSep, Date.wMonth, LocalDateSep, Date.wDay );
		else
			wsprintf( outString, L"%02u%c%02u%c%02u", Date.wYear % 100, LocalDateSep, Date.wMonth, LocalDateSep, Date.wDay );
		break;

	case 5:
		MakeISOWeekDate( Date, isoyear, isoweeknum, isoweekday );
		wsprintf( outString, L"%04d-W%02d-%d", isoyear, isoweeknum, isoweekday );
		break;

	case 6:
		wsprintf( outString, L"%04u-%03d", Date.wYear, doy( Date ) );
		break;

	case 16:
		j = Julian( Date ) - JulDOSEpoch;

		if ( j < 0 )
			wsprintf( outString, L"%d", j );
		else
			wsprintf( outString, L"+%d", j );

		break;

	case 17:

		if ( LongDateFormat == 2 )
			wsprintf( outString, L"%s, %u %s %u", DowNameLong[Julian( Date ) % 7], Date.wDay, MonthName[Date.wMonth], Date.wYear );
		else
			wsprintf( outString, L"%s, %s %u, %u", DowNameLong[Julian( Date ) % 7], MonthName[Date.wMonth], Date.wDay, Date.wYear );

		break;

	case 18:
		GetNow( FALSE );

		j = Julian( Date );
		tj = Julian( Now );

		if ( j >= tj )
			wsprintf( outString, L"T+%d", j - tj );
		else
			wsprintf( outString, L"T%d", j - tj );

		break;

	case 19:
		wsprintf( outString, L"A+%d", Julian( Date ) );
		break;

	case 20:
		wsprintf( outString, L"%04d%02d%02d", Date.wYear, Date.wMonth, Date.wDay );
		break;

	case 21 :
		j = Julian( Date ) - JulUnixEpoch;

		if ( j < 0 )
			wsprintf( outString, L"U%d", j );
		else
			wsprintf( outString, L"U+%d", j );

		break;

	case 32:
	case 33:
	case 34:
	case 35:
	case 36 :

		if ( fmt == 33 )
			wfmt = DATE_SHORTDATE;
		else if ( fmt == 34 )
			wfmt = DATE_USE_ALT_CALENDAR;
		else if ( fmt == 35 ) {
			GetEnvironmentVariable( L"DateForm", FormatPic, SHORT_BUF_LEN );

			if ( FormatPic[0] ) {
				PicPtr = FormatPic;
				wfmt = NULL;
			}
		}
		else if ( fmt == 36 )
			wfmt = DATE_YEARMONTH;

		GetDateFormat( LOCALE_USER_DEFAULT, wfmt, &Date, PicPtr, outString, SHORT_BUF_LEN );
		break;

	case 145:
		if ( LongDateFormat == 2 )
			wsprintf( outString, L"%s, %u %s %u", LocalDowNameLong[Julian( Date ) % 7], Date.wDay, LocalMonthNameLong[Date.wMonth], Date.wYear );
		else
			wsprintf( outString, L"%s, %s %u, %u", LocalDowNameLong[Julian( Date ) % 7], LocalMonthNameLong[Date.wMonth], Date.wDay, Date.wYear );
		break;

	case 4:
	default:
		wsprintf( outString, L"%04u-%02u-%02u", Date.wYear, Date.wMonth, Date.wDay );
		break;
	}

	return 0;
}

#else	//	QCAL_only

int OutputFormattedDate( SYSTEMTIME Date, int fmt, LPTSTR outString )
{
	wsprintf( outString, L"%04u-%02u-%02u", Date.wYear, Date.wMonth, Date.wDay );
	return 0;
}

#endif	//	QCAL_only


int MakeDate( long int julian, SYSTEMTIME &Date )
{
	//  Converts a julian to SYSTEMTIME format

	int tryyear, trymonth, tryday, ord;
	BOOL leap;

	tryyear = Epoch + ( julian / 366 );				// guess at year

	while ( Julian( tryyear, 12, 31 ) < julian )	// if guess is too low,
		tryyear++;									// adjust it

	ord = julian - Julian( tryyear, 1, 1 ) + 1;
	leap = Leapness( tryyear );

	trymonth = 1;
	tryday = 1;

	for ( int i = 1; ( Kalends[i][leap] < ord ) && ( i <= 12 ); i++ )
		trymonth = i;

	tryday = ord - Kalends[trymonth][leap];

	Date.wYear		= tryyear;
	Date.wMonth		= trymonth;
	Date.wDay		= tryday;
	Date.wDayOfWeek	= julian % 7;
	return 0;
}



//	Strips off any leading or trailing spaces from a string.

int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



BOOL IsDateSep( wchar_t ch )
{
	return ( ( ch == '/' ) || ( ch == '-' ) || ( ch == '.' ) );
}


#ifndef QCAL_only

BOOL IsTimeSep( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '.' ) );
}



BOOL IsDecimalSep( wchar_t ch )
{
	return ( ( ch == '.' ) || ( ch == ',' ) );
}
#endif	//	QCAL_only



BOOL IsSign( wchar_t ch )
{
	return ( ( ch == '-' ) || ( ch == '+' ) );
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



BOOL IsCommaOrSemi( wchar_t ch )
{
	return ( ( ch == ',' ) || ( ch == ';' ) );
}



BOOL SignedNumberStart( LPTSTR String )
{
	if ( iswdigit( String[0] ) )
		return TRUE;

	if ( IsSign( String[0] ) && iswdigit( String[1] ) )
		return TRUE;

	return FALSE;
}



int MakeInitCaps( LPTSTR inString, LPTSTR outString )
{
	if ( !inString || !outString )
		return -1;

	wchar_t ch = '.';
	BOOL start = TRUE;

	for ( int i = 0; ch; i++ ) {
		ch = inString[i];
		outString[i] = start ? towupper( ch ) : ch;
		start = iswspace( ch );
	}

	return 0;
}



int GetLocalDateFormat()
{
	wchar_t shortbuf[SHORT_BUF_LEN];
	long int i = -1;


	if ( QcalDoingDelayedExpansion )
		return 1;

	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_IDATE, shortbuf, MAX_ARG_LEN ) ) {
		i = wcstol( shortbuf, NULL, 10 );
		if ( ( i >= 0 ) && ( i <= 2 ) )
			LocalDateFormat = i + 1;
	}

#ifndef QCAL_only

	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_ILDATE, shortbuf, MAX_ARG_LEN ) ) {
		i = wcstol( shortbuf, NULL, 10 );
		if ( ( i >= 0 ) && ( i <= 2 ) )
			LongDateFormat = i + 1;
	}

	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_SDATE, shortbuf, MAX_ARG_LEN ) )
		if ( IsDateSep( shortbuf[0] ) && ( shortbuf[1] == '\0' ) )
			LocalDateSep = shortbuf[0];

	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_STIME, shortbuf, MAX_ARG_LEN ) )
		if ( IsTimeSep( shortbuf[0] ) && ( shortbuf[1] == '\0' ) )
			LocalTimeSep = shortbuf[0];

	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_SDECIMAL, shortbuf, MAX_ARG_LEN ) )
		if ( IsDecimalSep( shortbuf[0] ) && ( shortbuf[1] == '\0' ) )
			LocalDecSep = shortbuf[0];

#endif	//	QCAL_only

	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_IFIRSTDAYOFWEEK, shortbuf, MAX_ARG_LEN ) ) {
		i = wcstol( shortbuf, NULL, 10 );
		LeftColumnDow = ( i + 1 ) % 7;
	}

	for ( i = 0; i < 12; i++ ) {
		GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_SABBREVMONTHNAME1 + i, LocalMonthName[i + 1], LOCAL_NAME_LEN );
		GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_SMONTHNAME1 + i, LocalMonthNameLong[i + 1], LOCAL_NAME_LEN );
		MakeInitCaps( LocalMonthNameLong[i + 1], LocalMonthNameCaps[i + 1] );
	}

	for ( i = 0; i < 7; i++ ) {
		GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_SABBREVDAYNAME1 + i, LocalDowName[(i + 1) % 7], LOCAL_NAME_LEN );
		GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_SDAYNAME1 + i, LocalDowNameLong[(i + 1) % 7], LOCAL_NAME_LEN );
	}

	GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_SENGCOUNTRY, CountryName, 128 );

	return 0;
}



int CallInternal( LPTSTR Cmd, LPTSTR Args )
{
	if ( !Cmd )
		return -1;
	else if ( Cmd[0] == '\0' )
		return -1;


	wchar_t CmdBuf[MAX_ARG_LEN]		= L"";
	wchar_t ExpVar[SHORT_BUF_LEN]	= L"%_expansion";
	BOOL AddStar					= TRUE;


	ExpandVariables( ExpVar, 1 );
	if ( wcschr( ExpVar, '1' ) )
		AddStar = FALSE;


	if ( ( Cmd[0] != '*' ) && AddStar )
		wcscpy_s( CmdBuf, MAX_ARG_LEN, L"*" );

	wcscat_s( CmdBuf, MAX_ARG_LEN, Cmd );

	if ( Args ) {
		if ( Args[0] ) {
			if ( !iswspace( Args[0] ) )
				wcscat_s( CmdBuf, MAX_ARG_LEN, L" " );

			wcscat_s( CmdBuf, MAX_ARG_LEN, Args );
		}
	}

	return Command( CmdBuf, 0 );
}



long int ParseInt( LPTSTR inString )
{
	if ( !inString )
		return -1;


	long int num1	= -1;
	wchar_t *sep1	= NULL;


	_set_errno( 0 );

	if ( ( inString[0] < '0' ) || ( inString[0] > '9' ) )
		return -1;

	num1 = wcstol( inString, &sep1, 10 );

	if ( errno )
		return -1;

	if ( sep1 )
		if ( *sep1 != '\0')
			return -1;

	return num1;
}



int ParseDow( LPTSTR inString )
{
	if ( !inString )
		return -1;


	int num			= -1;
	wchar_t *sep1	= NULL;


	if ( iswdigit( inString[0] ) ) {
		num = wcstol( inString, &sep1, 10 );

		if ( *sep1 != '\0' )
			return -1;

		if ( ( num >= 0 ) && ( num <= 7 ) )
			return num % 7;
		else
			return -1;
	}


	for ( int i = 0; i < 7; i++ )
		if ( !_wcsicmp( inString, LocalDowName[i] ) )
			return i;

	for ( int i = 0; i < 7; i++ )
		if ( !_wcsnicmp( inString, DowName[i], 2 ) )
			return i;

	return -1;
}



int ParseMonth( LPTSTR inString )
{
	if ( !inString )
		return -1;


	int num			= -1;
	wchar_t *sep1	= '\0';


	if ( iswdigit( inString[0] ) ) {
		num = wcstol( inString, &sep1, 10 );

		if ( *sep1 != '\0' )
			return -1;

		if ( ( num >= 1 ) && ( num <= 12 ) )
			return num;
		else
			return -1;
	}


	for ( int i = 1; i <= 12; i++ )
		if ( !_wcsicmp( inString, LocalMonthName[i] ) )
			return i;

	for ( int i = 1; i <= 12; i++ )
		if ( !_wcsnicmp( inString, MonthName[i], 3 ) )
			return i;

	return -1;
}



int ParseDow2( LPTSTR inString, int *len )
{
	if ( !inString )
		return -1;


	wchar_t shortbuf[MAX_ARG_LEN]	= L"";
	int i		= 0;
	wchar_t ch	= '\0';


	do {
		ch = inString[i];
		if ( iswspace( ch ) || ( ch == ';' ) || ( ch == '-' ) || ( ch == '.' ) || ( ch == ',' ) )
			ch = '\0';

		shortbuf[i++] = ch;
	}
	while ( ch );


	if ( shortbuf[0] == '\0' )
		return -1;


	*len = i - 1;

	if ( !_wcsicmp( shortbuf, L"weekday" ) || !_wcsicmp( shortbuf, L"weekdays" ) )
		return WEEKDAYS_MASK;
	else if ( !_wcsicmp( shortbuf, L"weekend" ) || !_wcsicmp(shortbuf, L"weekends" ) )
		return WEEKENDS_MASK;
	else
		return ParseDow( shortbuf );
}



int ParseDowMask( LPTSTR inString )
{
	if ( !inString )
		return -1;


	int i = 0, mask = 0;
	BOOL invert = FALSE;
	int dow = 0, dow2 = 0;
	int len = 0;


	if ( ( inString[i] == '!' ) || ( inString[i] == '^' ) ) {
		invert = TRUE;
		i++;
	}


	while ( inString[i] ) {
		dow = ParseDow2( inString + i, &len );

		if ( dow < 0 )
			return -1;
		else
			i += len;

		dow2 = -1;

		if ( ( i <= ( MAX_ARG_LEN - 4 ) ) && ( dow <= 6 ) ) {
			if ( ( inString[i] == '-' ) || ( ( inString[i] == '.' ) && ( inString[i + 1] == '.' ) ) ) {
				// parse dow range:

				i += ( inString[i] == '.' ) ? 2 : 1;

				dow2 = ParseDow2( inString + i, &len );

				if ( ( dow2 < 0 ) || ( dow2 > 6 ) )
					return -1;
				else
					i += len;

				if ( ( inString[i] == '-' ) || ( inString[i] == '.' ) )
					return -1;
			}
		}

		while ( iswspace( inString[i] ) || IsCommaOrSemi( inString[i] ) )
			i++;


		if ( dow <= 6 )
			mask |= ( 1 << dow );
		else
			mask |= dow;

		while ( ( dow2 >= 0 ) && ( dow != dow2 ) ) {
			dow = ( ( dow + 1 ) % 7 );
			mask |= ( 1 << dow );
		}
	}

	if ( invert )
		mask ^= 0x7f;		//	mask = 0x7f - mask;

	return mask;
}



int ParseYear( LPTSTR inString )
{
	if ( !inString )
		return -1;


	long int n		= 0;
	wchar_t *sep1	= NULL;


	if ( inString[0] == '\0' )
		return Now.wYear;


	if ( IsSign( inString[0] ) && iswdigit( inString[1] ) ) {
		n = wcstol(inString, &sep1, 10);
		if ( *sep1 )
			return -1;

		n = Now.wYear + n;
		if ( ( n < Epoch ) || ( n > Ultima ) )
			return -1;

		return n;
	}


	if ( !iswdigit( inString[0] ) )
		return -1;

	n = ParseInt( inString );
	if ( ( n < Epoch ) || ( n > Ultima ) )
		return -1;

	return n;
}



BOOL JulianMatchesMask( long int jul, int mask )
{
	return ( ( mask & ( 1 << ( jul % 7 ) ) ) != 0 ) ;
}



int ParseDate( LPTSTR String, int fmt, SYSTEMTIME &Date )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return -1;


	long int num1 = -1, num2 = -1, num3 = -1;
	wchar_t nullch = '\0';
	wchar_t *sep1 = &nullch, *sep2 = &nullch, *ch = NULL;
	int tryyear = 0, trymonth = 0, tryday = 0;
	long int j = 0;
	BOOL leap = FALSE;
	BOOL isoweekhack = FALSE;
	

	GetNow( FALSE );


	if ( QcalDoingDelayedExpansion )
		fmt = 4;
	else {
		fmt &= 0xff;

		if ( ( fmt < 1 ) || ( fmt > 3 ) )
			fmt = LocalDateFormat;
	}


	if ( SignedNumberStart( String ) ) {
		num1 = wcstol( String, &sep1, 10 );

		if ( IsDateSep( *sep1 ) ) {
			ch = sep1 + 1;

			if ( towupper( *ch ) == 'W' ) {
				isoweekhack = TRUE;
				ch++;
			}

			if ( !iswdigit( *ch ) )
				return -1;

			num2 = wcstol( ch, &sep2, 10 );

			if ( IsDateSep( *sep2 ) ) {
				ch = sep2 + 1;

				if ( !iswdigit( *ch ) )
					return -1;

				num3 = wcstol( ch, &ch, 10 );

				if ( *ch )
					return -1;
			}
			else if ( *sep2 )
				return -1;
		}
		else if ( *sep1 )
			return -1;
	}
	else if ( !_wcsicmp( String, L"T" ) || !_wcsicmp( String, L"TODAY" ) ) {
		Date.wYear		= Now.wYear;
		Date.wMonth		= Now.wMonth;
		Date.wDay		= Now.wDay;
		Date.wDayOfWeek	= Now.wDayOfWeek;
		return 0;
	}
	else if ( ( towupper( String[0] ) == 'T' ) && IsSign( String[1] ) && iswdigit( String[2] ) ) {
		num1 = wcstol( String + 1, &sep1, 10 );
		if ( *sep1 )
			return -1;

		num1 = Julian( Now ) + num1 - JulDOSEpoch;
	}
	else if ( !_wcsnicmp( String, L"TODAY", 5 ) && IsSign( String[5] ) && iswdigit( String[6] ) ) {
		num1 = wcstol( String + 5, &sep1, 10 );
		if ( *sep1 )
			return -1;

		num1 = Julian( Now ) + num1 - JulDOSEpoch;
	}
	else if ( ( towupper( String[0] ) == 'D' ) && SignedNumberStart( String + 1 ) ) {
		num1 = wcstol( String + 1, &sep1, 10 );
		if (*sep1 )
			return -1;

		if ( ( num1 < MinimumDayNum ) || ( num1 > MaximumDayNum ) )
			return -1;

		MakeDate( num1 + JulDOSEpoch, Date );
		return 0;
	}
	else if ( ( towupper( String[0] ) == 'A' ) && ( ( ( String[1] == '+' ) && iswdigit( String[2] ) ) || iswdigit( String[1] ) ) ) {
		num1 = wcstol( String + 1, &sep1, 10 );
		if (*sep1 )
			return -1;

		if ( ( num1 < MinimumJulian ) || ( num1 > MaximumJulian ) )
			return -1;

		MakeDate( num1, Date );
		return 0;
	}
	else if ( ( towupper( String[0] ) == 'U' ) && SignedNumberStart( String + 1 ) ) {
		num1 = wcstol( String + 1, &sep1, 10 );
		if ( *sep1 )
			return -1;

		num1 += JulUnixEpoch;
		if ( ( num1 < MinimumJulian ) || ( num1 > MaximumJulian ) )
			return -1;

		MakeDate( num1, Date );
		return 0;
	}
	else
		return -1;


	if ( isoweekhack ) {
		//  Date entered in ISO yyyy-Www-d format:

		if ( ( num1 < Epoch ) || ( num1 > Ultima ) )
			return -1;
		if ( ( num2 < 1 ) || ( num2 > 53 ) )
			return -1;
		if ( ( num3 < 1 ) || ( num3 > 7 ) )
			return -1;

		if ( ( *sep1 != '-' ) || ( *sep2 != '-' ) )
			return -1;

		tryyear = num1;
		leap = Leapness( tryyear );

		if ( num2 > (int) ISOWeeksIn( tryyear ) )
			return -1;

		int offset = ISOWeekOffset( tryyear );
		int ord = ( 7 * ( num2 - 1 ) ) + num3 + offset;

		if ( ord < 1 ) {
			tryyear--;
			trymonth = 12;
			tryday = 31 + ord;
			if ( tryyear < Epoch )
				return -1;   // Can't happen...?
		}
		else {
			for ( int i=1; ( Kalends[i][leap] < ord ) && ( i <= 12 ); i++ )
				trymonth = i;

			tryday = ord - Kalends[trymonth][leap];

			if ( tryday > 31 ) {
				tryyear++;
				trymonth = 1;
				tryday -= 31;
				if ( tryyear > Ultima )
					return -1;
			}
		}

		Date.wYear		= tryyear;
		Date.wMonth		= trymonth;
		Date.wDay		= tryday;
		Date.wDayOfWeek	= num3 % 7;
		return 0;
	}
	else if ( ( num1 >= MinimumDayNum ) && ( num1 <= MaximumDayNum ) && ( *sep1 == '\0' ) ) {
		//  Date entered as a day number:

		MakeDate( num1 + JulDOSEpoch, Date );
		return 0;
	}
	else if ( ( num1 >= ( Epoch * 10000 + 101 ) ) && ( num1 <= ( Ultima * 10000 + 1231 ) ) && ( num2 == -1 ) ) {
		//  Date entered as YYYYMMDD :

		tryyear = num1 / 10000;
		num1 -= ( tryyear * 10000 );
		trymonth = num1 / 100;
		tryday = num1 - ( 100 * trymonth );

		if ( ( trymonth < 1 ) || ( trymonth > 12 ) || ( tryday < 1 ) || ( tryday > 31 ) )
			return -1;

		if ( tryday > MonthLength[trymonth][Leapness( tryyear )] )
			return -1;

		Date.wYear		= tryyear;
		Date.wMonth		= trymonth;
		Date.wDay		= tryday;
		Date.wDayOfWeek	= Julian( tryyear, trymonth, tryday ) % 7;
		return 0;
	}
	else if ( ( num1 >= Epoch ) && ( num1 <= Ultima ) && IsDateSep( *sep1 ) && IsDateSep( *sep2 ) ) {
		//  Date starts with a four-digit year; assume yyyy/mm/dd:

		j = Julian( num1, num2, num3 );

		if ( j < 0 )
			return -1;

		Date.wYear		= (WORD) num1;
		Date.wMonth		= (WORD) num2;
		Date.wDay		= (WORD) num3;
		Date.wDayOfWeek	= j % 7;
		return 0;
	}
	else if ( ( num1 >= Epoch ) && ( num1 <= Ultima ) && ( num2 >= 1 ) && ( num2 <= 366 ) && ( *sep1 == '-' ) && ( *sep2 == '\0' ) ) {
		//  ISO ordinal-day format, yyyy-ddd:

		tryyear = num1;
		leap = Leapness( tryyear );
		if ( !leap && ( num2 > 365 ) )
			return -1;

		trymonth = 1;
		tryday = 1;

		for ( int i = 1; ( Kalends[i][leap] < num2 ) && ( i <= 12 ); i++ )
			trymonth = i;

		tryday = num2 - Kalends[trymonth][leap];

		Date.wYear		= tryyear;
		Date.wMonth		= trymonth;
		Date.wDay		= tryday;
		Date.wDayOfWeek	= Julian( tryyear, trymonth, tryday ) % 7;
		return 0;
	}
	else if ( ( num1 >= 1 ) && ( num1 <= 31 ) && ( num2 >= 1 ) && ( num2 <= 31 ) && IsDateSep( *sep1 ) && ( *sep2 == '\0' ) ) {
		// Date in m/d or d/m format; assume the current or default year:

		if ( StrictYearInput && !QcalDoingDelayedExpansion )
			return -1;

		if ( QcalDoingDelayedExpansion )
			tryyear = QcalDoingDelayedExpansion;
		else
			tryyear = Now.wYear;

		if ( fmt == 2 ) {
			trymonth = num2;
			tryday   = num1;
		}
		else {
			trymonth = num1;
			tryday   = num2;
		}

		if ( trymonth > 12 )
			return -1;

		if ( tryday > MonthLength[trymonth][Leapness( tryyear )] )
			return -1;

		Date.wYear		= tryyear;
		Date.wMonth		= trymonth;
		Date.wDay		= tryday;
		Date.wDayOfWeek	= Julian( tryyear, trymonth, tryday ) % 7;
		return 0;
	}
	else if ( IsDateSep( *sep1 ) && IsDateSep( *sep2 ) ) {
		//  Date entered as three numbers; assume local or user-
		//  specified date format:

		switch ( fmt ) {

		case 1:					//  m/d/y order
			tryyear		= num3;
			trymonth	= num1;
			tryday		= num2;
			break;

		case 2:					//  d/m/y order
			tryyear		= num3;
			trymonth	= num2;
			tryday		= num1;
			break;

		default:				//  y/m/d order
			tryyear		= num1;
			trymonth	= num2;
			tryday		= num3;
			break;
		}

		if ( ( tryyear >= 0 ) && ( tryyear <= 79 ) && !StrictYearInput )
			tryyear += 2000;
		else if ( ( tryyear >= 80 ) && ( tryyear <= 99 ) && !StrictYearInput )
			tryyear += 1900;

		if ( ( tryyear < Epoch ) || ( tryyear > Ultima ) )
			return -1;

		if ( ( trymonth < 1 ) || ( trymonth > 12 ) )
			return -1;

		leap = Leapness( tryyear );

		if ( ( tryday < 1 ) || ( tryday > MonthLength[trymonth][leap] ) )
			return -1;

		Date.wYear		= tryyear;
		Date.wMonth		= trymonth;
		Date.wDay		= tryday;
		Date.wDayOfWeek	= Julian( tryyear, trymonth, tryday ) % 7;
		return 0;
	}
	else
		return -1;
}



#ifndef QCAL_only


#include "TimeFunctions.cpp"


int ParseDateAndFormat( LPTSTR String, SYSTEMTIME &Date )
{
	if ( !String )
		return -1;


	int fmt		= 0x0100;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );

	fmt = LocalDateFormat;


	if ( arglen[1] ) {
		fmt = ParseInt( arg[1] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[0] ) {
		if ( ParseDate( arg[0], fmt, Date ) < 0) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else
		Date = Now;

	return 0;
}



DLLExports int WINAPI f_agedate( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t *sep1 = '\0';
	long long int jiffies = 0;
	int fmt = 4, timefmt = OT_SHOWSECONDS | OT_SHOWMSEC | OT_LEADINGZERO, flags = 0;
	wchar_t DateSep[] = L",";


	FILETIME FileTime;
	SYSTEMTIME fileStampUnpacked;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	if ( arglen[0] )
		jiffies = _wcstoi64( arg[0], &sep1, 10 );
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No age\r\n" );
		return -1;
	}

	if ( *sep1 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid age \"%s\"\r\n", arg[0] );
		return -1;
	}

	if ( ( jiffies < 0 ) || ( jiffies > MaximumAge ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Age out of range \"%s\"\r\n", arg[0] );
		return -1;
	}

	if ( arglen[1] ) {
		fmt = ParseInt( arg[1] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[2] ) {
		timefmt = ParseInt(arg[2]);

		if ( timefmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		flags = ParseInt( arg[3] );

		if ( flags < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad option \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	if ( flags & 1 )
		DateSep[0] = 'T';
	else if ( flags & 2 )
		DateSep[0] = '@';


	FileTime.dwHighDateTime = (DWORD) ( jiffies >> 32 );
	FileTime.dwLowDateTime  = jiffies & 0xffffffff;
	FileTimeToSystemTime( &FileTime, &fileStampUnpacked );

	OutputFormattedDate( fileStampUnpacked, fmt, String );
	wcscat_s( String, MAX_ARG_LEN, DateSep );
	OutputTime( fileStampUnpacked, timefmt, String + wcslen( String ) );
	return 0;
}



DLLExports int WINAPI f_date( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;

	wsprintf( String, L"%d", Julian( Date ) - JulDOSEpoch );
	return 0;
}



DLLExports int WINAPI f_dateplus( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;

	int fmt = 0x100;
	wchar_t nullch = '\0';
	wchar_t *sep1 = '\0';
	long int offset = 0, julian = 0;;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No date\r\n" );
		return -1;
	}

	if ( arglen[2] ) {
		fmt = ParseInt(arg[2]);

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( ParseDate( arg[0], fmt, Date ) < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[0] );
		return -1;
	}

	if ( arglen[1] ) {
		if ( SignedNumberStart( arg[1] ) ) {
			offset = wcstol(arg[1], &sep1, 10);

			if ( *sep1 != '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad offset \"%s\"\r\n", arg[1] );
				return -1;
			}
		}
		else {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad offset \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No offset\r\n" );
		return -1;
	}
	

	julian = Julian( Date ) + offset;

	if ( ( julian < MinimumJulian ) || ( julian > MaximumJulian ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Resulting date out of range\r\n" );
		return -1;
	}

	MakeDate( julian, Date );
	OutputFormattedDate( Date, fmt, String );
	return 0;
}



DLLExports int WINAPI f_day( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;

	wsprintf( String, L"%u", Date.wDay );
	return 0;
}



DLLExports int WINAPI f_days( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date1, Date2;
	int fmt = 0, flag = 0;
	long int diff = 0;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	fmt = LocalDateFormat;


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No start date\r\n" );
		return -1;
	}

	if ( arglen[1] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No end date\r\n" );
		return -1;
	}

	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( ParseDate( arg[0], fmt, Date1 ) < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid start date \"%s\"\r\n", arg[0] );
		return -1;
	}

	if ( ParseDate( arg[1], fmt, Date2 ) < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid end date \"%s\"\r\n", arg[1] );
		return -1;
	}

	if ( arglen[3] ) {
		flag = ParseInt( arg[3] );

		if ( ( flag < 0 ) || ( flag > 1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid mode \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	diff = Julian( Date2 ) - Julian( Date1 );

	if ( flag & 1 ) {
		if ( diff < 0 )
			diff = -diff;

		diff++;
	}

	wsprintf( String, L"%d", diff );
	return 0;
}



DLLExports int WINAPI f_dow( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;

	wcscpy_s( String, MAX_ARG_LEN, DowName[Julian( Date ) % 7] );
	return 0;
}



DLLExports int WINAPI f_dowf( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;

	wcscpy_s( String, MAX_ARG_LEN, DowNameLong[Julian( Date ) % 7] );
	return 0;
}



DLLExports int WINAPI f_dowi( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;

	wsprintf( String, L"%u", ( Julian( Date ) % 7 ) + 1 );
	return 0;
}



DLLExports int WINAPI f_dowis( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;
	int mask = 0;
	int fmt = 0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetNow( FALSE );
	GetLocalDateFormat();


	fmt = LocalDateFormat;

	if ( arglen[0] ) {
		mask = ParseDowMask( arg[0] );

		if ( mask <= 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid days-of-the-week list \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing days-of-the-week list\r\n" );
		return -1;
	}

	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[1] ) {
		if ( ParseDate( arg[1], fmt, Date ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else
		Date = Now;


	if ( JulianMatchesMask( Julian( Date ), mask ) )
		wcscpy_s( String, MAX_ARG_LEN, L"1" );
	else
		wcscpy_s( String, MAX_ARG_LEN, L"0" );

	return 0;
}



DLLExports int WINAPI f_doy( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;

	wsprintf( String, L"%d", doy( Date ) );
	return 0;
}



DLLExports int WINAPI f_idow( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;


	wcscpy_s( String, MAX_ARG_LEN, LocalDowName[Julian( Date ) % 7] );
	return 0;
}



DLLExports int WINAPI f_idowf( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;


	wcscpy_s( String, MAX_ARG_LEN, LocalDowNameLong[Julian( Date ) % 7] );
	return 0;
}



DLLExports int WINAPI f_inrange( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date, StartDate, EndDate;
	long int TestJul, StartJul, EndJul;
	int fmt = 0;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	fmt = LocalDateFormat;


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No date\r\n" );
		return -1;
	}

	if ( arglen[1] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No start date\r\n" );
		return -1;
	}

	if ( arglen[2] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No end date\r\n" );
		return -1;
	}

	if ( arglen[3] ) {
		fmt = ParseInt( arg[3] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	if ( ParseDate( arg[0], fmt, Date ) < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[0] );
		return -1;
	}

	if ( ParseDate( arg[1], fmt, StartDate ) < 0) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid start date \"%s\"\r\n", arg[0] );
		return -1;
	}

	if ( ParseDate( arg[2], fmt, EndDate ) < 0) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid end date \"%s\"\r\n", arg[0] );
		return -1;
	}


	TestJul		= Julian( Date );
	StartJul	= Julian( StartDate );
	EndJul		= Julian( EndDate );


	if ( ( TestJul < StartJul ) && ( TestJul < EndJul ) )
		wcscpy_s( String, SHORT_BUF_LEN, L"-1" );
	else if ( ( TestJul > StartJul ) && ( TestJul > EndJul ) )
		wcscpy_s( String, SHORT_BUF_LEN, L"1" );
	else
		wcscpy_s( String, SHORT_BUF_LEN, L"0" );

	return 0;
}



DLLExports int WINAPI f_isdate( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;
	int fmt		= 0;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );


	fmt = LocalDateFormat;


	if ( arglen[0] == 0 ) {
		wcscpy_s(String, MAX_ARG_LEN, L"0");
		return 0;
	}

	if ( arglen[1] ) {
		fmt = ParseInt( arg[1] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	if ( ParseDate( arg[0], fmt, Date ) < 0 )
		wcscpy_s(String, MAX_ARG_LEN, L"0");
	else
		wcscpy_s(String, MAX_ARG_LEN, L"1");

	return 0;
}



DLLExports int WINAPI f_istime( LPTSTR String )
{
	if ( !String )
		return -1;


	SYSTEMTIME inTime;


	if ( ParseTime( String, inTime, NULL ) )
		wcscpy_s( String, MAX_ARG_LEN, L"0" );
	else
		wcscpy_s( String, MAX_ARG_LEN, L"1" );

	return 0;
}



DLLExports int WINAPI f_timeparts( LPTSTR String )
{
	if ( !String )
		return -1;


	SYSTEMTIME inTime;
	int Flags = 0;


	if ( ParseTime( String, inTime, &Flags ) )
		wcscpy_s( String, MAX_ARG_LEN, L"0" );
	else
		wsprintf( String, L"%d", Flags );

	return 0;
}



DLLExports int WINAPI f_isholiday( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;
	unsigned short int Holidays[12][31];
	int fmt = 0;


	memset( Holidays, 0x00, sizeof( Holidays ) );


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );


	fmt = LocalDateFormat;


	if ( arglen[1] ) {
		fmt = ParseInt( arg[1] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[0] ) {
		if ( ParseDate( arg[0], fmt, Date ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else
		Date = Now;


	if ( arglen[2] )
		wcscpy_s( CountryName, 128, arg[2] );


	FreeHolidayDefs();

	if ( QcalReadHolidaysFile() <= 0 )
		QcalDefaultHolidays();

	ExpandHolidays( Date.wYear, Holidays );
	FreeHolidayDefs();
	wsprintf( String, L"%u", Holidays[Date.wMonth - 1][Date.wDay - 1] );
	return 0;
}



/*
DLLExports int WINAPI f_holidayname( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;
	unsigned short int Holidays[12][31];
	int fmt = 0;

	long int julian			= 0;
	_HolidayDef *This		= NULL;
	_HolidayDef *Found		= NULL;


	memset( Holidays, 0x00, sizeof( Holidays ) );


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );


	fmt = LocalDateFormat;


	if ( arglen[1] ) {
		fmt = ParseInt( arg[1] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[0] ) {
		if ( ParseDate( arg[0], fmt, Date ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else
		Date = Now;


	if ( arglen[2] )
		wcscpy_s( CountryName, 128, arg[2] );


	wcscpy_s( String, MAX_ARG_LEN, L"-" );


	FreeHolidayDefs();

	if ( QcalReadHolidaysFile() <= 0 )
		QcalDefaultHolidays();

	julian = Julian( Date );
	ExpandHolidays( Date.wYear, Holidays );

	if ( FirstHolidayDef ) {
		This = FirstHolidayDef;

		while ( This ) {
			if ( This->Enabled && ( This->Julian == julian ) && This->Name[0] ) {
				wcscpy_s( String, MAX_ARG_LEN, This->Name );
				break;
			}

			This = This->Next;
		}
	}

	FreeHolidayDefs();
	return 0;
}
*/



DLLExports int WINAPI f_isodowi( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;


	wsprintf( String, L"%u", ( ( Julian( Date ) + 6 ) % 7 ) + 1 );
	return 0;
}



DLLExports int WINAPI f_isoweek( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;
	unsigned int WeekYear = 0, WeekNum = 0, WeekDay = 0;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;


	MakeISOWeekDate( Date, WeekYear, WeekNum, WeekDay );
	wsprintf( String, L"%u", WeekNum );
	return 0;
}



DLLExports int WINAPI f_isowyear( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;
	unsigned int WeekYear = 0, WeekNum = 0, WeekDay = 0;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;


	MakeISOWeekDate( Date, WeekYear, WeekNum, WeekDay );
	wsprintf( String, L"%u", WeekYear );
	return 0;
}



DLLExports int WINAPI f_julian( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;

	wsprintf( String, L"%d", Julian( Date ) );
	return 0;
}



DLLExports int WINAPI f_leap( LPTSTR String )
{
	if ( !String )
		return 1;


	int year = 0;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetNow( FALSE );


	if ( arglen[0] ) {
		year = ParseYear( arg[0] );

		if ( year < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid year \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else
		year = Now.wYear;


	wsprintf( String, L"%u", Leapness( year ) ? 1 : 0 );
	return 0;
}



DLLExports int WINAPI f_localdatetime( LPTSTR String )
{
	if ( !String )
		return -1;


	SYSTEMTIME UTCTime, LocalTime;

	int fmt = 0x100, timefmt = OT_SHOWSECONDS | OT_SHOWMSEC;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetSystemTime( &UTCTime );


	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		timefmt = ParseInt( arg[3] );

		if ( timefmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time format \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	if ( arglen[0] ) {
		if ( ParseDate( arg[0], fmt, UTCTime ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[0] );
			return -1;
		}
	}

	if ( arglen[1] ) {
		if ( ParseTime( arg[1], UTCTime, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	if ( !SystemTimeToTzSpecificLocalTime( NULL, &UTCTime, &LocalTime ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error converting UTC to local\r\n" );
		return -1;
	}


	OutputFormattedDate( LocalTime, fmt, String );
	wcscat_s( String, MAX_ARG_LEN, L"," );
	OutputTime( LocalTime, timefmt, String + wcslen( String ) );
	return 0;
}



DLLExports int WINAPI f_utcdatetime( LPTSTR String )
{
	if ( !String )
		return -1;


	SYSTEMTIME LocalTime, UTCTime;

	int fmt = 0x100, timefmt = OT_SHOWSECONDS | OT_SHOWMSEC;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetLocalTime( &LocalTime );


	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		timefmt = ParseInt( arg[3] );

		if ( timefmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time format \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	if ( arglen[0] ) {
		if ( ParseDate( arg[0], fmt, LocalTime ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[0] );
			return -1;
		}
	}

	if ( arglen[1] ) {
		if ( ParseTime( arg[1], LocalTime, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	if ( !TzSpecificLocalTimeToSystemTime( NULL, &LocalTime, &UTCTime ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error converting local to UTC\r\n" );
		return -1;
	}


	OutputFormattedDate( UTCTime, fmt, String );
	wcscat_s( String, MAX_ARG_LEN, L"," );
	OutputTime( UTCTime, timefmt, String + wcslen( String ) );
	return 0;
}



DLLExports int WINAPI f_makeage( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME DateTime;
	FILETIME Age;
	int fmt		= 0;


	memset( &DateTime, 0x00, sizeof( SYSTEMTIME ) );


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();

	fmt = LocalDateFormat;


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No date\r\n" );
		return -1;
	}

	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( ParseDate( arg[0], fmt, DateTime ) < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[0] );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseTime( arg[1], DateTime, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	SystemTimeToFileTime( &DateTime, &Age );

	wsprintf( String, L"%I64u", ( (unsigned __int64) Age.dwHighDateTime << 32 ) | Age.dwLowDateTime );
	return 0;
}



DLLExports int WINAPI f_makedate( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;

	long int jul = 0;
	BOOL valid = FALSE;
	wchar_t *chtemp = L'\0';


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No day number\r\n" );
		return -1;
	}
	else if ( SignedNumberStart( arg[0] ) ) {
		jul = wcstol(arg[0], &chtemp, 10);
		jul += JulDOSEpoch;

		if ( ( jul >= MinimumJulian ) && ( jul <= MaximumJulian ) && ( *chtemp == L'\0' ) )
			valid = TRUE;
	}
	
	if ( !valid ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid day number \"%s\"\r\n", arg[0] );
		return -1;
	}

	int fmt = LocalDateFormat;

	if ( arglen[1] ) {
		fmt = ParseInt( arg[1] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	MakeDate( jul, Date );
	OutputFormattedDate( Date, fmt, String );
	return 0;
}



DLLExports int WINAPI f_dateconv( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	int infmt = LocalDateFormat, fmt = LocalDateFormat;


	if ( arglen[1] ) {
		fmt = infmt = ParseInt( arg[1] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No date\r\n" );
		return -1;
	}

	if ( ParseDate( arg[0], fmt, Date ) < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[0] );
		return -1;
	}


	OutputFormattedDate( Date, fmt, String );
	return 0;
}



DLLExports int WINAPI f_month( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;


	wsprintf( String, L"%u", Date.wMonth );
	return 0;
}



DLLExports int WINAPI f_monthlen( LPTSTR String )
{
	if ( !String )
		return 1;


	int month = 0, year = 0;


	GetNow( FALSE );
	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	if ( arglen[0] == 0 )
		month = Now.wMonth;
	else
		month = ParseMonth( arg[0] );


	if ( arglen[1] == 0 )
		year = Now.wYear;
	else if ( !_wcsicmp( arg[1], L"0" ) || !_wcsicmp( arg[1], L"1" ) )
		year = ParseInt( arg[1] );
	else
		year = ParseYear( arg[1] );


	if ( ( month < 1 ) || ( month > 12 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid month \"%s\"\r\n", arg[0] );
		return -1;
	}

	if ( year < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid year \"%s\"\r\n", arg[1] );
		return -1;
	}

	if ( year > 1 )
		year = Leapness(year);

	wsprintf( String, L"%d", MonthLength[month][year] );
	return 0;
}



DLLExports int WINAPI f_monthname( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned int month = 0;
	int flags = 0;


	GetNow( FALSE );
	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	if ( arglen[0] == 0 )
		month = Now.wMonth;
	else
		month = ParseMonth( arg[0] );


	if ( arglen[1] ) {
		flags = ParseInt( arg[1] );

		if ( flags < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid arg \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( ( month < 1 ) || ( month > 12 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid month \"%s\"\r\n", arg[0] );
		return -1;
	}


	switch ( flags & 3 ) {

	case 1 :
		wcscpy_s( String, MAX_ARG_LEN, MonthName[month] );
		String[3] = '\0';
		break;

	case 2 :
		wcscpy_s( String, MAX_ARG_LEN, LocalMonthNameLong[month] );
		break;

	case 3 :
		wcscpy_s( String, MAX_ARG_LEN, LocalMonthName[month] );
		break;

	default:
		wcscpy_s( String, MAX_ARG_LEN, MonthName[month] );
	}

	return 0;
}



DLLExports int WINAPI f_nextdate( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;
	int fmt = 0x100, year = 0, month = 0, day = 0;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );


	if ( arglen[0] ) {
		month = ParseMonth( arg[0] );

		if ( month < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad month \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No month\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		day = ParseInt( arg[1] );

		if ( ( day < 1 ) || ( day > MonthLength[month][1] ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad day \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No day\r\n" );
		return -1;
	}

	if ( arglen[3] ) {
		fmt = ParseInt( arg[3] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	if ( arglen[2] ) {
		if ( ParseDate( arg[2], fmt, Date ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[2] );
			return -1;
		}
	}
	else
		Date = Now;


	year = Date.wYear;

	if ( ( (unsigned) month < Date.wMonth ) || ( ( month == Date.wMonth ) && ( (unsigned) day <= Date.wDay ) ) )
		year++;

	while ( ( month == 2 ) && ( day == 29 ) && ( Leapness( year ) == 0 ) )
		year++;

	if ( year > Ultima ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Resulting date out of range\r\n" );
		return -1;
	}


	Date.wYear		= year;
	Date.wMonth		= month;
	Date.wDay		= day;
	OutputFormattedDate( Date, fmt, String );
	return 0;
}



#include "FileStamp.cpp"



DLLExports int _useconds( LPTSTR String )
{
	if ( !String )
		return -1;


	__int64 BigSeconds		= 0;


	GetNow( TRUE );

	BigSeconds = (unsigned __int64) ( Julian( Now ) - JulUnixEpoch ) * 24 * 60 * 60;
	BigSeconds += ( Now.wHour * 60 * 60 ) + ( Now.wMinute * 60 ) + Now.wSecond;

	wsprintf( String, L"%I64d", BigSeconds );
	return 0;
}



DLLExports int f_useconds( LPTSTR String )
{
	if ( !String )
		return -1;


	SYSTEMTIME DateTime;
	SYSTEMTIME UtcDateTime;
	__int64 BigSeconds			= 0;

	int fmt						= 0x100;
	int Flags					= 0;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format: \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		Flags = ParseInt( arg[3] );

		if ( Flags < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid flags value: \"%s\"\r\n", arg[3] );
			return -1;
		}
	}


	if ( Flags & 1 )
		GetSystemTime( &DateTime );
	else
		GetLocalTime( &DateTime );


	if ( arglen[0] ) {
		if ( ParseDate( arg[0], fmt, DateTime ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date: \"%s\"\r\n", arg[0] );
			return -1;
		}
	}

	if ( arglen[1] ) {
		if ( ParseTime( arg[1], DateTime, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	if ( Flags & 1 )
		UtcDateTime = DateTime;
	else {
		if ( !TzSpecificLocalTimeToSystemTime( NULL, &DateTime, &UtcDateTime ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Local-to-UTC time conversion failed\r\n" );
			return -1;
		}

		if ( ( UtcDateTime.wYear < Epoch ) || ( UtcDateTime.wYear > Ultima ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Resulting UTC time out of range: %04u-%02u-%02u\r\n", UtcDateTime.wYear, UtcDateTime.wMonth, UtcDateTime.wDay );
			return -1;
		}
	}


	BigSeconds = (unsigned __int64) ( Julian( UtcDateTime ) - JulUnixEpoch ) * 24 * 60 * 60;
	BigSeconds += ( UtcDateTime.wHour * 60 * 60 ) + ( UtcDateTime.wMinute * 60 ) + UtcDateTime.wSecond;

	wsprintf( String, L"%I64d", BigSeconds );
	return 0;
}



DLLExports int WINAPI f_uxtodatetime( LPTSTR String )
{
	if ( !String )
		return -1;


	__int64 BigSeconds				= 0;
	wchar_t *sep1					= NULL;
	int DateFmt						= 4;
	int TimeFmt						= OT_SHOWSECONDS | OT_LEADINGZERO;
	int Flags						= 0;

	const __int64 MinSeconds		= -11644473600;
	const __int64 MaxSeconds		= 253402300799;

	SYSTEMTIME DateTime;
	SYSTEMTIME LocalDateTime;

	long int Julian					= 0;

	size_t len						= 0;



	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	if ( arglen[0] ) {
		_set_errno( 0 );

		if ( ( arg[0][0] == '0' ) && ( towupper( arg[0][1] ) == 'X' ) )
			BigSeconds = _wcstoi64( arg[0], &sep1, 16 );
		else if ( SignedNumberStart( arg[0] ) )
			BigSeconds = _wcstoi64( arg[0], &sep1, 10 );
		else
			BigSeconds = LLONG_MAX;

		if ( sep1 )
			if ( *sep1 )
				BigSeconds = LLONG_MAX;

		if ( errno || ( BigSeconds == LLONG_MAX ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad seconds value: \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( ( BigSeconds < MinSeconds ) || ( BigSeconds > MaxSeconds ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Seconds out of range: %I64u\r\n", BigSeconds );
			//	Qprintf( ERRHANDLE, L"   Must be %I64d - %I64d\r\n", MinSeconds, MaxSeconds );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing seconds\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		Flags = ParseInt( arg[1] );

		if ( Flags < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags value: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[2] ) {
		DateFmt = ParseInt( arg[2] );

		if ( DateFmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad date format: \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		TimeFmt = ParseInt( arg[3] );

		if ( TimeFmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad time format: \"%s\"\r\n", arg[3] );
			return -1;
		}
	}


	BigSeconds -= MinSeconds;

	Julian = (long int) ( BigSeconds / 86400 + 1 );
	BigSeconds %= 86400;

	MakeDate( Julian, DateTime );

	DateTime.wHour			= (unsigned int) ( BigSeconds / 3600 );
	BigSeconds				%= 3600;
	DateTime.wMinute		= (unsigned int) ( BigSeconds / 60 );
	DateTime.wSecond		= (unsigned int) ( BigSeconds % 60 );
	DateTime.wMilliseconds	= 0;


	if ( Flags & 1 ) {
		//	return as utc date and time :

		OutputFormattedDate( DateTime, DateFmt, String );
	}
	else {
		//	return as local date and time :

		if ( !SystemTimeToTzSpecificLocalTime( NULL, &DateTime, &LocalDateTime ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"UTC-to-local time conversion failed\r\n" );
			return -1;
		}

		if ( ( LocalDateTime.wYear < 1601 ) || ( LocalDateTime.wYear > 9999 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Resulting local time out of range: %04u-%02u-%02u %02u:%02u:%02u\r\n", LocalDateTime.wYear, LocalDateTime.wMonth, LocalDateTime.wDay, LocalDateTime.wHour, LocalDateTime.wMinute, LocalDateTime.wSecond );
			return -1;
		}

		OutputFormattedDate( LocalDateTime, DateFmt, String );
	}


	if ( ( Flags & 128 ) == 0 ) {

		switch ( Flags & 768 ) {

		case 256 :
			wcscat_s( String, MAX_ARG_LEN, L"@" );
			break;

		case 512 :
			wcscat_s( String, MAX_ARG_LEN, L"T" );
			break;

		case 0 :
			wcscat_s( String, MAX_ARG_LEN, L" " );
		}
	}


	len = wcslen( String );

	if ( Flags & 1 )
		OutputTime( DateTime, TimeFmt, String + len );
	else
		OutputTime( LocalDateTime, TimeFmt, String + len );

	return 0;
}



DLLExports int WINAPI f_usweek( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;

	int WeekStart = 0, WeekEnd = 6;
	int Jan1dow = 0, Week2Start = 0, DoY = 0;
	int Dec31dow = 0, FinalWeekEnds = 0;
	BOOL leap = FALSE;
	int usweek = 0;
	int flags = 0;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;

	if ( arglen[2] ) {
		WeekStart = ParseDow( arg[2] );

		if ( WeekStart < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad day-of-the-week \"%s\"\r\n", arg[2] );
			return -1;
		}
	}
	else
		WeekStart = 0;

	if ( arglen[3] ) {
		flags = ParseInt( arg[3] );

		if ( flags < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags \"%s\"\r\n", arg[3] );
			return -1;
		}
	}


	WeekEnd = ( WeekStart + 6 ) % 7;

	leap = Leapness( Date.wYear );

	Jan1dow = Julian( Date.wYear, 1, 1 ) % 7;
	Week2Start = 8 - ( ( Jan1dow - WeekStart + 7 ) % 7 );
	DoY = doy( Date );

	if ( leap ) {
		Dec31dow = ( Jan1dow + 1 ) % 7;
		FinalWeekEnds = 366;
	}
	else {
		Dec31dow = Jan1dow;
		FinalWeekEnds = 365;
	}

	if ( Dec31dow != WeekEnd )
		FinalWeekEnds = FinalWeekEnds + ( ( WeekEnd - Dec31dow + 7 ) % 7 ) - 7;

	usweek = ( DoY - Week2Start + 14 ) / 7;

	if ( ( ( flags & 0x01 ) == 0 ) && ( DoY > FinalWeekEnds ) )
		usweek = 1;

	wsprintf( String, L"%d", usweek );
	return 0;
}



DLLExports int WINAPI f_year( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;


	if ( ParseDateAndFormat( String, Date ) < 0 )
		return -1;


	wsprintf( String, L"%u", Date.wYear );
	return 0;
}



DLLExports int WINAPI _isodowi( LPTSTR String )
{
	if ( !String )
		return 1;


	GetNow( FALSE );

	wsprintf( String, L"%u", ( ( Julian( Now ) + 6 ) % 7 ) + 1 );
	return 0;
}



DLLExports int WINAPI _isordate( LPTSTR String )
{
	if ( !String )
		return 1;


	GetNow( FALSE );

	OutputFormattedDate( Now, 6, String );
	return 0;
}



DLLExports int WINAPI _isowdate( LPTSTR String )
{
	if ( !String )
		return 1;


	GetNow( FALSE );

	OutputFormattedDate( Now, 5, String );
	return 0;
}



DLLExports int WINAPI _isoweek( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned int WeekYear = 0, WeekNum = 0, WeekDay = 0;


	GetNow( FALSE );

	MakeISOWeekDate( Now, WeekYear, WeekNum, WeekDay );
	wsprintf( String, L"%u", WeekNum );
	return 0;
}



DLLExports int WINAPI _isowyear( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned int WeekYear = 0, WeekNum = 0, WeekDay = 0;


	GetNow( FALSE );

	MakeISOWeekDate( Now, WeekYear, WeekNum, WeekDay );
	wsprintf( String, L"%u", WeekYear );
	return 0;
}



//   f_pickdate is implemented in ....
#include "popcal\popcal.cpp"



int NewEvaluate( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return -1;


	wchar_t temp[MAX_ARG_LEN]	= L"";
	HANDLE OldStderrHandle		= GetStdHandle( STD_ERROR_HANDLE );
	HANDLE OldStdoutHandle		= GetStdHandle( STD_OUTPUT_HANDLE );
	HANDLE NewOutputHandle		= CreateFile( L"nul", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL );
	int rv						= 0;


	if ( NewOutputHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, NewOutputHandle );
		SetStdHandle( STD_OUTPUT_HANDLE, NewOutputHandle );
	}

	wcscpy_s( temp, MAX_ARG_LEN, L"%@eval[" );
	wcscat_s( temp, MAX_ARG_LEN, String );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	ExpandVariables( temp, 1 );

	if ( NewOutputHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, OldStderrHandle );
		SetStdHandle( STD_OUTPUT_HANDLE, OldStdoutHandle );
		CloseHandle( NewOutputHandle );
	}

	if ( rv )
		return 1;
	else if ( temp[0] == '\0' )
		return 1;

	wcscpy_s( String, MAX_ARG_LEN, temp );
	return 0;
}



int InsertCommas( LPTSTR Buffer, unsigned int BufferLen )
{
	if ( !Buffer )
		return -1;

	if ( wcslen( Buffer ) > ( MAX_ARG_LEN - 10 ) )
		return 1;

	wchar_t temp[MAX_ARG_LEN]	= L"%@comma[";
	wcscat_s( temp, MAX_ARG_LEN, Buffer );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	if ( ExpandVariables( temp, 1 ) )
		return 2;

	if ( wcslen( temp ) >= BufferLen )
		return 3;

	wcscpy_s( Buffer, BufferLen, temp );
	return 0;
}



DLLExports int WINAPI f_ord( LPTSTR String )
{
	if ( !String )
		return 1;


	long int num = 0, n = 0;
	size_t i = 0;
	wchar_t *sep1 = '\0';
	const wchar_t ordSuffix[4][3] = { L"th", L"st", L"nd", L"rd" };


	for ( size_t i = 0; String[i]; i++ )
		if ( !iswspace( String[i] ) )
			n++;

	if ( !n ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No expression\r\n" );
		return -1;
	}

	if ( NewEvaluate( String ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad expression \"%s\"\r\n", String );
		return -1;
	}


	_set_errno( 0 );
	num = wcstol( String, &sep1, 10 );


	if ( *sep1 != '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid number \"%s\"\r\n", String );
		return -1;
	}
	else if ( num < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Number must be nonnegative \"%d\"\r\n", num );
		return -1;
	}
	else if ( errno ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Numeric overflow: \"%s\"\r\n", String );
		return -1;
	}


	n = num % 100;
	
	if ( n > 19 )
		n %= 10;

	if ( n > 3 )
		n = 0;


	wsprintf( String, L"%d", num );
	InsertCommas( String, MAX_ARG_LEN );			//		AddCommas( String );
	wcscat_s( String, MAX_ARG_LEN, ordSuffix[n] );
	return 0;
}



int RomanLetterValue( wchar_t ch )
{
	switch ( towupper( ch ) ) {

	case 'I' :
		return 1;

	case 'V' :
		return 5;

	case 'X' :
		return 10;

	case 'L' :
		return 50;

	case 'C' :
		return 100;

	case 'D' :
		return 500;

	case 'M' :
		return 1000;

	default :
		return -1;
	}
}



DLLExports int WINAPI f_roman( LPTSTR String )
{
	if ( !String )
		return 1;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );

	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No number\r\n" );
		return -1;
	}


	if ( !iswalpha( arg[0][0] ) ) {

		long int num = 0;
		wchar_t *sep1 = '\0';
		int i = 0, j = 0, k = 0;

		const wchar_t romNumerals[4][10][5] = {
			{ L"", L"I", L"II", L"III", L"IV", L"V", L"VI", L"VII", L"VIII", L"IX" },
			{ L"", L"X", L"XX", L"XXX", L"XL", L"L", L"LX", L"LXX", L"LXXX", L"XC" },
			{ L"", L"C", L"CC", L"CCC", L"CD", L"D", L"DC", L"DCC", L"DCCC", L"CM" },
			{ L"", L"M", L"MM", L"MMM", L"MMMM" } };

		num = wcstol(arg[0], &sep1, 10);

		if ( *sep1 != '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid number \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( ( num < 1 ) || ( num >= 5000 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Number out of range \"%s\"\r\n", arg[0] );
			return -1;
		}

		String[0] = '\0';
		k = 1000;

		for ( i = 3; i >= 0; i-- ) {
			j = num / k;
			num -= ( j * k );

			if ( j )
				wcscat_s(String, MAX_ARG_LEN, romNumerals[i][j]);

			k /= 10;
		}

		return 0;
	}
	else {

		wchar_t ch = '\0';
		unsigned int i = 0;
		int letterval = 0, subval = 0, val = 0, nextval = 0, subtrahend = 0;

		while ( arg[0][i] != '\0' ) {

			ch = towupper( arg[0][i] );

			i++;
			if ( ( ch == 'J' ) && ( arg[0][i] == '\0' ) )
				ch = 'I';

			subval = letterval = RomanLetterValue(ch);

			if ( letterval < 0 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad Roman numeral \"%s\"\r\n", arg[0] );
				return -1;
			}

			while ( towupper( arg[0][i] ) == ch ) {
				i++;
				subval += letterval;
			}

			nextval = RomanLetterValue( arg[0][i] );

			if ( nextval > letterval ) {
				subtrahend = subval - subtrahend;
			}
			else {
				val = val + subval - subtrahend;
				subtrahend = 0;

				if ( val <= 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad Roman numeral \"%s\"\r\n", arg[0] );
					return -1;
				}
			}
		}

		wsprintf( String, L"%d", val );
		return 0;
	}
}



DLLExports int WINAPI f_century( LPTSTR String )
{
	if ( !String )
		return -1;


	const wchar_t ordSuffix[4][3]		= { L"th", L"st", L"nd", L"rd" };
	
	int fmt						= 0x100;
	unsigned int Year			= 0;
	unsigned int Century		= 0;
	int Flags					= 0;
	SYSTEMTIME Date;
	wchar_t *sep1				= NULL;
	int n						= 0;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );


	if ( arglen[1] ) {
		fmt = ParseInt( arg[1] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[2] ) {
		Flags = ParseInt( arg[2] );

		if ( Flags < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid flags value: \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[0] ) {

		if ( iswdigit( arg[0][0] ) ) {
			_set_errno( 0 );

			Year = wcstoul( arg[0], &sep1, 10 );

			if ( errno )
				Year = 0;

			if ( sep1 )
				if ( *sep1 )
					Year = 0;
			
			if ( Year ) {
				if ( ( Year < Epoch ) || ( Year > Ultima ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Year out of range: %u\r\n", Year );
					return -1;
				}
			}
		}

		if ( !Year ) {
			if ( ParseDate( arg[0], fmt, Date ) < 0 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Invalid date: \"%s\"\r\n", arg[0] );
				return -1;
			}

			Year = Date.wYear;
		}
	}
	else
		Year = Now.wYear;


	Century = ( Year + 99 ) / 100;

	_ultow_s( Century, String, SHORT_BUF_LEN, 10 );

	if ( Flags & 2 )
		f_roman( String );

	if ( Flags & 1 ) {
		n = Century % 100;

		if ( n > 19 )
			n %= 10;

		if ( n > 3 )
			n = 0;

		wcscat_s( String, SHORT_BUF_LEN, ordSuffix[n] );
	}

	return 0;
}



/*
int Random( int Min, int Max )
{
	unsigned int r			= 0;


	if ( Min > Max ) {
		int t = Min;
		Min = Max;
		Max = t;
	}


	rand_s( &r );

	return Min + ( r % ( Max - Min + 1 ) );
}



DLLExports int WINAPI f_randdate( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;
	int StartYear			= Epoch;
	int EndYear				= Ultima;
	int fmt					= 4;


	memset( &Date, 0x00, sizeof( SYSTEMTIME ) );


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		StartYear = ParseInt( arg[0] );

		if ( ( StartYear < Epoch ) || ( StartYear > Ultima ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad start year: \"%s\"\r\n", arg[0] );
			return -1;
		}
	}

	if ( arglen[1] ) {
		EndYear = ParseInt( arg[1] );

		if ( ( EndYear < StartYear ) || ( EndYear > Ultima ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad end year: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad format: \"%s\"\r\n", arg[2] );
			return -1;
		}
	}


	Date.wYear	= (WORD) Random( StartYear, EndYear );
	Date.wMonth	= (WORD) Random( 1, 12 );
	Date.wDay	= (WORD) Random( 1, MonthLength[Date.wMonth][Leapness( Date.wYear )] );

	OutputFormattedDate( Date, fmt, String );
	return 0;
}



DLLExports int WINAPI f_randtime( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Time;
	int tfmt				= OT_SHOWSECONDS | OT_SHOWMSEC | OT_LEADINGZERO;


	memset( &Time, 0x00, sizeof( SYSTEMTIME ) );


	TrimLine( String );


	if ( iswdigit( String[0] ) ) {
		tfmt = ParseInt( String );

		if ( tfmt < 0 ) {
			_wcsupr_s( String, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad format: \"%s\"\r\n", String );
			return -1;
		}
	}


	Time.wHour		= (WORD) Random( 0, 23 );
	Time.wMinute	= (WORD) Random( 0, 59 );
	Time.wSecond	= (WORD) Random( 0, 59 );
	Time.wMilliseconds	= (WORD) Random( 0, 999 );

	OutputTime( Time, tfmt, String );
	return 0;
}
*/



int DisableOneFeature( LPTSTR FeaturesList, LPTSTR word )
{
	if ( !FeaturesList || !word )
		return -1;

	if ( ( FeaturesList[0] == '\0' ) || ( word[0] == '\0' ) )
		return -1;

	size_t i = 0, j = 0, base = 0;
	wchar_t ch = '\0', ch2 = '\0';
	BOOL done = FALSE, match = FALSE;

	do {
		// Compare word in 'word' against word at 'base' :

		i = 0;
		match = TRUE;

		do {
			ch = towupper(FeaturesList[base + i]);
			ch2 = ch;
			if ( IsCommaOrSemi(ch2) )
				ch2 = '\0';

			match &= ( ch2 == towupper(word[i]) );

			i++;
		} while ( ch2 != '\0' );

		if ( match ) {
			// disable the feature:

			if ( ch == '\0' )
				FeaturesList[base] = ch;
			else {
				j = 0;

				do {
					ch = FeaturesList[base + i + j];
					FeaturesList[base + j] = ch;
					j++;
				} while ( ch != '\0' );
			}
		}
		else
			base += i;

	} while ( ch != '\0' );

	i = wcslen(FeaturesList);
	if ( i > 0 )
		if ( IsCommaOrSemi(FeaturesList[i - 1]) )
			FeaturesList[i - 1] = '\0';

	return 0;
}



int DisableFeatures( LPTSTR FeaturesList )
{
	if ( !FeaturesList )
		return -1;

	const int MaxLen = 8192;
	wchar_t envvar[MaxLen] = L"";
	wchar_t word[MaxLen] = L"";
	int i = 0, j = 0;
	wchar_t ch = '\0';
	BOOL done = FALSE;

	if ( !GetEnvironmentVariable( L"Disable_Features", envvar, MaxLen ) )
		return 0;

	do {
		ch = envvar[i];
		done = ( ch == '\0' );
		if ( ( ch == ',' ) || ( ch == ';' ) || ( iswspace(ch) != 0 ) )
			ch = '\0';
		word[j] = ch;
		i++;
		j++;

		if ( ch == '\0' ) {
			DisableOneFeature( FeaturesList, word );
			j = 0;
			word[j] = ch;
		}
	} while ( !done );

	return 0;
}



#endif	//	QCAL_only



int QcalDefaultColors()
{

	HKEY UserKeyHandle = NULL, SystemKeyHandle = NULL;
	wchar_t RegValueName[SHORT_BUF_LEN] = L"";
	BYTE bytesbuf[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	DWORD bufsize = 0;


	RegOpenKeyEx( HKEY_CURRENT_USER,  PluginRegKey, 0, KEY_READ, &UserKeyHandle );
	RegOpenKeyEx( HKEY_LOCAL_MACHINE, PluginRegKey, 0, KEY_READ, &SystemKeyHandle );

	for ( int i = 0; i < 16; i++ ) {
		for ( int j = 0; j < 5; j++ )
			DefColors[i][j] = 0x11;

		wsprintf( RegValueName, L"QCal default colors %X", i );

		bufsize = sizeof( bytesbuf );

		if ( RegQueryValueEx( UserKeyHandle, RegValueName, NULL, NULL, (LPBYTE) &bytesbuf, &bufsize ) == ERROR_SUCCESS ) {
			for ( int j = 0; j < 5; j++ )
				DefColors[i][j] = bytesbuf[j];
		}
		else {
			bufsize = sizeof( bytesbuf );

			if ( RegQueryValueEx( SystemKeyHandle, RegValueName, NULL, NULL, (LPBYTE) &bytesbuf, &bufsize ) == ERROR_SUCCESS ) {
				for ( int j = 0; j < 5; j++ )
					DefColors[i][j] = bytesbuf[j];
			}
		}

	}

	if ( UserKeyHandle )
		RegCloseKey( UserKeyHandle );

	if ( SystemKeyHandle )
		RegCloseKey( SystemKeyHandle );

	return 0;
}



int MakeEaster( int year, int church, long int offset, SYSTEMTIME &Date )
{
	int month = 0, day = 0;
	int C = 0, G = 0, H = 0, I = 0, J = 0, L = 0, n = 0;
	long int julian = 0;


	if ( ( year < Epoch ) || ( year > Ultima ) || ( church < 0 ) || ( church > 1 ) )
		return -1;

	//  Stolen shamelessly from Claus Tondering's Calendar FAQ....

	switch ( church ) {

	case 1 :
		G = year % 19;
		I = ( ( 19 * G ) + 15 ) % 30;
		J = ( year + ( year / 4 ) + I ) % 7;
		L = I - J;

		month = 3 + ( (L + 40) / 44 );
		day = L + 28 - ( 31 * ( month / 4 ) );

		//  add a few days to convert julian calendar to gregorian :

		n = 10 + ( ( year - 1600 ) / 100 ) - ( ( year - 1600 ) / 400 );
		day += n;

		if ( day > MonthLength[month][0] ) {
			day -= MonthLength[month][0];
			month++;
		}

		break;

	default :
		G = year % 19;
		C = year / 100;
		H = ( C - (C / 4) - ( ( (8 * C) + 13 ) / 25 ) + (19 * G) + 15 ) % 30;
		I = H - (H / 28) * ( 1 - ( 29 / (H + 1) ) * ( ( 21 - G ) / 11 ) );
		J = ( year + ( year / 4 ) + I + 2 - C + ( C / 4 ) ) % 7;
		L = I - J;

		month = 3 + ( (L + 40) / 44 );
		day = L + 28 - ( 31 * ( month / 4 ) );
	}

	Date.wYear	= year;
	Date.wMonth	= month;
	Date.wDay	= day;


	if ( offset ) {
		julian = Julian( Date ) + offset;

		if ( ( julian < MinimumJulian ) || ( julian > MaximumJulian ) )
			return -1;
		else
			MakeDate( julian, Date );
	}

	return 0;
}



DLLExports int WINAPI f_easter( LPTSTR String )
{
	if ( !String )
		return 1;


	int year = 0, church = 0, fmt = 0x100, testrv = 0;
	SYSTEMTIME Date, TestDate;
	long int offset = 0;
	BOOL FindNext = FALSE;
	wchar_t *sep1 = '\0';


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetNow( FALSE );


	if ( QcalDoingDelayedExpansion )
		year = QcalDoingDelayedExpansion;
	else
		year = Now.wYear;


	if ( arglen[0] ) {
		if ( !_wcsicmp( arg[0], L"*" ) )
			FindNext = TRUE;
		else
			year = ParseYear( arg[0] );

		if ( year < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid year \"%s\"\r\n", arg[0] );
			return -1;
		}
	}

	if ( arglen[1] ) {
		church = ParseInt( arg[1] );

		if ( ( church < 0 ) || ( church > 1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid church \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		if ( SignedNumberStart( arg[3] ) ) {
			offset = wcstol( arg[3], &sep1, 10 );

			if ( *sep1 != '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad offset \"%s\"\r\n", arg[3] );
				return -1;
			}
		}
		else {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad offset \"%s\"\r\n", arg[3] );
			return -1;
		}

		if ( ( offset < -366 ) || ( offset > 366 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad offset \"%s\"\r\n", arg[3] );
			return -1;
		}
	}
	

	if ( MakeEaster( year, church, offset, Date ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Resulting date out of range\r\n" );
		return -1;
	}

	if ( FindNext ) {

		while ( !InTheFuture( Date ) ) {
			year++;

			if ( MakeEaster( year, church, offset, Date ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Resulting date out of range\r\n" );
				return -1;
			}
		}

		do {
			testrv = MakeEaster( year - 1, church, offset, TestDate );

			if ( testrv == 0 )
				if ( InTheFuture( TestDate ) ) {
					year--;
					Date = TestDate;
				}
				else
					testrv = -31;

		}
		while ( !testrv );

	}

	OutputFormattedDate( Date, fmt, String );
	return 0;
}



int MakeHoliday( int index, int whatday, int month, int year, long int offset, SYSTEMTIME &Date )
{
	BOOL leap = Leapness( year );
	long int julian;
	int kdow = Julian( year, month, 1 ) % 7;


	if ( ( index < 1 ) || ( index > 5 ) || ( whatday < 0 ) || ( whatday > 7 ) || ( month < 1 ) || ( month > 12 ) || ( year < Epoch ) || ( year > Ultima ) )
		return -1;


	whatday %= 7;

	Date.wYear  = year;
	Date.wMonth = month;

	Date.wDay = ( ( whatday + 7 - kdow ) % 7 ) + 1 + ( ( index - 1 ) * 7 );
	if ( Date.wDay > (unsigned) MonthLength[Date.wMonth][leap] )
		Date.wDay -= 7;

	if ( offset ) {
		julian = Julian( Date ) + offset;

		if ( ( julian < MinimumJulian ) || ( julian > MaximumJulian ) )
			return -1;
		else
			MakeDate( julian, Date );
	}

	return 0;
}



DLLExports int WINAPI f_holiday( LPTSTR String )
{
	if ( !String )
		return 1;


	int index = 0, whatday = 0, whatmonth = 0, whatyear = 0, fmt = 0x100, testrv = 0;
	long int offset = 0;
	BOOL FindNext = FALSE;
	SYSTEMTIME Date, TestDate;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );


	if ( arglen[0] ) {
		index = ParseInt( arg[0] );

		if ( index < 0 )
			if ( !_wcsicmp( arg[0], L"L" ) || !_wcsicmp( arg[0], L"F" ) || !_wcsicmp( arg[0], L"last" ) || !_wcsicmp( arg[0], L"final" ) )
				index = 5;

		if ( ( index < 1 ) || ( index > 5 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad index \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No index\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		whatday = ParseDow( arg[1] );

		if ( whatday < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad day-of-the-week \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No day-of-the-week\r\n" );
		return -1;
	}

	if ( arglen[2] ) {
		whatmonth = ParseMonth(arg[2]);

		if ( ( whatmonth < 1 ) || ( whatmonth > 12 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad month \"%s\"\r\n", arg[2] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No month\r\n" );
		return -1;
	}


	if ( QcalDoingDelayedExpansion )
		whatyear = QcalDoingDelayedExpansion;
	else
		whatyear = Now.wYear;


	if ( arglen[3] ) {
		if ( !_wcsicmp( arg[3], L"*" ) )
			FindNext = TRUE;
		else
			whatyear = ParseYear( arg[3] );

		if ( whatyear < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad year \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	if ( arglen[4] ) {
		fmt = ParseInt( arg[4] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad format \"%s\"\r\n", arg[4] );
			return -1;
		}
	}

	if ( arglen[5] ) {
		if ( SignedNumberStart( arg[5] ) )
			offset = wcstol( arg[5], NULL, 10 );
		else {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad offset \"%s\"\r\n", arg[5] );
			return -1;
		}

		if ( ( offset < -31 ) || ( offset > 31 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad offset \"%s\"\r\n", arg[5] );
			return -1;
		}

	}


	if ( MakeHoliday( index, whatday, whatmonth, whatyear, offset, Date ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Resulting date out of range\r\n" );
		return -1;
	}

	if ( FindNext ) {
		while ( !InTheFuture( Date ) ) {
			whatyear++;

			if ( MakeHoliday( index, whatday, whatmonth, whatyear, offset, Date ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Resulting date out of range\r\n" );
				return -1;
			}
		}

		do {
			testrv = MakeHoliday( index, whatday, whatmonth, whatyear - 1, offset, TestDate );

			if ( testrv == 0 )
				if ( InTheFuture( TestDate ) ) {
					whatyear--;
					Date = TestDate;
				}
				else
					testrv = -31;

		}
		while ( !testrv );
	}

	OutputFormattedDate( Date, fmt, String );
	return 0;
}



DLLExports int WINAPI f_inlieu( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;

	long int jul	= 0;
	int dow			= 0;
	int fmt			= 256;
	int mode		= 0;
	int flags		= 0;
	BOOL changed	= FALSE;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );


	if ( arglen[1] ) {
		fmt = ParseInt( arg[1] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[0] ) {
		if ( ParseDate( arg[0], fmt, Date ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else
		Date = Now;

	if ( arglen[2] ) {
		mode = ParseInt( arg[2] );

		if ( ( mode < 0 ) || ( mode > 5 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid mode \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		flags = ParseInt( arg[3] );

		if ( flags < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid flags \"%s\"\r\n", arg[3] );
			return -1;
		}
	}


	jul = Julian( Date );
	dow = jul % 7;


	if ( mode == 4 ) {				//  special mode for Da del Descubrimiento de Dos Mundos (Chile)
		switch ( dow ) {

		case 2 :
			jul--;
			changed = TRUE;
			break;

		case 3 :
			jul -= 2;
			changed = TRUE;
			break;

		case 4 :
			jul -= 3;
			changed = TRUE;
			break;

		case 5 :
			jul += 3;
			changed = TRUE;
			break;
		}
	}
	else if ( mode == 5 ) {			//  special mode for Da Nacional de las Iglesias Evanglicas y Protestantes (Chile)
		if ( dow == 2 ) {
			jul -= 4;
			changed = TRUE;
		}
		else if ( dow == 3 ) {
			jul += 2;
			changed = TRUE;
		}
	}
	else if ( dow == 0 ) {			// Sunday
		if ( mode == 1 )
			jul -= 2;
		else
			jul++;

		changed = TRUE;
	}
	else if ( dow == 6 ) {			// Saturday
		if ( mode == 2 )
			jul += 2;
		else if ( mode != 3 )
			jul--;

		changed = TRUE;
	}


	if ( ( jul < MinimumJulian ) || ( jul > MaximumJulian ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Resulting date out of range\r\n" );
		return -1;
	}


	if ( ( flags & 1 ) && !changed )
		String[0] = '\0';
	else {
		MakeDate( jul, Date );
		OutputFormattedDate( Date, fmt, String );
	}

	return 0;
}



DLLExports int WINAPI f_nextdow( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Date;
	int fmt = 0x100, index = 1;
	int mask = 0;
	long int jul = 0;
	int nd = 0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );


	if ( arglen[0] ) {
		mask = ParseDowMask( arg[0] );

		if ( mask <= 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid days-of-the-week list \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing days-of-the-week list\r\n" );
		return -1;
	}

	if ( arglen[3] ) {
		fmt = ParseInt( arg[3] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	if ( arglen[1] ) {
		if ( ParseDate( arg[1], fmt, Date ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else
		Date = Now;

	if ( arglen[2] ) {
		if ( SignedNumberStart( arg[2] ) )
			index = wcstol( arg[2], NULL, 10 );
		else {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad index \"%s\"\r\n", arg[2] );
			return -1;
		}

		if ( ( index < -127 ) || ( index > 127 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad index \"%s\"\r\n", arg[2] );
			return -1;
		}
	}
	else
		index = 1;


	jul = Julian( Date );

	nd = 0;
	for ( int i = 0; i < 7; i++ )
		if ( mask & ( 1 << i ) )
			nd++;

	while ( index > nd ) {
		jul += 7;
		index -= nd;
	}

	while ( index < -nd ) {
		jul -= 7;
		index += nd;
	}

	while ( index || !JulianMatchesMask( jul, mask ) ) {

		if ( index >= 0 ) {
			// search forward:
			jul++;
			if ( JulianMatchesMask( jul, mask ) && ( index > 0 ) )
				index--;
		}
		else {
			// search backward:
			jul--;
			if ( JulianMatchesMask( jul, mask ) )
				index++;
		}
	}

	if ( ( jul < MinimumJulian ) || ( jul > MaximumJulian ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Resulting date out of range\r\n" );
		return -1;
	}


	MakeDate( jul, Date );
	OutputFormattedDate( Date, fmt, String );
	return 0;
}



DLLExports int WINAPI _t4( LPTSTR outString )
{
	if ( !outString )
		return 1;


	GetNow( FALSE );
	wsprintf( outString, L"%02u%02u", Now.wHour, Now.wMinute );
	return 0;
}



DLLExports int WINAPI _tomorrow( LPTSTR outString )
{
	if ( !outString )
		return 1;


	long int jul			= 0;
	SYSTEMTIME Tomorrow;


	GetNow( FALSE );
	jul = Julian( Now );

	if ( ( jul < MinimumJulian ) || ( jul >= MaximumJulian ) ) {
		wcscpy_s( outString, MAX_ARG_LEN, L"-1" );
		return 0;
	}


	MakeDate( jul + 1, Tomorrow );
	OutputFormattedDate( Tomorrow, 4, outString );
	return 0;
}



DLLExports int WINAPI _yesterday( LPTSTR outString )
{
	if ( !outString )
		return 1;


	long int jul			= 0;
	SYSTEMTIME Yesterday;

	
	GetNow( FALSE );
	jul = Julian( Now );


	if ( ( jul <= MinimumJulian ) || ( jul > MaximumJulian ) ) {
		wcscpy_s( outString, MAX_ARG_LEN, L"-1" );
		return 0;
	}


	MakeDate( jul - 1, Yesterday );
	OutputFormattedDate( Yesterday, 4, outString );
	return 0;
}



int ShowPluginVersion()
{
	Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );
	return 0;
}



int Indent( int n )
{
	if ( n < 1 )
		return 0;
	else if ( n >= SHORT_BUF_LEN )
		n = SHORT_BUF_LEN - 1;


	wchar_t shortbuf[SHORT_BUF_LEN];


	for ( int i = 0; i < n; i++ )
		shortbuf[i] = ' ';


	shortbuf[n] = '\0';
	QPuts( shortbuf );
	return 0;
}



int PagedCrLf()
{
	if ( OutputToConsole && EnableColor ) {
		SetColors( NormalColor );

		if ( OSVerInfo.dwMajorVersion >= 10 )
			QPuts( L"\u00a0" );
	}

	CrLf();

	if ( !OutputToConsole || !paged )
		return 0;

	LineCount++;
	if ( LineCount < ScreenRows )
		return 0;

	_cputws( L"    \u2014 More \u2014 " );

	while ( !QueryKeyWaiting() )
		tty_yield( 0 );

	EatKeystrokes();
	_cputws( L"\r               \r" );
	LineCount = 0;
	return 0;
}



int ShowDowLine(BOOL showweeks, BOOL localnames)
{
	wchar_t shortbuf[SHORT_BUF_LEN];
	int j = 0;

	if ( showweeks )
		Indent( 3 );

	if ( localnames ) {
		wchar_t ch = '\0';

		for ( int i = 0; i < 7; i++ ) {
			j = ( i + LeftColumnDow ) % 7;
			shortbuf[( 3 * i )] = LocalDowName[j][0];
			ch = LocalDowName[j][1];
			shortbuf[( 3 * i ) + 1] = ( ch == '\0' ) ? ' ' : ch;
			shortbuf[( 3 * i ) + 2] = ' ';
		}
	}
	else {
		for ( int i = 0; i < 7; i++ ) {
			j = ( i + LeftColumnDow ) % 7;
			shortbuf[( 3 * i ) + 0] = DowName[j][0];
			shortbuf[( 3 * i ) + 1] = DowName[j][1];
			shortbuf[( 3 *  i) + 2] = ' ';
		}
	}

	shortbuf[21] = 0x20;
	shortbuf[22] = 0x20;
	shortbuf[23] = '\0';
	QPuts( shortbuf );
	return 0;
}



int ShowMonthHeader( unsigned int year, unsigned int month, BOOL showyear, BOOL showweeks, BOOL localnames )
{
	const wchar_t *name = NULL;

	if ( localnames )
		name = &LocalMonthNameCaps[month][0];
	else
		name = &MonthName[month][0];

	int l = (int) wcslen(name);
	int k = showweeks ? 23 : 20;
	int s = 0;

	int m = showyear ? ( k - 5 ) : k;		// max length for month name
	if ( l > m )
		l = m;

	if ( showyear ) {
		s = (k - l - 5) / 2;
		Indent(s);
		Printf(L"%.*s %u", m, name, year);
		Indent(k - 2 - l - s);
	}
	else {
		s = (k - l) / 2;
		Indent(s);
		Printf(L"%.*s", m, name);
		Indent(k + 3 - l - s);
	}

	return 0;
}



unsigned int GetHLColor()
{
	unsigned int normal = 0, inverse = 0;
	GetAttribute(&normal, &inverse);

	NormalColor = normal;
	unsigned int fore = (normal & 0x0f);
	unsigned int back = (normal / 0x10);


	if ( DefColors[back][0] != 0x11 ) {
		HLColor       = DefColors[back][0] ? DefColors[back][0] : NormalColor;
		HLSundayColor = DefColors[back][1] ? DefColors[back][1] : HLColor;
		SundayColor   = DefColors[back][2] ? DefColors[back][2] : NormalColor;
		HolidayColor  = DefColors[back][3] ? DefColors[back][3] : NormalColor;
		WeeksColor    = DefColors[back][4] ? DefColors[back][4] : NormalColor;
		return 0;
	}


	HLColor = ( 0x10 * fore ) + back;
	if ( NormalColor == 0xf0 )
		HLColor = 0x1f;

	if ( back < 0x08 ) {
		if ( fore != 0x0c )
			SundayColor = ( 0x10 * back ) + 0x0c;
		else
			SundayColor = ( 0x10 * back ) + 0x0a;

		if ( back != 0x01 )
			HolidayColor = 0x10 + fore;
		else
			HolidayColor = 0x20 + fore;
	}
	else {
		if ( fore != 0x04 )
			SundayColor = ( 0x10 * back ) + 0x04;
		else
			SundayColor = ( 0x10 * back );

		if ( back != 0x0e )
			HolidayColor = 0xe0;
		else
			HolidayColor = 0xb1;
	}

	HLSundayColor = ( 0x10 * ( SundayColor & 0x0f ) ) + ( SundayColor / 0x10 );

	if ( ( ( back < 0x02 ) || ( back == 0x0f ) ) && ( fore != 0x03 ) )
		WeeksColor = ( 0x10 * back ) + 0x03;
	else
		WeeksColor = normal;

	return 0;
}



int QcalCheckYearRange( LPTSTR inString, int year )
{
	//  Checks for a year range specifier in the input string; if found,
	//  removes it from the string, and tests year against it.
	//  Returns -1 for any error, 0 if the year matches the range,
	//  or 1 if the year does not match the range.

	if ( inString[0] != '(' )
		return 0;

	size_t i = 0;
	int j = 0;
	int minyear = Epoch, maxyear = Ultima;
	int period = 0, remainder = 0;
	wchar_t ch = '\0';
	long int tempint = 0;
	wchar_t *sep1;

	while ( inString[0] == '(' ) {
		i = 1;

		if ( iswdigit(inString[i]) ) {
			// parse minyear
			tempint = wcstol(inString + i, &sep1, 10);

			if ( ( tempint < Epoch ) || ( tempint > Ultima ) )
				return -1;

			minyear = tempint;
			i = sep1 - inString;
		}

		if ( inString[i] == '-' )
			if ( ( inString[i + 1] == ')' ) || ( inString[i + 1] == '?' ) )
				i++;

		if ( inString[i] == '-' ) {
			if ( iswdigit(inString[i + 1]) ) {
				// parse maxyear
				tempint = wcstol(inString + i + 1, &sep1, 10);

				if ( ( tempint < Epoch ) || ( tempint > Ultima ) )
					return -1;

				maxyear = tempint;
				i = sep1 - inString;
			}
			else
				return -1;
		}

		if ( inString[i] == '?' ) {
			if ( iswdigit(inString[i + 1]) ) {
				// parse period
				tempint = wcstol(inString + i + 1, &sep1, 10);

				if ( ( tempint < 1 ) || ( tempint > 100 ) )
					return -1;

				period = tempint;
				i = sep1 - inString;

				if ( inString[i] == '=' ) {
					if ( iswdigit(inString[i + 1]) ) {
						// parse remainder
						tempint = wcstol(inString + i + 1, &sep1, 10);

						if ( tempint >= period )
							return -1;

						remainder = tempint;
						i = sep1 - inString;
					}
					else
						return -1;
				}
			}
			else
				return -1;
		}

		if ( inString[i] != ')' )
			return -1;
		i++;

		while ( iswspace(inString[i]) )
			i++;

		ch = '*';
		for ( j = 0; ch != '\0'; j++ ) {
			ch = inString[i + j];
			inString[j] = ch;
		}

		if ( ( year < minyear ) || ( year > maxyear ) )
			return 1;

		if ( period > 1 )
			if ( ( year % period ) != remainder )
				return 1;
	}

	return 0;
}



int ExpandHolidays( int year, unsigned short int Holidays[][31] )
{
	wchar_t shortbuf[MAX_ARG_LEN]	= L"";
	wchar_t yearbuf[16]				= L"";
	int k = 0;
	BOOL percent					= FALSE;
	SYSTEMTIME Holiday;
	int NumberAdded					= 0;
	_HolidayDef *This				= FirstHolidayDef;
	long int basejulian				= Julian( year, 1, 1 ) - 1;
	BOOL leap						= Leapness( year );


	wsprintf( yearbuf, L"%u", year );
	QcalDoingDelayedExpansion = year;
	memset( Holidays, 0x00, 12 * 31 * sizeof( unsigned short int ) );


	while ( This ) {
		percent = FALSE;
		k = 0;
		shortbuf[k] = '\0';

		This->Julian = 0;

		for ( wchar_t *ch = This->Text; *ch; ch++ ) {
			if ( *ch == '%' )
				percent = TRUE;

			if ( *ch == '*' ) {
				wcscat_s( shortbuf, MAX_ARG_LEN, yearbuf );
				k = (int) wcslen( shortbuf );
			}
			else {
				shortbuf[k++] = *ch;
				shortbuf[k] = '\0';
			}
		}

		if ( QcalCheckYearRange( shortbuf, year ) == 1 ) {
			This = This->Next;
			continue;
		}

		if ( percent )
			ExpandVariables( shortbuf, 1 );

		memset( &Holiday, 0x00, sizeof( Holiday ) );
		if ( ParseDate( shortbuf, 0, Holiday ) >= 0 ) {
			if ( ( Holiday.wYear == year ) && ( Holiday.wMonth > 0 ) && ( Holiday.wMonth <= 12 ) && ( Holiday.wDay > 0 ) && ( Holiday.wDay <= 31 ) ) {
				Holidays[Holiday.wMonth - 1][Holiday.wDay - 1] = 1;
				This->Julian	= basejulian + Kalends[Holiday.wMonth][leap] + Holiday.wDay;
				This->Enabled	= TRUE;
			}
		}

		This = This->Next;
	}

	QcalDoingDelayedExpansion = 0;
	return NumberAdded;
}



int ShowDay( unsigned int day, BOOL selected, BOOL holiday, unsigned int dow )
{
	if ( !selected && !holiday && ( dow != 0 ) ) {
		Printf( L"%2u ", day );
		return 0;
	}

	if ( !EnableColor ) {
		if ( selected )
			QPuts( L" * " );
		else if ( holiday )
			QPuts( L" + " );
		else
			Printf( L"%2u ", day );
		return 0;
	}

	int color = 0x0b;

	if ( selected && ( dow == 0 ) )
		color = HLSundayColor;
	else if ( selected )
		color = HLColor;
	else if ( holiday )
		color = HolidayColor;
	else if ( dow == 0 )
		color = SundayColor;

	ColorPrintf( color, L"%2u", day );
	QPuts( L" " );
	return 0;
}



int QcalShowWeekNum( unsigned int year, unsigned int month, int day )
{
	if ( day < 1 )
		day = 1;

	int ord = Kalends[month][Leapness( year )] + day - 1;
	int offset = ISOWeekOffset( year );
	unsigned int isoweek = ( ( ord - offset ) / 7 ) + 1;
	
	if ( isoweek > ISOWeeksIn( year ) )
		isoweek = 1;
	else if ( ord < offset )
		isoweek = ISOWeeksIn( year - 1 );

	if ( EnableColor )
		ColorPrintf( WeeksColor, L"%02u ", isoweek );
	else
		Printf( L"%02u ", isoweek );

	return 0;
}



int ShowMonthCal( unsigned int year, unsigned int month, unsigned int selday, BOOL showweeks, BOOL localnames, BOOL showholidays, long int &holminjul, long int &holmaxjul )
{
	long int julian = Julian(year, month, 1);
	BOOL leap = Leapness(year);
	int b = 0, day = 0;
	unsigned short int Holidays[12][31];

	memset( Holidays, 0x00, sizeof( Holidays ) );
	ExpandHolidays( year, Holidays );

	holminjul = Julian( year, month, 1 );
	holmaxjul = Julian( year, month, MonthLength[month][Leapness( year )] );

	PagedCrLf();
	ShowMonthHeader( year, month, TRUE, showweeks, localnames );
	PagedCrLf();
	PagedCrLf();
	ShowDowLine( showweeks, localnames );
	PagedCrLf();

	b = ( julian + 7 - LeftColumnDow ) % 7;

	for ( int y = 0; y < 6; y++ ) {
		if ( showweeks ) {
			day = ( 7 * y ) - b + 1;

			if ( day > MonthLength[month][leap] )
				QPuts( L"   " );
			else
				QcalShowWeekNum( year, month, day );
		}

		for ( int x = 0; x < 7; x++ ) {
			day = ( 7 * y ) + x - b + 1;
			if ( ( day < 1 ) || ( day > MonthLength[month][leap] ) )
				QPuts(L"   ");
			else
				ShowDay( day, ( day == selday ), showholidays && Holidays[month - 1][day - 1], ( x + LeftColumnDow ) % 7 ) ;
		}
		PagedCrLf();
	}

	PagedCrLf();
	return 0;
}



int ShowYearCal( unsigned int year, unsigned int selmonth, unsigned int selday, BOOL showweeks, BOOL localnames, BOOL showholidays, long int &holminjul, long int &holmaxjul )
{
	long int julian = 0;
	BOOL leap = Leapness( year );
	int b[3] = {0, 0, 0};
	int month = 0, day = 0;
	unsigned short int Holidays[12][31];

	memset( Holidays, 0x00, sizeof( Holidays ) );
	ExpandHolidays( year, Holidays );

	holminjul = Julian( year, 1, 1 );
	holmaxjul = Julian( year, 12, 31 );

	PagedCrLf();
	Indent( showweeks ? 30 : 26 );
	Printf( L"\u2014\u2014\u2014  %u  \u2014\u2014\u2014", year );
	PagedCrLf();
	PagedCrLf();

	for ( int j = 0; j < 4; j++ ) {
		for ( int i = 0; i < 3; i++ ) {
			month = ( j * 3 ) + i + 1;
			julian = Julian( year, month, 1 );
			b[i] = ( julian + 7 - LeftColumnDow ) % 7;

			ShowMonthHeader( year, month, FALSE, showweeks, localnames );
		}

		PagedCrLf();
		PagedCrLf();
		for ( int i = 0; i < 3; i++ )
			ShowDowLine( showweeks, localnames );
		PagedCrLf();

		for ( int y = 0; y < 6; y++ ) {
			for ( int i = 0; i < 3; i++ ) {
				month = ( j * 3 ) + i + 1;

				if ( showweeks ) {
					day = ( 7 * y ) - b[i] + 1;
					if ( day > MonthLength[month][leap] )
						QPuts( L"   " );
					else
						QcalShowWeekNum( year, month, day );
				}

				for ( int x = 0; x < 7; x++ ) {
					day = ( 7 * y ) + x - b[i] + 1;
					if ( ( day < 1 ) || ( day > MonthLength[month][leap] ) )
						QPuts( L"   " );
					else
						ShowDay( day, ( ( month == selmonth ) && ( day == selday ) ), showholidays && Holidays[month - 1][day - 1], ( x + LeftColumnDow ) % 7 );
				}
				QPuts( L"  " );
			}
			PagedCrLf();
		}
		PagedCrLf();
	}
	return 0;
}



int ShowThreeMonthCal( unsigned int startyear, unsigned int selmonth, unsigned int selday, BOOL showweeks, BOOL localnames, BOOL showholidays, long int &holminjul, long int &holmaxjul )
{
	long int julian = 0;
	BOOL leap[3] = {FALSE, FALSE, FALSE};
	int b[3]     = {0, 0, 0};
	int year[3]  = {0, 0, 0};
	int month[3] = {0, 0, 0};
	int day = 0;
	unsigned short int Holidays[3][12][31];

	memset( Holidays, 0x00, sizeof( Holidays ) );


	int startmonth = selmonth - 1;
	if ( startmonth < 1 ) {
		startmonth = 12;
		startyear -= 1;
		if ( startyear < Epoch ) {
			startyear = Epoch;
			startmonth = 1;
		}
	}

	if ( ( startyear == Ultima ) && ( startmonth > 10 ) )
		startmonth = 10;

	PagedCrLf();
	for ( int i = 0; i < 3; i++ ) {
		month[i] = startmonth + i;
		year[i]  = startyear;
		if ( month[i] > 12 ) {
			month[i] -= 12;
			year[i]++;
		}
		leap[i] = Leapness( year[i] );
		julian = Julian( year[i], month[i], 1 );
		b[i] = ( julian + 7 - LeftColumnDow ) % 7;

		ShowMonthHeader( year[i], month[i], TRUE, showweeks, localnames );
	}

	holminjul = Julian( year[0], month[0], 1 );
	holmaxjul = Julian( year[2], month[2], MonthLength[month[2]][Leapness( year[2] )] );

	ExpandHolidays( year[0], Holidays[0] );

	if ( year[1] == year[0] )
		memcpy( Holidays[1], Holidays[0], 12 * 31 * sizeof( unsigned short int ) );
	else
		ExpandHolidays( year[1], Holidays[1] );

	if ( year[2] == year[1] )
		memcpy( Holidays[2], Holidays[1], 12 * 31 * sizeof( unsigned short int ) );
	else
		ExpandHolidays( year[2], Holidays[2] );

	PagedCrLf();
	PagedCrLf();
	for ( int i = 0; i < 3; i++ )
		ShowDowLine( showweeks, localnames );
	PagedCrLf();

	for ( int y = 0; y < 6; y++ ) {
		for ( int i = 0; i < 3; i++ ) {
			if ( showweeks ) {
				day = ( 7 * y ) - b[i] + 1;
				if ( day > MonthLength[month[i]][leap[i]] )
					QPuts( L"   " );
				else
					QcalShowWeekNum( year[i], month[i], day );
			}

			for ( int x = 0; x < 7; x++ ) {
				day = ( 7 * y ) + x - b[i] + 1;
				if  ( ( day < 1 ) || ( day > MonthLength[month[i]][leap[i]] ) )
					QPuts( L"   " );
				else
					ShowDay( day, ( ( month[i] == selmonth ) && ( day == selday ) ), showholidays && Holidays[i][month[i] - 1][day - 1], ( x + LeftColumnDow ) % 7 );
			}
			QPuts( L"  " );
		}
		PagedCrLf();
	}

	PagedCrLf();
	return 0;
}



int ShowSixMonthCal( unsigned int year, unsigned int selmonth, unsigned int selday, BOOL showweeks, BOOL localnames, BOOL showholidays, long int &holminjul, long int &holmaxjul )
{
	long int julian = 0;
	BOOL leap = Leapness( year );
	int b[3] = { 0, 0, 0 };
	int startmonth = ( selmonth < 7 ) ? 1 : 7;
	int month = 0, day = 0;
	unsigned short int Holidays[12][31];

	memset( Holidays, 0x00, sizeof( Holidays ) );
	ExpandHolidays( year, Holidays );

	holminjul = Julian( year, startmonth, 1 );
	holmaxjul = Julian( year, startmonth + 5, MonthLength[startmonth + 5][Leapness( year )] );

	PagedCrLf();
	Indent( showweeks ? 30 : 26 );
	Printf( L"\u2014\u2014\u2014  %u  \u2014\u2014\u2014", year );
	PagedCrLf();
	PagedCrLf();

	for ( int j = 0; j < 2; j++ ) {
		for ( int i = 0; i < 3; i++ ) {
			month = ( j * 3 ) + startmonth + i;
			julian = Julian( year, month, 1 );
			b[i] = ( julian + 7 - LeftColumnDow ) % 7;

			ShowMonthHeader( year, month, FALSE, showweeks, localnames );
		}

		PagedCrLf();
		PagedCrLf();

		for ( int i = 0; i < 3; i++ )
			ShowDowLine( showweeks, localnames );

		PagedCrLf();

		for ( int y = 0; y < 6; y++ ) {
			for ( int i = 0; i < 3; i++ ) {
				month = ( j * 3 ) + startmonth + i;

				if ( showweeks ) {
					day = ( 7 * y ) - b[i] + 1;
					if ( day > MonthLength[month][leap] )
						QPuts( L"   " );
					else
						QcalShowWeekNum( year, month, day );
				}

				for ( int x = 0; x < 7; x++ ) {
					day = ( 7 * y ) + x - b[i] + 1;
					if ( ( day < 1 ) || ( day > MonthLength[month][leap] ) )
						QPuts( L"   " );
					else
						ShowDay( day, ( ( month == selmonth ) && ( day == selday ) ), showholidays && Holidays[month - 1][day - 1], ( x + LeftColumnDow ) % 7 );
				}

				QPuts( L"  " );
			}

			PagedCrLf();
		}

		PagedCrLf();
	}

	return 0;
}



int AddHolidayDef( LPTSTR Text, LPTSTR Name )
{
	if ( !Text )
		return -1;
	else if ( Text[0] == '\0' )
		return -1;


	_HolidayDef *New	= NULL;
	_HolidayDef *This	= FirstHolidayDef;

	size_t TextLen		= wcslen( Text );
	size_t NameLen		= 0;
	int Count			= 0;


	if ( Name )
		if ( Name[0] )
			NameLen = wcslen( Name ) + 1;


	New = (_HolidayDef *) malloc( sizeof( _HolidayDef ) + ( ( TextLen + NameLen ) * sizeof( wchar_t ) ) );
	if ( !New )
		return 1;

	wcscpy_s( New->Text, TextLen + 1, Text );

	New->Julian		= 0;
	New->Enabled	= FALSE;

	if ( NameLen ) {
		New->Name = New->Text + TextLen + 1;
		wcscpy_s( New->Name, NameLen, Name );
	}
	else
		New->Name	= New->Text + TextLen;

	New->Next = NULL;
	Count++;

	if ( FirstHolidayDef ) {
		This = FirstHolidayDef;

		while ( This->Next )
			This = This->Next;

		This->Next = New;
	}
	else
		FirstHolidayDef = New;

	return 0;
}



int AddHolidayDef2( LPTSTR Text )
{
	if ( !Text )
		return -1;


	int brackets		= 0;
	LPTSTR Name			= NULL;


	for ( wchar_t *ch = Text; *ch; ch++ ) {
		if ( *ch == '[' )
			brackets++;
		else if ( ( *ch == ']' ) && brackets )
			brackets--;
		else if ( ( *ch == '~' ) && !brackets ) {
			if ( ch[1] ) {
				Name = ch + 1;
				*ch = '\0';
				break;
			}
		}
	}


	if ( Name ) {
		TrimLine( Text );
		TrimLine( Name );
	}

	return AddHolidayDef( Text, Name );
}



int QcalAddHolidays( LPTSTR inString )
{
	if ( !inString )
		return 0;

	wchar_t shortbuf[MAX_ARG_LEN];
	BOOL done = FALSE;
	int i = 2, j = 0;
	wchar_t ch = '\0';
	int brackets = 0;
	
	if ( IsColonOrEquals( inString[i] ) )
		i++;

	do {
		ch = inString[i];

		if ( ch == '[' )
			brackets++;

		if ( ( ch == ']' ) && ( brackets > 0 ) )
			brackets--;

		if ( ( ch == '\0' ) || ( !brackets && IsCommaOrSemi( ch ) ) ) {
			shortbuf[j] = '\0';
			if ( ch == '\0' )
				done = TRUE;

			if ( shortbuf[0] )
				AddHolidayDef2( shortbuf );

			j = 0;
		}
		else {
			shortbuf[j] = ch;
			j++;
		}

		i++;
	}
	while ( !done );

	return 0;
}



int QcalDefaultHolidays()
{
	// New Year's Day:
	AddHolidayDef( L"*-01-01", L"New Year\u2019s Day" );

	// Martin Luther King, Jr. Day:
	AddHolidayDef( L"(1986)%@holiday[3,1,1]", L"Martin Luther King Jr. Day" );

	// Washington's Birthday:
	AddHolidayDef( L"(1880-1970)*-02-22", L"Washington\u2019s Birthday" );
	AddHolidayDef( L"(1971)%@holiday[3,1,2]", L"Washington\u2019s Birthday" );

	// Easter:
	AddHolidayDef( L"%@easter[]", L"Easter" );

	// Mother's Day:
	AddHolidayDef( L"(1914)%@holiday[2,7,5]", L"Mother\u2019s Day" );

	// Memorial Day:
	AddHolidayDef( L"(1868-1970)*-05-30", L"Memorial Day" );
	AddHolidayDef( L"(1971)%@holiday[5,1,5]", L"Memorial Day" );

	// Father's Day:
	AddHolidayDef( L"(1910)%@holiday[3,7,6]", L"Father\u2019s Day" );

	// Juneteenth:
	AddHolidayDef( L"(1865)*-06-19", L"Juneteenth" );
	AddHolidayDef( L"(2021)%@inlieu[*-06-19,,,1]", L"Juneteenth (observed)" );

	// Independence Day:
	AddHolidayDef( L"(1776)*-07-04", L"Independence Day" );
	AddHolidayDef( L"(1776)%@inlieu[*-07-04,,,1]", L"Independence Day (observed)" );

	// Labor Day:
	AddHolidayDef( L"(1894)%@holiday[1,1,9]", L"Labor Day" );

	// Columbus Day:
	AddHolidayDef( L"(1934-1970)*-10-12", L"Columbus Day" );
	AddHolidayDef( L"(1971)%@holiday[2,1,10]", L"Columbus Day" );

	// Halloween:
	AddHolidayDef( L"*-10-31", L"Halloween" );

	// Election Day:
	AddHolidayDef( L"(1848?2)%@holiday[1,1,11,,,1]", L"Election Day" );

	// Veterans Day:
	AddHolidayDef( L"(1919-1970)*-11-11", L"Veterans Day" );
	AddHolidayDef( L"(1971-1977)%@holiday[4,1,10]", L"Veterans Day" );
	AddHolidayDef( L"(1978)*-11-11", L"Veterans Day" );

	// Thanksgiving:
	AddHolidayDef( L"(1942)%@holiday[4,4,11]", L"Thanksgiving Day" );
	AddHolidayDef( L"(1939-1941)%@holiday[5,4,11,,,-7]", L"Thanksgiving Day" );
	AddHolidayDef( L"(1863-1938)%@holiday[5,4,11]", L"Thanksgiving Day" );

	// Christmas:
	AddHolidayDef( L"*-12-25", L"Christmas" );

	// New Year's Eve:
	AddHolidayDef( L"*-12-31", L"New Year\u2019s Eve" );

	return 0;
}



int PrepSetDos()
{
	wcscpy_s( svSetDos, 32, L"%_expansion" );

	ExpandVariables( svSetDos, 1 );
	if ( svSetDos[0] == '\0' )
		wcscpy_s( svSetDos, 32, L"0" );

	CallInternal( L"setdos", L" /x-1478+3" );
	return 0;
}



int ResetSetDos()
{
	wcscpy_s( fxSetDos, 32, L"/X" );
	wcscat_s( fxSetDos, 32, svSetDos );

	CallInternal( L"setdos", fxSetDos );
	return 0;
}



int SplitHolidayLine( LPTSTR Line, wchar_t **HolidayName )
{
	if ( !Line || !HolidayName )
		return -1;


	wchar_t *ch			= Line;

	int brackets		= 0;
	wchar_t *sep1		= NULL;


	for ( ch = Line; *ch; ch++ ) {

		if ( *ch == '[' )
			brackets++;
		else if ( brackets && ( *ch == ']' ) )
			brackets--;
		else if ( !brackets && ( *ch == ';' ) ) {
			sep1 = ch;
			break;
		}
	}


	if ( !sep1 )
		return 0;

	*sep1 = '\0';
	*HolidayName = sep1 + 1;

	TrimLine( Line );
	TrimLine( sep1 + 1 );
	return 1;
}



int SortHolidaysByName$( const void *a, const void *b )
{
	_HolidayDef **a1	= (_HolidayDef**) a;
	_HolidayDef **b1	= (_HolidayDef**) b;

	return _wcsicmp( (*a1)->Name, (*b1)->Name );
}



int SortHolidaysByDate$( const void *a, const void *b )
{
	_HolidayDef **a1	= (_HolidayDef**) a;
	_HolidayDef **b1	= (_HolidayDef**) b;

	if ( (*a1)->Julian < (*b1)->Julian )
		return -1;
	else if ( (*a1)->Julian > (*b1)->Julian )
		return 1;
	else
		return _wcsicmp( (*a1)->Name, (*b1)->Name );
}



int ListHolidaysInRange( long int MinJulian, long int MaxJulian, int HolidaySort )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	_HolidayDef *This		= FirstHolidayDef;
	_HolidayDef **Sorted	= NULL;

	int Count				= 0;
	int i					= 0;
	SYSTEMTIME Date;


	if ( !This )
		return -1;

	while ( This ) {
		if ( This->Enabled && ( This->Julian >= MinJulian ) && ( This->Julian <= MaxJulian ) )
			Count++;

		This = This->Next;
	}

	if ( !Count )
		return 0;

	Sorted = (_HolidayDef**) malloc( Count * sizeof( _HolidayDef* ) );
	if ( !Sorted )
		return -1;

	This = FirstHolidayDef;
	i = 0;

	while ( This ) {
		if ( This->Enabled && ( This->Julian >= MinJulian ) && ( This->Julian <= MaxJulian ) )
			Sorted[i++] = This;

		This=This->Next;
	}

	if ( HolidaySort && ( Count > 1 ) ) {

		if ( HolidaySort == 1 )
			qsort( Sorted, Count, sizeof( _HolidayDef* ), SortHolidaysByName$ );
		else
			qsort( Sorted, Count, sizeof( _HolidayDef* ), SortHolidaysByDate$ );
	}


	PagedCrLf();

	for ( int i = 0; i < Count; i++ ) {
		MakeDate( Sorted[i]->Julian, Date );
		//	OutputFormattedDate( Date, 4, temp );

		if ( LocalDateFormat == 2 )
			wsprintf( temp, L"%2u %s", Date.wDay, LocalMonthName[Date.wMonth] );
		else
			wsprintf( temp, L"%s %2u", LocalMonthName[Date.wMonth], Date.wDay );

		Printf( L"  %s:  %s", temp, Sorted[i]->Name );
		PagedCrLf();
	}

	return Count;
}



int DeHex( LPTSTR inTwoChars )
{
	int val = -1;
	wchar_t ch = toupper( inTwoChars[0] );

	if ( ( ch >= '0' ) && ( ch <= '9' ) )
		val = ch - '0';
	else if ( ( ch >= 'A' ) && ( ch <= 'F' ) )
		val = ch - 'A' + 10;

	ch = toupper( inTwoChars[1] );

	if ( ch == '\0' )
		return val;
	else
		val *= 16;

	if ( ( ch >= '0' ) && ( ch <= '9' ) )
		val += ch - '0';
	else if ( ( ch >= 'A' ) && ( ch <= 'F' ) )
		val += ch - 'A' + 10;
	else
		return -1;

	return val;
}



BOOL IsInvalidColor( int inColor )
{
	if ( inColor == 0 )
		return FALSE;

	if ( ( inColor < 0x00 ) || ( inColor > 0xff ) )
		return TRUE;

	return ( ( inColor / 16 ) == ( inColor & 0x0f ) );
}


#include "mmfiles.cpp"


int QcalReadHolidaysFile()
{
	wchar_t FullName[MAX_FILENAME_LEN] = L"";
	wchar_t *lastbs;
	wchar_t SectionName[128] = L"";
	BOOL CorrectSection		= FALSE;

	_MMFile MMFile;

	int i = 0, rv = 0;
	wchar_t FileLine[MAX_ARG_LEN] = L"";
	wchar_t *HolidayName	= NULL;
	int Count		= 0;


	GetModuleFileName( MyDLLHandle, FullName, MAX_FILENAME_LEN );
	lastbs = wcsrchr( FullName, L'\\' );
	if ( !lastbs )
		return -1;

	lastbs++;
	
	wcscpy_s( lastbs, MAX_FILENAME_LEN - ( lastbs - FullName ), L"holidays.ini" );

	SectionName[0] = '[';
	wcscat_s(SectionName, 128, CountryName);
	wcscat_s(SectionName, 128, L"]");


	if ( OpenMMFile( &MMFile, FullName, 0 ) )
		return -1;

	//	wprintf( L"   * Reading holidays for '%s' from file '%s' :\n", CountryName, FullName );

	while ( MMFile.Index < MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, FileLine, MAX_ARG_LEN, FALSE ) )
			break;

		TrimLine( FileLine );

		if ( FileLine[0] == '[' )			// Start of section :
			CorrectSection = !_wcsicmp( FileLine, SectionName );

		else if ( CorrectSection && ( FileLine[0] != '\0' ) ) {		// Definition line :
			HolidayName		= NULL;

			SplitHolidayLine( FileLine, &HolidayName );

			if ( !AddHolidayDef( FileLine, HolidayName ) )
				Count++;
		}

		// Printf(L"   %03u : \"%s\" \r\n", i, FileLine);
		// i++;
	}

	CloseMMFile( &MMFile );

	return Count;
}



int QcalDumpHolidays()
{
	_HolidayDef *This		= FirstHolidayDef;
	int i					= 0;



	while ( This ) {

		i++;
		Printf( L"   * %d  ->  '%s'  '%s'\n", i, This->Name,  This->Text );

		This = This->Next;
	}

	return i;
}



DLLExports int WINAPI qcal( LPTSTR String )
{
	if ( !String )
		return 1;

	unsigned int year = 0, month = 0, day = 0;
	SYSTEMTIME SelectDate;
	BOOL datefound		= FALSE;
	int mode			= 0;
	BOOL showweeks		= FALSE;
	BOOL readholidays	= FALSE;
	BOOL showholidays	= FALSE;
	BOOL listholidays	= FALSE;
	BOOL localnames		= TRUE;
	BOOL ShowToday		= TRUE;
	int HolidaySort		= 2;
	int leftcol = -1;

	long int holminjul		= 0;
	long int holmaxjul		= 0;

	BOOL numeric = FALSE;
	long int numin = 0;

	EnableColor = OutputToConsole = QueryIsConsole( GetStdHandle( STD_OUTPUT_HANDLE ) );


	GetNow( FALSE );
	GetLocalDateFormat();
	FreeHolidayDefs();
	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );

	paged = FALSE;
	ScreenRows = GetScrRows();
	LineCount = 0;
	GetHLColor();


	for ( int i = 0; i < MAX_ARGS; i++ ) {
		if ( arglen[i] > 0 ) {
			if ( arg[i][0] == '/' ) {
				switch ( towupper( arg[i][1] ) ) {

				case 'P' :
					paged = TRUE;
					break;

				case 'W' :
					if ( IsColonOrEquals( arg[i][2] ) )
						leftcol = ParseDow(arg[i] + 3);
					else
						leftcol = ParseDow(arg[i] + 2);

					if ( ( leftcol < 0 ) || ( leftcol > 7 ) ) {
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad day of the week: \"%s\"\r\n", arg[i] );
						return 1;
					}

					break;

				case 'D' :
					EnableColor = FALSE;
					break;

				case '3' :
					mode = 3;
					break;

				case '6' :
					mode = 6;
					break;

				case '1' :
				case 'M' :
					mode = 1;
					break;

				case 'Y' :
					mode = 12;
					break;

				case 'H' :
					if ( arg[i][2] == '\0' )
						showholidays = readholidays = TRUE;
					else if ( ( ( towupper( arg[i][2] ) == 'H' ) || ( towupper( arg[i][2] ) == 'L' ) ) && ( arg[i][3] == '\0' ) )
						showholidays = readholidays = listholidays = TRUE;
					else {
						if ( QcalAddHolidays( arg[i] ) )
							return 1;

						showholidays = TRUE;
					}

					break;

				case 'L' :
					listholidays = TRUE;
					break;

				case 'I' :
					showweeks = TRUE;
					break;

				case 'E' :
					localnames = FALSE;
					break;

				case 'C' :
					{
						int j = 2;
						if ( IsColonOrEquals(arg[i][j]) )
							j++;
						if ( arg[i][j] )
							wcscpy_s(CountryName, 128, arg[i] + j);
						else
							CountryName[0] = '\0';
					}

					readholidays = TRUE;
					break;

				case 'Z' :
					{
						int j = 2;
						BOOL Save = FALSE;
						_wcsupr_s( arg[i], MAX_ARG_LEN );

						if ( arg[i][j] == 'Z' ) {
							j++;
							Save = TRUE;
						}

						if ( IsColonOrEquals( arg[i][j] ) )
							j++;

						int k = 0;
						while ( iswxdigit( arg[i][j + k] ) )
							k++;

						if ( ( k < 2 ) || ( k & 1 ) ) {
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Invalid colors \"%s\"\r\n", arg[i] );
							return 1;
						}

						int hl = 0, hlsun = 0, sun = NormalColor, hol = NormalColor, weeks = NormalColor;
						int bg = NormalColor / 16;

						hl = hlsun = DeHex( arg[i] + j );

						if ( k > 2 )
							hlsun = DeHex( arg[i] + j + 2 );

						if ( k > 4 )
							sun = DeHex( arg[i] + j + 4 );

						if ( k > 6 )
							hol = DeHex( arg[i] + j + 6 );

						if ( k > 8 )
							weeks = DeHex( arg[i] + j + 8 );

						if ( IsInvalidColor(hl) || IsInvalidColor(hlsun) || IsInvalidColor(sun) || IsInvalidColor(hol) || IsInvalidColor(weeks) ) {
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Invalid colors \"%s\"\r\n", arg[i] );
							return 1;
						}

						HLColor       = hl ? hl : NormalColor;
						SundayColor   = sun ? sun : NormalColor;
						HLSundayColor = hlsun ? hlsun : HLColor;
						WeeksColor    = weeks ? weeks : NormalColor;
						HolidayColor  = hol ? hol : NormalColor;

						if ( Save ) {
							DefColors[bg][0] = hl;
							DefColors[bg][1] = hlsun;
							DefColors[bg][2] = sun;
							DefColors[bg][3] = hol;
							DefColors[bg][4] = weeks;

							HKEY KeyHandle = NULL;
							wchar_t RegValueName[SHORT_BUF_LEN] = L"";
							BYTE bytesbuf[10] = { hl, hlsun, sun, hol, weeks, 0, 0, 0, 0, 0 };

							wsprintf( RegValueName, L"QCal default colors %X", bg );

							if ( RegCreateKeyEx( HKEY_CURRENT_USER, PluginRegKey, 0, NULL, 0, KEY_WRITE, NULL, &KeyHandle, NULL ) == ERROR_SUCCESS ) {
								RegSetValueEx( KeyHandle, RegValueName, 0, REG_BINARY, (LPBYTE) &bytesbuf, sizeof(bytesbuf) );
								RegCloseKey( KeyHandle );
							}

						}

					}
					break;

				case 'N' :
					ShowToday = FALSE;
					break;

				case 'O' :
					{
						int j = 2;
						if ( IsColonOrEquals( arg[i][j] ) )
							j++;

						switch ( towupper( arg[i][j] ) ) {

						case 'A' :
						case 'N' :
							HolidaySort = 1;
							break;

						case 'D' :
						case '\0' :
							HolidaySort = 2;
							break;

						case 'U' :
							HolidaySort = 0;
							break;

						default :
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Unknown sort key:  /O:%c\r\n", towupper( arg[i][j] ) );
							return 1;
						}
					}
					break;

				case '?' :
					QPuts( (LPTSTR) QcalHelpText );
					if ( arg[i][2] == '?' )
						ShowPluginVersion();

					return 0;

				default:
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown QCAL option: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else {

				numeric = TRUE;
				numin = -1;
				_wcsupr_s( arg[i], MAX_ARG_LEN );

				for ( int j = 0; arg[i][j] != '\0'; j++ )
					if ( !iswdigit( arg[i][j] ) )
						numeric = FALSE;

				if ( numeric )
					numin = ParseInt( arg[i] );

				if ( ( numin >= ( 10000 * Epoch + 101 ) ) && ( numin <= ( 10000 * Ultima + 1231 ) ) ) {
					//  date entered as YYYYMMDD -- pretend it's not numeric:
					numin = -1;
					numeric = FALSE;
				}

				if ( ( numin >= 1 ) && ( numin <= 12 ) && ( month == 0 ) )
					month = numin;
				else if ( ( numin >= Epoch ) && ( numin <= Ultima ) && ( year == 0 ) )
					year = numin;
				else if ( numeric ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"QCAL syntax error \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( !numeric ) {
					if ( !datefound ) {
						if ( ParseDate(arg[i], 0, SelectDate ) >= 0 )
							datefound = TRUE;
					}

					if ( ( month == 0 ) && !datefound )
						for ( int q = 1; q <= 12; q++ )
							if ( !_wcsnicmp( arg[i], MonthName[q], 3 ) )
								month = q;

					if ( ( month == 0 ) && !datefound )
						for ( int q = 1; q <= 12; q++ )
							if ( !_wcsnicmp( arg[i], LocalMonthName[q], 3 ) )
								month = q;

					if ( ( month == 0 ) && !datefound ) {
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", arg[i] );
						return 1;
					}
				}
			}
		}
	}


	if ( datefound ) {
		year  = SelectDate.wYear;
		month = SelectDate.wMonth;
		day   = SelectDate.wDay;
	}

	if ( year && !month )
		mode = 12;

	if ( ( month >= 1 ) && ( month <= 12 ) && ( year == 0 ) ) {
		year = Now.wYear;
		if ( ( month < Now.wMonth ) && ( year < Ultima ) )
			year++;

		day = 0;
	}

	if ( !year ) {
		year  = Now.wYear;
		month = Now.wMonth;
		day   = ShowToday ? Now.wDay : 0;
	}
	

	if ( readholidays | listholidays )
		if ( QcalReadHolidaysFile() <= 0 )
			QcalDefaultHolidays();

	//	QcalDumpHolidays();


	if ( leftcol >= 0 )
		LeftColumnDow = leftcol;
	else if ( showweeks && ( leftcol < 0 ) )
		LeftColumnDow = 1;



	switch ( mode ) {

	case 12 :
		ShowYearCal( year, month, day, showweeks, localnames, showholidays, holminjul, holmaxjul );
		break;

	case 6 :
		ShowSixMonthCal( year, month, day, showweeks, localnames, showholidays, holminjul, holmaxjul );
		break;

	case 3 :
		ShowThreeMonthCal( year, month, day, showweeks, localnames, showholidays, holminjul, holmaxjul );
		break;

	default :
		ShowMonthCal( year, month, day, showweeks, localnames, showholidays, holminjul, holmaxjul );
	}

	if ( listholidays )
		ListHolidaysInRange( holminjul, holmaxjul, HolidaySort );

	FreeHolidayDefs();
	return 0;
}


#ifndef QCAL_only


DLLExports int WINAPI longyears( LPTSTR String )
{
	if ( !String )
		return 1;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] == 0 )
		Printf(L"LONGYEARS is %s\r\n", StrictYearInput ? L"STRICT" : ( FourDigitsAlways ? L"ON" : L"OFF" ) );
	else if ( !wcscmp( arg[0], L"ON" ) ) {
		FourDigitsAlways = TRUE;
		StrictYearInput  = FALSE;
	}
	else if ( !wcscmp( arg[0], L"OFF" ) ) {
		FourDigitsAlways = FALSE;
		StrictYearInput  = FALSE;
	}
	else if ( !wcscmp( arg[0], L"STRICT" ) ) {
		FourDigitsAlways = TRUE;
		StrictYearInput  = TRUE;
	}
	else if ( !wcscmp( arg[0], L"/?" ) )
		QPuts( (LPTSTR) LongYearsHelpText );
	else if ( !wcscmp( arg[0], L"/??" ) ) {
		QPuts( (LPTSTR) LongYearsHelpText );
		ShowPluginVersion();
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Unknown LONGYEARS option \"%s\"\r\n", arg[0] );
		return 1;
	}

	return 0;
}



#include "NewHelp.cpp"



DLLExports int WINAPI iso8601help( LPTSTR String )
{
	return PluginHelpCmd( String, ISO8601HelpHelpText );
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	DisableFeatures( PluginFeatures );

	if ( !wcslen( PluginFeatures ) )
		return NULL;


	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/iso8601.html";
	piInfo.pszDescription	= (LPTSTR) L"Adds support for ISO 8601 date formats";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports int WINAPI Usage(LPTSTR Name, LPTSTR Usage)
{
	if ( !Name || !Usage )
		return 0;

	if ( !_wcsicmp( Name, L"iso8601help" ) ) {
		wcscpy_s( Usage, 1024, UsageText( ISO8601HelpHelpText ) );
		return 1;
	}
	else if ( !_wcsicmp( Name, L"longyears" ) ) {
		wcscpy_s( Usage, 1024, UsageText( LongYearsHelpText ) );
		return 1;
	}
	else if ( !_wcsicmp( Name, L"qcal" ) ) {
		wcscpy_s( Usage, 1024, UsageText( QcalHelpText ) );
		return 1;
	}

	return 0;
}

#else	//	QCAL_only


DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/iso8601.html";
	piInfo.pszDescription	= (LPTSTR) L"Displays a quick-and-dirty calendar";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}


//	Normally from NewHelp.cpp, but QCal-only build doesn't include that
//	so copy it in here:

LPTSTR UsageText( const wchar_t *String )
{
	if ( !String )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) String, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Help( LPTSTR Name )
{
	if ( !Name )
		return 0;


	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";
	const int bufsize = 2048;
	wchar_t helpbuf[bufsize];


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );


	if ( !_wcsicmp( Name, L"qcal" ) ) {
		wcscpy_s( helpbuf, bufsize, QcalHelpText );
		wcscat_s( helpbuf, bufsize, QcalAdditionalHelpText );
		MessageBox( GetForegroundWindow(), helpbuf, WindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( Name, L"@easter" ) ) {
		MessageBox( GetForegroundWindow(), F_EasterHelpText, WindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( Name, L"@holiday" ) ) {
		MessageBox( GetForegroundWindow(), F_HolidayHelpText, WindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( Name, L"@inlieu" ) ) {
		MessageBox( GetForegroundWindow(), F_InlieuHelpText, WindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( Name, L"@nextdow" ) ) {
		MessageBox( GetForegroundWindow(), F_NextdowHelpText, WindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( Name, L"_t4" ) ) {
		MessageBox( GetForegroundWindow(), V_T4HelpText, WindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( Name, L"_tomorrow" ) ) {
		MessageBox( GetForegroundWindow(), V_TomorrowHelpText, WindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( Name, L"_yesterday" ) ) {
		MessageBox( GetForegroundWindow(), V_YesterdayHelpText, WindowTitle, MB_OK );
		return 1;
	}

	return 0;
}



DLLExports int WINAPI Usage( LPTSTR Name, LPTSTR Usage )
{
	if ( !Name || !Usage )
		return 0;

	if ( !_wcsicmp( Name, L"qcal" ) ) {
		wcscpy_s( Usage, 1024, UsageText( QcalHelpText ) );
		return 1;
	}

	return 0;
}

#endif	//	QCAL_only



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	MyDLLHandle = (HINSTANCE) hModule;

	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports BOOL WINAPI InitializePlugin()
{
	GetVersion2( &OSVerInfo );
	QcalDefaultColors();
	

	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s plugin v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );


	wsprintf( shortbuf, L"%%@if[plugin %s,1,0]", UNLOAD_PLUGIN );
	ExpandVariables( shortbuf, 1 );
	if ( !wcscmp( shortbuf, L"1" ) )
		UnloadOnePlugin( UNLOAD_PLUGIN );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{

#ifndef QCAL_only

	CloseHelpWindow();

#endif	//	QCAL_only

	return 0;
}

