#ifndef ISO8601TIME
#define ISO8601TIME		1.0.0



//		Time-handling functions for ISO8601 plugin



#define PT_12HOURFMT	0x0010
#define PT_GOTHOURS		0x0008
#define PT_GOTMINUTES	0x0004
#define PT_GOTSECONDS	0x0002
#define PT_GOTMSEC		0x0001

#define OT_FORMAT		0x000f
#define OT_SHOWSECONDS	0x0080
#define OT_SHOWMSEC		0x0100
#define OT_LOCALSEPS	0x0200
#define OT_LEADINGZERO	0x0400
#define OT_PADLEFT		0x0800
#define OT_NOTIMESEPS	0x1000

#define TIME_FUDGE		1.0e-08




int ParseTime( LPTSTR String, SYSTEMTIME &Time, int *ParseTimeFlags )
{
	if ( !String )
		return -1;


	if ( ParseTimeFlags )
		*ParseTimeFlags = 0;


	long int Hour = 0, Minute = 0, Second = 0, Msec = 0;
	int outFlags = 0;
	int AmPm = 0;
	wchar_t *ch = String;


	while ( iswspace( *ch ) )
		ch++;


	if ( towupper( *ch ) == 'A' )
		AmPm = 1;
	else if ( towupper( *ch ) == 'P' )
		AmPm = 2;


	if ( AmPm ) {
		outFlags |= PT_12HOURFMT;
		ch++;

		while ( ( towupper( *ch ) == 'M' ) || ( *ch == '.' ) || ( iswspace( *ch ) ) )
			ch++;
	}


	if ( !iswdigit( *ch ) )
		return -1;


	//  get hours:

	_set_errno( 0 );
	Hour = wcstol( ch, &ch, 10 );
	if ( errno )
		return -1;

	if ( ( Hour < 0 ) || ( Hour > 23 ) )
		return -1;


	outFlags |= PT_GOTHOURS;


	if ( IsTimeSep ( *ch ) ) {
		//  get minutes:

		ch++;
		if ( !iswdigit( *ch ) )
			return -1;

		_set_errno( 0 );
		Minute = wcstol( ch, &ch, 10 );
		if ( errno )
			return -1;

		if ( ( Minute < 0 ) || ( Minute > 59 ) )
			return -1;

		outFlags |= PT_GOTMINUTES;


		if ( IsTimeSep( *ch ) ) {
			//  get seconds:

			ch++;
			if ( !iswdigit( *ch ) )
				return -1;

			_set_errno( 0 );
			Second = wcstol( ch, &ch, 10 );
			if ( errno )
				return -1;

			if ( ( Second < 0 ) || ( Second > 59 ) )
				return -1;

			outFlags |= PT_GOTSECONDS;


			if ( IsDecimalSep( *ch ) ) {
				//  get milliseconds:

				ch++;
				int roller = 100;
				Msec = 0;

				while ( iswdigit( *ch ) ) {
					Msec += roller * ( *ch - '0' );
					roller /= 10;
					ch++;
					outFlags |= PT_GOTMSEC;
				}
			}
		}
	}


	if ( !AmPm && *ch ) {
		
		while ( iswspace( *ch ) )
			ch++;

		if ( towupper( *ch ) == 'A' )
			AmPm = 1;
		else if ( towupper( *ch ) == 'P' )
			AmPm = 2;
		else if ( *ch != '\0' )
			return -1;

		if ( AmPm ) {
			outFlags |= PT_12HOURFMT;
			ch++;
			while ( ( towupper( *ch ) == 'M' ) || ( *ch == '.' ) || ( iswspace( *ch ) ) )
				ch++;
		}
	}

	if ( *ch != '\0' )
		return -1;


	if ( AmPm ) {
		if ( ( Hour < 1 ) || ( Hour > 12 ) )
			return -1;

		Hour %= 12;
		if ( AmPm == 2 )
			Hour += 12;
	}
	else if ( ( outFlags & PT_GOTMINUTES ) == 0 )
		return -1;


	Time.wHour			= (WORD) Hour;
	Time.wMinute		= (WORD) Minute;
	Time.wSecond		= (WORD) Second;
	Time.wMilliseconds	= (WORD) Msec;

	if ( ParseTimeFlags )
		*ParseTimeFlags = outFlags;

	return 0;
}



int OutputTime( SYSTEMTIME Time, int fmt, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( ( Time.wHour > 23 ) || ( Time.wMinute > 59 ) || ( Time.wSecond > 59 ) || ( Time.wMilliseconds > 999 ) )
		return -1;


	if ( ( fmt & 0x0f ) == 0x0f ) {
		double DecimalTime = ( Time.wHour / 24.0 ) + ( Time.wMinute / 1440.0 )  + ( Time.wSecond / 86400.0 ) + ( Time.wMilliseconds / 86400000.0 );
		wchar_t temp[SHORT_BUF_LEN] = L"";
		swprintf_s( temp, SHORT_BUF_LEN, L"%.8f", DecimalTime );
		wcscpy_s( outString, SHORT_BUF_LEN, temp + ( ( fmt & OT_LEADINGZERO ) ? 0 : 1 ) );
		return 0;
	}


	int hours = Time.wHour;
	int AmPm = 0;
	wchar_t AmPmString[SHORT_BUF_LEN] = L"";
	wchar_t PadHours[2] = { '\0', '\0' };
	wchar_t TimeSep[2]  = L":";
	wchar_t DecSep[2]   = L".";


	if ( fmt & OT_FORMAT ) {
		hours = hours %= 12;
		if ( !hours )
			hours = 12;

		AmPm = 1;
		if ( Time.wHour > 11 )
			AmPm = 2;
	}

	if ( hours < 10 ) {
		if ( fmt & ( OT_LEADINGZERO | OT_NOTIMESEPS ) )
			PadHours[0] = '0';
		else if ( fmt & OT_PADLEFT )
			PadHours[0] = 0x20;
	}

	if ( fmt & OT_LOCALSEPS ) {
		TimeSep[0] = LocalTimeSep;
		DecSep[0]  = LocalDecSep;
	}

	if ( fmt & OT_NOTIMESEPS ) {
		TimeSep[0] = '\0';
		if ( ( fmt & ( OT_LEADINGZERO | OT_LOCALSEPS ) ) == 0 )
			DecSep[0] = '\0';
	}
	

	if ( fmt & OT_SHOWMSEC )
		wsprintf( outString, L"%s%u%s%02u%s%02u%s%03u", PadHours, hours, TimeSep, Time.wMinute, TimeSep, Time.wSecond, DecSep, Time.wMilliseconds );
	else if ( fmt & OT_SHOWSECONDS )
		wsprintf( outString, L"%s%u%s%02u%s%02u", PadHours, hours, TimeSep, Time.wMinute, TimeSep, Time.wSecond );
	else
		wsprintf( outString, L"%s%u%s%02u", PadHours, hours, TimeSep, Time.wMinute );

	if ( !AmPm )
		return 0;

	fmt = ( ( fmt & OT_FORMAT ) - 1 );

	if ( fmt >= 12 ) {
		GetLocaleInfo( LOCALE_USER_DEFAULT, ( AmPm == 1 ) ? LOCALE_S1159 : LOCALE_S2359, AmPmString, SHORT_BUF_LEN );

		if ( fmt < 14 )
			wcscat_s( outString, MAX_ARG_LEN, L" " );

		if ( fmt == 13 )
			_wcslwr_s( AmPmString, SHORT_BUF_LEN );

		wcscat_s( outString, MAX_ARG_LEN, AmPmString );
		return 0;
	}

	if ( ( fmt < 4 ) || ( fmt == 8 ) || ( fmt == 9 ) )
		wcscat_s( outString, MAX_ARG_LEN, L" " );

	if ( ( fmt & 10 ) == 2 )
		wcscpy_s( AmPmString, SHORT_BUF_LEN, L"A.M." );
	else if ( fmt < 8 )
		wcscpy_s( AmPmString, SHORT_BUF_LEN, L"AM" );
	else
		wcscpy_s( AmPmString, SHORT_BUF_LEN, L"A" );

	if ( AmPm == 2 )
		AmPmString[0] = 'P';

	if ( fmt & 1 )
		_wcslwr_s( AmPmString, SHORT_BUF_LEN );

	wcscat_s( outString, MAX_ARG_LEN, AmPmString );
	return 0;
}



int TimePlus( SYSTEMTIME StartTime, long int AddDays, long int AddHours, long int AddMinutes, long int AddSeconds, long int AddMsec, long int &EndDate, SYSTEMTIME &EndTime )
{
	long int a = 0;
	EndDate = 0;
	long int EndHours = StartTime.wHour, EndMinutes = StartTime.wMinute, EndSeconds = StartTime.wSecond, EndMsec = StartTime.wMilliseconds;

	EndMsec += AddMsec;
	if ( EndMsec >= 1000 ) {
		a = EndMsec / 1000;
		EndMsec -= ( 1000 * a );
		EndSeconds += a;
	}
	else if ( EndMsec < 0 ) {
		a = ( ( 999 - EndMsec ) / 1000 );
		EndMsec += ( 1000 * a );
		EndSeconds -= a;
	}

	EndSeconds += AddSeconds;
	if ( EndSeconds >= 60 ) {
		a = EndSeconds / 60;
		EndSeconds -= ( 60 * a );
		EndMinutes += a;
	}
	else if ( EndSeconds < 0 ) {
		a = ( ( 59 - EndSeconds ) / 60 );
		EndSeconds += ( 60 * a );
		EndMinutes -= a;
	}

	EndMinutes += AddMinutes;
	if ( EndMinutes >= 60 ) {
		a = EndMinutes / 60;
		EndMinutes -= ( 60 * a );
		EndHours += a;
	}
	else if ( EndMinutes < 0 ) {
		a = ( ( 59 - EndMinutes ) / 60 );
		EndMinutes += ( 60 * a );
		EndHours -= a;
	}

	EndHours += AddHours;
	if ( EndHours >= 24 ) {
		a = EndHours / 24;
		EndHours -= ( 24 * a );
		EndDate += a;
	}
	else if ( EndHours < 0 ) {
		a = ( ( 23 - EndHours ) / 24 );
		EndHours += ( 24 * a );
		EndDate -= a;
	}

	EndDate += AddDays;

	EndTime.wHour			= (WORD) EndHours;
	EndTime.wMinute			= (WORD) EndMinutes;
	EndTime.wSecond			= (WORD) EndSeconds;
	EndTime.wMilliseconds	= (WORD) EndMsec;

	//  Printf( L"   * EndDate = %d  Hour = %d  Minute = %d  Second = %d  %Msec = %d\r\n", EndDate, EndTime.hour, EndTime.minute, EndTime.second, EndTime.ms );
	return 0;
}



int TimeCmp( SYSTEMTIME Time1, SYSTEMTIME Time2 )
{
	if ( Time1.wHour < Time2.wHour )
		return -1;
	else if ( Time1.wHour > Time2.wHour )
		return 1;

	if ( Time1.wMinute < Time2.wMinute )
		return -1;
	else if ( Time1.wMinute > Time2.wMinute )
		return 1;

	if ( Time1.wSecond < Time2.wSecond )
		return -1;
	else if ( Time1.wSecond > Time2.wSecond )
		return 1;

	if ( Time1.wMilliseconds < Time2.wMilliseconds )
		return -1;
	else if ( Time1.wMilliseconds > Time2.wMilliseconds )
		return 1;

	return 0;
}



int TimeDifference( SYSTEMTIME DateTime1, SYSTEMTIME DateTime2, long int &DaysDiff, SYSTEMTIME &TimeDiff, BOOL &Negative )
{
	long int Julian1 = 0, Julian2 = 0;
	int a = 0;


	if ( DateTime1.wYear ) {
		Julian1 = Julian( DateTime1 );
		Julian2 = Julian( DateTime2 );
	}

	if ( Julian1 > Julian2 )
		Negative = TRUE;
	else if ( ( Julian1 == Julian2 ) && ( TimeCmp( DateTime1, DateTime2 ) > 0 ) )
		Negative = TRUE;

	if ( Negative ) {
		SYSTEMTIME temp = DateTime1;
		DateTime1 = DateTime2;
		DateTime2 = temp;
		long int jt = Julian1;
		Julian1 = Julian2;
		Julian2 = jt;
	}

	DaysDiff = 0;

	a = DateTime2.wMilliseconds - DateTime1.wMilliseconds;
	if ( a >= 0 )
		TimeDiff.wMilliseconds = a;
	else {
		TimeDiff.wMilliseconds = 1000 + a;
		DateTime1.wSecond++;
	}

	a = DateTime2.wSecond - DateTime1.wSecond;
	if ( a >= 0 )
		TimeDiff.wSecond = a;
	else {
		TimeDiff.wSecond = 60 + a;
		DateTime1.wMinute++;
	}

	a = DateTime2.wMinute - DateTime1.wMinute;
	if ( a >= 0 )
		TimeDiff.wMinute = a;
	else {
		TimeDiff.wMinute = 60 + a;
		DateTime1.wHour++;
	}

	a = DateTime2.wHour - DateTime1.wHour;
	if ( a >= 0 )
		TimeDiff.wHour = a;
	else {
		TimeDiff.wHour = 24 + a;
		Julian1++;
	}

	DaysDiff = Julian2 - Julian1;
	return 0;
}



int ParseTimeWithOptionalDate( LPTSTR String, SYSTEMTIME &DateTime, BOOL &DateSpecified, int fmt, wchar_t *SepChar )
{
	wchar_t *ch = String, *sep1 = NULL, sv = '\0';


	if ( wcslen( String ) > 1 ) {
		sep1 = wcschr( String + 1, '@' );

		if ( !sep1 )
			sep1 = wcschr( String + 1, 'T' );

		if ( !sep1 )
			sep1 = wcschr( String + 1, 't' );
	}

	if ( sep1 ) {
		sv = *sep1;
		*sep1 = '\0';

		if ( ParseDate( String, fmt, DateTime ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date \"%s\"\r\n", String );
			return -1;
		}

		DateSpecified = TRUE;
		*sep1 = sv;
		ch = sep1 + 1;

		if ( *ch == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No time \"%s\"\r\n", String );
			return -1;
		}
	}
	else
		DateSpecified = FALSE;
	
	if ( ParseTime( ch, DateTime, NULL ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid time \"%s\"\r\n", String );
		return -1;
	}

	if ( SepChar )
		*SepChar = towupper( sv );

	return 0;
}



int ParseTimeOffset( LPTSTR inString, long int *AddDays, long int *AddHours, long int *AddMinutes, long int *AddSeconds, long int *AddMsec )
{
	long int Days = 0, Hours = 0, Minutes = 0, Seconds = 0, Msec = 0;
	BOOL Subtract = FALSE;

	long int a = 0;
	wchar_t *ch = inString, *sep1 = NULL;

	if ( *ch == '-' ) {
		Subtract = TRUE;
		ch++;
	}
	else if ( *ch == '+' )
		ch++;

	if ( !iswdigit( *ch ) )
		return -1;

	_set_errno( 0 );
	a = wcstol( ch, &sep1, 10 );

	if ( errno || ( sep1 == ch ) )
		return -1;

	if ( *sep1 == '@' ) {
		Days = a;
		ch = sep1 + 1;
	}
	else if ( *sep1 == 0x20 ) {
		Days = a;

		ch = sep1;
		while ( *ch == 0x20 )
			ch++;

		if ( !_wcsnicmp( ch, L"day", 3 ) ) {
			ch += 3;

			if ( towlower( *ch ) == 's' )
				ch++;
		}

		while ( *ch == 0x20 )
			ch++;

		if ( *ch == '\0' ) {
			if ( Subtract )
				*AddDays = -Days;
			else
				*AddDays = Days;

			*AddHours   = 0;
			*AddMinutes = 0;
			*AddSeconds = 0;
			*AddMsec    = 0;
			return 0;
		}
	}

	if ( !iswdigit( *ch ) )
		return -1;

	_set_errno( 0 );
	Hours = wcstol( ch, &sep1, 10 );

	if ( errno || ( sep1 == ch ) )
		return -1;

	if ( IsTimeSep( *sep1 ) ) {
		ch = sep1 + 1;

		if ( !iswdigit( *ch ) )
			return -1;

		_set_errno( 0 );
		Minutes = wcstol( ch, &sep1, 10 );

		if ( errno || ( sep1 == ch ) )
			return -1;

		if ( IsTimeSep( *sep1 ) ) {
			ch = sep1 + 1;

			if ( !iswdigit( *ch ) )
				return -1;

			_set_errno( 0 );
			Seconds = wcstol( ch, &sep1, 10 );

			if ( errno || ( sep1 == ch ) )
				return -1;

			if ( IsDecimalSep( *sep1 ) ) {
				ch = sep1 + 1;
				int roller = 100;
				while ( iswdigit( *ch ) ) {
					Msec += roller * ( *ch - '0' );
					roller /= 10;
					ch++;
				}
			}
			else if ( *sep1 != 0x00 )
				return -1;

		}
		else if ( *sep1 != 0x00 )
			return -1;

	}
	else if ( *sep1 != 0x00 )
		return -1;

	
	if ( Subtract ) {
		*AddDays    = -Days;
		*AddHours   = -Hours;
		*AddMinutes = -Minutes;
		*AddSeconds = -Seconds;
		*AddMsec    = -Msec;
	}
	else {
		*AddDays    = Days;
		*AddHours   = Hours;
		*AddMinutes = Minutes;
		*AddSeconds = Seconds;
		*AddMsec    = Msec;
	}
	
	return 0;
}



DLLExports int WINAPI f_timefmt( LPTSTR String )
{
	if ( !String )
		return 1;


	SYSTEMTIME Time;
	int timefmt = 0, ParseTimeFlags = 0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();


	if ( arglen[1] ) {
		timefmt = ParseInt( arg[1] );

		if ( timefmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time format \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[0] ) {
		if ( ParseTime(arg[0], Time, &ParseTimeFlags ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( arglen[1] == 0 ) {
			if ( ParseTimeFlags & PT_GOTSECONDS )
				timefmt |= OT_SHOWSECONDS;

			if ( ParseTimeFlags & PT_GOTMSEC )
				timefmt |= ( OT_SHOWSECONDS | OT_SHOWMSEC );
		}
	}
	else
		GetLocalTime( &Time );


	OutputTime( Time, timefmt, String );
	return 0;
}



DLLExports int WINAPI f_dectotime( LPTSTR String )
{
	if ( !String )
		return -1;


	int timefmt = OT_SHOWSECONDS;
	long int Days = 0;
	SYSTEMTIME Time;
	double DecimalTime = 0.0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	memset( &Time, 0x00, sizeof( SYSTEMTIME ) );


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No decimal time\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		timefmt = ParseInt( arg[1] );

		if ( timefmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time format \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	_set_errno( 0 );
	DecimalTime = _wtof( arg[0] );

	if ( errno ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad decimal time \"%s\"\r\n", arg[0] );
		return -1;
	}


	if ( DecimalTime < 0.0 ) {
		Days = (int) DecimalTime;

		DecimalTime = DecimalTime - Days;
		if ( DecimalTime < 0.0 ) {
			Days--;
			DecimalTime += 1.0;
		}
	}
	else if ( DecimalTime >= 1.0 ) {
		Days = (int) DecimalTime;
		DecimalTime -= Days;
	}

	if ( ( DecimalTime + TIME_FUDGE ) < 1.0 )
		DecimalTime += TIME_FUDGE;

	DecimalTime *= 24.0;

	Time.wHour		= (int) DecimalTime;
	DecimalTime		= ( DecimalTime - Time.wHour ) * 60.0;

	Time.wMinute	= (int) DecimalTime;
	DecimalTime		= ( DecimalTime - Time.wMinute ) * 60.0;

	Time.wSecond	= (int) DecimalTime;
	DecimalTime		= ( DecimalTime - Time.wSecond ) * 1000.0;

	Time.wMilliseconds	= (int) DecimalTime;


	String[0] = '\0';

	if ( Days ) {
		if ( ( Days == 1 ) || ( Days == -1 ) )
			swprintf_s( String, SHORT_BUF_LEN, L"%+d day ", Days );
		else
			swprintf_s( String, SHORT_BUF_LEN, L"%+d days ", Days );
	}

	OutputTime( Time, timefmt, String + wcslen( String ) );
	return 0;
}



DLLExports int WINAPI f_timetodec( LPTSTR String )
{
	if ( !String )
		return -1;


	int Flags = 0;
	long int Days = 0;
	SYSTEMTIME Time;
	double DecimalTime = 0.0;


	memset( &Time, 0x00, sizeof( SYSTEMTIME ) );


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No time\r\n" );
		return -1;
	}

	wchar_t *ch = arg[0], *sep1 = NULL;

	if ( iswdigit( *ch ) || ( ( ( *ch == '+' ) || ( *ch == '-' ) ) && iswdigit( *(ch + 1) ) ) ) {
		_set_errno( 0 );
		Days = wcstol( ch, &sep1, 10 );

		if ( ( errno == 0 ) && ( *sep1 == 0x20 ) ) {

			if ( ( Days >= ( MaximumJulian + 1 ) ) || ( Days < ( -MaximumJulian ) ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad days count\r\n" );
				return -1;
			}

			ch = sep1;
			while ( *ch == 0x20 )
				ch++;

			if ( !_wcsnicmp( ch, L"day", 3 ) ) {
				ch += 3;

				if ( towlower( *ch ) == 's' )
					ch++;

				while ( *ch == 0x20 )
					ch++;
			}
		}
		else
			Days = 0;
	}

	if ( ParseTime( ch, Time, &Flags ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad time \"%s\"\r\n", ch );
		return -1;
	}

	if ( ( Flags & ( PT_GOTHOURS | PT_GOTMINUTES ) ) != ( PT_GOTHOURS | PT_GOTMINUTES ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad time \"%s\"\r\n", ch );
		return -1;
	}

	DecimalTime = Days + ( Time.wHour / 24.0 ) + ( Time.wMinute / 1440.0 )  + ( Time.wSecond / 86400.0 ) + ( Time.wMilliseconds / 86400000.0 );

	swprintf_s( String, SHORT_BUF_LEN, L"%.9f", DecimalTime );
	return 0;
}



DLLExports int WINAPI f_timeplus( LPTSTR String )
{
	if ( !String )
		return -1;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );


	int fmt = 0x0100, timefmt = OT_SHOWSECONDS;
	wchar_t *ch = NULL, *sep1 = NULL, DateSepChar[2] = L"@";
	BOOL DateSpecified = FALSE;
	SYSTEMTIME StartDateTime;
	SYSTEMTIME EndDateTime;
	long int StartDay = 0, EndDay;
	long int AddDays = 0, AddHours = 0, AddMinutes = 0, AddSeconds = 0, AddMsec;
	wchar_t temp[SHORT_BUF_LEN] = L"";


	memset( &StartDateTime, 0x00, sizeof( SYSTEMTIME ) );
	memset( &EndDateTime, 0x00, sizeof( SYSTEMTIME ) );


	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		timefmt = ParseInt( arg[3] );

		if ( timefmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time format \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	if ( arglen[0] ) {
		if ( ParseTimeWithOptionalDate( arg[0], StartDateTime, DateSpecified, fmt, DateSepChar ) )
			return -1;
	}
	else
		StartDateTime	= Now;

	if ( arglen[1] ) {
		if ( ParseTimeOffset( arg[1], &AddDays, &AddHours, &AddMinutes, &AddSeconds, &AddMsec ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad offset \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	TimePlus( StartDateTime, AddDays, AddHours, AddMinutes, AddSeconds, AddMsec, EndDay, EndDateTime );

	OutputTime( EndDateTime, timefmt, temp );

	if ( DateSpecified ) {
		EndDay = EndDay + Julian( StartDateTime );
		MakeDate( EndDay, EndDateTime );
		OutputFormattedDate( EndDateTime, fmt, String );
		wcscat_s( String, SHORT_BUF_LEN, DateSepChar );
	}
	else if ( EndDay ) {
		if ( ( timefmt & OT_FORMAT ) == 0x0f ) {
			swprintf_s( String, SHORT_BUF_LEN, L"%+d", EndDay );
			timefmt = 0x0f;
		}
		else if ( ( EndDay == 1 ) || ( EndDay == -1 ) )
			swprintf_s( String, SHORT_BUF_LEN, L"%+d day %c", EndDay, DateSepChar[0] );
		else
			swprintf_s( String, SHORT_BUF_LEN, L"%+d days %c", EndDay, DateSepChar[0] );
	}
	else
		String[0] = '\0';

	wcscat_s( String, SHORT_BUF_LEN, temp );
	return 0;
}



DLLExports int WINAPI f_timediff( LPTSTR String )
{
	if ( !String )
		return -1;


	int fmt = 0x0100, timefmt = OT_SHOWSECONDS;
	SYSTEMTIME DateTime1;
	SYSTEMTIME DateTime2;
	SYSTEMTIME TimeDiff;
	BOOL Date1Specified = FALSE, Date2Specified = FALSE;
	long int DaysDiff = 0;
	BOOL Negative = FALSE;
	int Flags = 0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );
	GetLocalDateFormat();
	GetNow( FALSE );


	memset( &DateTime1, 0x00, sizeof( SYSTEMTIME ) );
	memset( &DateTime2, 0x00, sizeof( SYSTEMTIME ) );
	memset( &TimeDiff, 0x00, sizeof( SYSTEMTIME ) );


	if ( arglen[2] ) {
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid date format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		timefmt = ParseInt( arg[3] );

		if ( timefmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid time format \"%s\"\r\n", arg[3] );
			return -1;
		}

		if ( ( timefmt & OT_FORMAT ) != 0x0f )
			timefmt &= ~OT_FORMAT;
	}

	if ( arglen[0] ) {
		if ( ParseTimeWithOptionalDate( arg[0], DateTime1, Date1Specified, fmt, NULL ) )
			return -1;
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No time1\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseTimeWithOptionalDate( arg[1], DateTime2, Date2Specified, fmt, NULL ) )
			return -1;

		if ( Date1Specified && !Date2Specified ) {
			DateTime2.wYear			= DateTime1.wYear;
			DateTime2.wMonth		= DateTime1.wMonth;
			DateTime2.wDay			= DateTime1.wDay;
			DateTime2.wDayOfWeek	= DateTime1.wDayOfWeek;
		}
		else if ( Date2Specified && !Date1Specified ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No date \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DateTime2		= Now;
		Date2Specified	= Date1Specified;
	}

	if ( arglen[4] ) {
		Flags = ParseInt( arg[4] );

		if ( Flags < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags \"%s\"\r\n", arg[4] );
			return -1;
		}
	}


	TimeDifference( DateTime1, DateTime2, DaysDiff, TimeDiff, Negative );

	if ( ( Flags & 0x01 ) == 0 )
		Negative = FALSE;


	String[0] = '\0';

	if ( Negative )
		wcscpy_s( String, SHORT_BUF_LEN, L"-" );


	if ( DaysDiff ) {
		if ( ( timefmt & OT_FORMAT ) == 0x0f ) {
			wsprintf( String + wcslen( String ), L"%d", DaysDiff );
			timefmt = 0x0f;
		}
		else if ( DaysDiff == 1 )
			wsprintf( String + wcslen( String ), L"%d day ", DaysDiff );
		else
			wsprintf( String + wcslen( String ), L"%d days ", DaysDiff );
	}

	OutputTime( TimeDiff, timefmt, String + wcslen( String ) );
	return 0;
}


#endif	//	ISO8601TIME
