#ifndef POPCAL
#define POPCAL

#define POPCAL_VER_MAJOR		2
#define POPCAL_VER_MINOR		0
#define POPCAL_VER_BUILD		0


#define BoxFlagDisabled		0x0100
#define BoxFlagSunday		0x0200
#define BoxFlagSelected		0x0400
#define BoxFlagHoliday		0x0800

#define FlagNoCancel		0x0002
#define FlagRestrict		0x0004
#define FlagNoWeekends		0x0008
#define FlagPreferEarlier	0x0010
#define FlagPreferLater		0x0020
#define FlagPreferMask		0x0030
#define FlagNoHolidays      0x0080
#define FlagUseEnglish		0x0100
#define FlagNoFront         0x0200
#define FlagDiagnose		0x0400

#define inKeyMask			0xff00

#define NumBoxes			37

#define TickLength			1000



struct _PopCalData {
	SYSTEMTIME PickDate;
	SYSTEMTIME PickView;
	SYSTEMTIME MinDate;
	SYSTEMTIME MaxDate;
	SYSTEMTIME InitDate;
};



SYSTEMTIME pickDate, pickView;
int pickdatefmt = 0;
int DowMask = 0;
int ProgramFlags = 0;
int HolidaysYear = 0;

unsigned short int PopCalHolidays[12][31];

int BoxVal[NumBoxes];

HWND ThisWindowHandle = NULL;
HICON CalDateIcon     = NULL;
HHOOK KeyboardHookHandle = NULL;

HFONT HeaderFont = NULL;
HFONT DateFont = NULL;
HFONT OutOfRangeFont = NULL;
HFONT SelectedFont = NULL;
HFONT SmallFont = NULL;
HFONT ButtonFont = NULL;

HBRUSH SelectedDateBoxBackground = NULL;

UINT_PTR TimerID = NULL;
int CountDown = 10;
BOOL DoCountDown = FALSE;
BOOL KeystrokeEndsCountDown = FALSE;


int SundayTextColor  = 0;
int HolidayTextColor = 0;
int DisabledColor    = 0;



int MyCenterDialog( HWND hwnd )
{
	RECT ConsoleRect, DialogRect;
	long int ConsoleWidth, ConsoleHeight, DialogWidth, DialogHeight;

	GetWindowRect( GetConsoleWindow(), &ConsoleRect );
	GetWindowRect( hwnd, &DialogRect );

	ConsoleWidth = ConsoleRect.right - ConsoleRect.left;
	ConsoleHeight = ConsoleRect.bottom - ConsoleRect.top;
	DialogWidth = DialogRect.right - DialogRect.left;
	DialogHeight = DialogRect.bottom - DialogRect.top;
	
	SetWindowPos( hwnd, HWND_TOP,
		ConsoleRect.left + (ConsoleWidth - DialogWidth) / 2,
		ConsoleRect.top + (ConsoleHeight - DialogHeight) / 2,
		DialogWidth, DialogHeight, SWP_SHOWWINDOW );

	return 0;
}



int RgbLum( COLORREF Color )
{
	int r		= ( Color & 0x0000ff );
	int g		= ( Color & 0x00ff00 ) >> 8;
	int b		= ( Color & 0xff0000 ) >> 16;

	return ( r * 299 / 255 ) + ( g * 587 / 255 ) + ( b * 114 / 255 );
}



BOOL DrawStatusLine( HWND hwnd )
{
	wchar_t statusbuf[81];

	OutputFormattedDate( pickDate, ( ProgramFlags & FlagDiagnose ) ? pickdatefmt : ( ( ProgramFlags & FlagUseEnglish ) ? 17 : 32 ), statusbuf );
	SetDlgItemText( hwnd, STATUSLINE, statusbuf );
	return FALSE;
}



BOOL FillDowLine( HWND hwnd )
{
	for ( int i=0; i < 7; i++ ) {
		SetDlgItemText( hwnd, DOWBOX0 + i, ( ProgramFlags & FlagUseEnglish ) ? DowName[( i + LeftColumnDow ) % 7] : LocalDowName[( i + LeftColumnDow ) % 7] );
		SendDlgItemMessage( hwnd, DOWBOX0 + i, WM_SETFONT, (WPARAM) SmallFont, TRUE );
	}
	return 0;
}



BOOL DrawPopCalendar( HWND hwnd, _PopCalData *Data )
{
	wchar_t shortbuf[MAX_ARG_LEN + 1];

	long julian = Julian( pickView.wYear, pickView.wMonth, 1 );
	BOOL leap = Leapness( pickView.wYear );
	int b, boxnum;
	BOOL atStart, atEnd;


	if ( ( pickView.wYear != HolidaysYear ) && ( ( ProgramFlags & FlagNoHolidays ) == 0 ) && FirstHolidayDef ) {
		HolidaysYear = pickView.wYear;
		ExpandHolidays( HolidaysYear, PopCalHolidays );
	}

	b = ( julian + 7 - LeftColumnDow ) % 7;

	for ( int i=0; i < NumBoxes; i++ ) {
		boxnum = ( i - b + 1 );
		if ( ( boxnum < 1 ) || ( boxnum > MonthLength[pickView.wMonth][leap] ) ) {
			BoxVal[i] = 0;
			ShowWindow( GetDlgItem( hwnd, DATEBOX0 + i ), FALSE );
		}
		else if ( ( ( pickView.wYear == Data->MinDate.wYear ) && ( pickView.wMonth == Data->MinDate.wMonth ) && ( boxnum < (int) Data->MinDate.wDay ) ) ||
				  ( ( pickView.wYear == Data->MaxDate.wYear ) && ( pickView.wMonth == Data->MaxDate.wMonth ) && ( boxnum > (int) Data->MaxDate.wDay ) ) ||
				  ( DowMask & ( 1 << ( ( i + LeftColumnDow ) % 7 ) ) ) ) {
			BoxVal[i] = boxnum | BoxFlagDisabled;
			wsprintf( shortbuf, L"%u", boxnum );
			SetDlgItemText( hwnd, DATEBOX0 + i, shortbuf );
			SendDlgItemMessage( hwnd, DATEBOX0 + i, WM_SETFONT, (WPARAM) OutOfRangeFont, TRUE );
			ShowWindow( GetDlgItem( hwnd, DATEBOX0 + i ), TRUE );
		}
		else if ( ( boxnum == pickDate.wDay ) && ( pickView.wMonth == pickDate.wMonth ) && ( pickView.wYear == pickDate.wYear ) ) {
			BoxVal[i] = boxnum | BoxFlagSelected;
			if ( ( ( i + LeftColumnDow ) % 7 ) == 0 )
				BoxVal[i] |= BoxFlagSunday;

			if ( PopCalHolidays[pickView.wMonth - 1][boxnum - 1] )
				BoxVal[i] |= BoxFlagHoliday;

			wsprintf( shortbuf, L"%u", boxnum );
			SetDlgItemText( hwnd, DATEBOX0 + i, shortbuf );
			SendDlgItemMessage( hwnd, DATEBOX0 + i, WM_SETFONT, (WPARAM) SelectedFont, TRUE );
			ShowWindow( GetDlgItem( hwnd, DATEBOX0 + i ), TRUE );
		}
		else {
			BoxVal[i] = boxnum;
			if ( ( ( i + LeftColumnDow ) % 7 ) == 0 )
				BoxVal[i] |= BoxFlagSunday;

			if ( PopCalHolidays[pickView.wMonth - 1][boxnum - 1] )
				BoxVal[i] |= BoxFlagHoliday;

			wsprintf( shortbuf, L"%u", boxnum );
			SetDlgItemText( hwnd, DATEBOX0 + i, shortbuf );
			SendDlgItemMessage( hwnd, DATEBOX0 + i, WM_SETFONT, (WPARAM) DateFont, TRUE );
			ShowWindow( GetDlgItem( hwnd, DATEBOX0 + i ), TRUE );
		}
	}

	atStart = ( ( pickView.wYear == Data->MinDate.wYear ) && ( pickView.wMonth == Data->MinDate.wMonth ) );
	atEnd = ( ( pickView.wYear == Data->MaxDate.wYear ) && ( pickView.wMonth == Data->MaxDate.wMonth ) );

	EnableWindow( GetDlgItem( hwnd, BACKBUTTON ), !atStart );
	EnableWindow( GetDlgItem( hwnd, YEARBACKBUTTON ), !atStart );
	EnableWindow( GetDlgItem( hwnd, FORWARDBUTTON ), !atEnd );
	EnableWindow( GetDlgItem( hwnd, YEARFORWARDBUTTON ), !atEnd );
	
	wsprintf( shortbuf, L"%s %u", ( ProgramFlags & FlagUseEnglish ) ? MonthName[pickView.wMonth] : LocalMonthNameLong[pickView.wMonth], pickView.wYear );
	SetDlgItemText( hwnd, HEADERLINE, shortbuf );

	return 0;
}



int TerminateMyTimer()
{
	if ( TimerID ) {
		SetDlgItemText( ThisWindowHandle, IDOK, L"OK" );
		KillTimer( NULL, TimerID );
		TimerID = NULL;
	}

	return 0;
}



BOOL SeekUp( long int &julian, _PopCalData *Data )
{
	if ( ( DowMask & 0x7f ) == 0x7f )
		return FALSE;

	if ( ( DowMask & ( 1 << ( julian % 7 ) ) ) == 0 )
		return TRUE;

	while ( DowMask & ( 1 << ( julian % 7 ) ) )
		julian++;

	if ( julian > Julian( Data->MaxDate ) )
		return FALSE;

	return TRUE;
}



BOOL SeekDown( long int &julian, _PopCalData *Data )
{
	if ( ( DowMask & 0x7f ) == 0x7f )
		return FALSE;

	if ( ( DowMask & ( 1 << ( julian % 7 ) ) ) == 0 )
		return TRUE;

	while ( DowMask & ( 1 << ( julian % 7 ) ) )
		julian--;

	if ( julian < Julian( Data->MinDate ) )
		return FALSE;

	return TRUE;
}



BOOL BackMonth( HWND hwnd, _PopCalData *Data )
{
	unsigned int trymonth = pickView.wMonth;
	unsigned int tryyear = pickView.wYear;

	trymonth--;
	if ( trymonth < 1 ) {
		trymonth = 12;
		tryyear--;
	}

	if ( tryyear < Data->MinDate.wYear )
		return FALSE;

	if ( ( tryyear == Data->MinDate.wYear ) && ( trymonth < Data->MinDate.wMonth ) )
		return FALSE;

	pickView.wMonth = trymonth;
	pickView.wYear = tryyear;
	DrawPopCalendar( hwnd, Data );
	return TRUE;
}



BOOL AheadMonth( HWND hwnd, _PopCalData *Data )
{
	unsigned int trymonth = pickView.wMonth;
	unsigned int tryyear = pickView.wYear;

	trymonth++;
	if (trymonth > 12) {
		trymonth = 1;
		tryyear++;
	}

	if ( tryyear > Data->MaxDate.wYear )
		return FALSE;

	if ( ( tryyear == Data->MaxDate.wYear ) && ( trymonth > Data->MaxDate.wMonth ) )
		return FALSE;

	pickView.wMonth = trymonth;
	pickView.wYear = tryyear;
	DrawPopCalendar( hwnd, Data );
	return TRUE;
}



BOOL BackYear( HWND hwnd, _PopCalData *Data )
{
	unsigned int tryyear = pickView.wYear;
	unsigned int trymonth = pickView.wMonth;

	if ( ( tryyear == Data->MinDate.wYear ) && ( trymonth == Data->MinDate.wMonth ) )
		return FALSE;

	if ( tryyear == Data->MinDate.wYear )
		trymonth = Data->MinDate.wMonth;
	else {
		tryyear--;

		if ( ( tryyear == Data->MinDate.wYear ) && ( trymonth < Data->MinDate.wMonth ) )
			trymonth = Data->MinDate.wMonth;
	}

	pickView.wYear  = tryyear;
	pickView.wMonth = trymonth;

	DrawPopCalendar( hwnd, Data );
	return TRUE;
}



BOOL AheadYear( HWND hwnd, _PopCalData *Data )
{
	unsigned int tryyear = pickView.wYear;
	unsigned int trymonth = pickView.wMonth;

	if ( ( tryyear == Data->MaxDate.wYear ) && ( trymonth == Data->MaxDate.wMonth ) )
		return FALSE;

	if ( tryyear == Data->MaxDate.wYear )
		trymonth = Data->MaxDate.wMonth;
	else {
		tryyear++;

		if ( ( tryyear == Data->MaxDate.wYear ) && ( trymonth > Data->MaxDate.wMonth ) )
			trymonth = Data->MaxDate.wMonth;
	}

	pickView.wYear  = tryyear;
	pickView.wMonth = trymonth;

	DrawPopCalendar( hwnd, Data );
	return TRUE;
}



BOOL SelBackDay( HWND hwnd, _PopCalData *Data )
{
	long int julian = Julian( pickDate ) - 1;
	if ( julian < Julian( Data->MinDate ) )
		return FALSE;

	if ( !SeekDown( julian, Data ) )
		return FALSE;

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelAheadDay( HWND hwnd, _PopCalData *Data )
{
	long int julian = Julian( pickDate ) + 1;
	if ( julian > Julian( Data->MaxDate ) )
		return FALSE;

	if ( !SeekUp( julian, Data ) )
		return FALSE;

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelBackWeek( HWND hwnd, _PopCalData *Data )
{
	long int julian = Julian( pickDate ) - 7;
	if ( julian < Julian( Data->MinDate ) )
		return FALSE;

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelAheadWeek( HWND hwnd, _PopCalData *Data )
{
	long int julian = Julian( pickDate ) + 7;
	if ( julian > Julian( Data->MaxDate ) )
		return FALSE;

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelBackMonth( HWND hwnd, _PopCalData *Data )
{
	int tryyear  = pickDate.wYear;
	int trymonth = pickDate.wMonth;
	int tryday   = pickDate.wDay;

	if ( ( tryyear == Data->MinDate.wYear ) && ( trymonth == Data->MinDate.wMonth ) && ( tryday == Data->MinDate.wDay ) )
		return FALSE;

	long int kjul = Julian( tryyear, trymonth, 1 );
	long int ljul = kjul;
	if ( !SeekUp( ljul, Data ) )
		return FALSE;

	int first = ljul - kjul + 1;

	if ( ( ( tryyear != Data->MinDate.wYear ) || ( trymonth != Data->MinDate.wMonth) ) && ( tryday == first ) ) {
		trymonth--;
		if (trymonth < 1) {
			tryyear--;
			trymonth = 12;
		}
		tryday = 1;
	}
	else if ( ( tryyear != Data->MinDate.wYear ) || ( trymonth != Data->MinDate.wMonth ) )
		tryday = first;
	else
		tryday = Data->MinDate.wDay;

	long int julian = Julian( tryyear, trymonth, tryday );

	if ( !SeekUp( julian, Data ) )
		return FALSE;

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelAheadMonth( HWND hwnd, _PopCalData *Data )
{
	int tryyear  = pickDate.wYear;
	int trymonth = pickDate.wMonth;
	int tryday   = pickDate.wDay;

	if ( ( tryyear == Data->MaxDate.wYear ) && ( trymonth == Data->MaxDate.wMonth ) && ( tryday == Data->MaxDate.wDay ) )
		return FALSE;

	if ( ( tryyear != Data->MaxDate.wYear ) || ( trymonth != Data->MaxDate.wMonth ) ) {
		trymonth++;
		if (trymonth > 12) {
			tryyear++;
			trymonth = 1;
		}
		tryday = 1;
	}
	else
		tryday = Data->MaxDate.wDay;

	long int julian = Julian( tryyear, trymonth, tryday );

	if ( ( tryyear == Data->MaxDate.wYear ) && ( trymonth == Data->MaxDate.wMonth ) && ( tryday > 15 ) ) {
		if ( !SeekDown( julian, Data ) )
			return FALSE;
	}
	else {
		if ( !SeekUp( julian, Data ) )
			return FALSE;
	}

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}


BOOL SelBackYear( HWND hwnd, _PopCalData *Data )
{
	unsigned int tryyear  = pickDate.wYear;
	unsigned int trymonth = pickDate.wMonth;
	unsigned int tryday   = pickDate.wDay;

	if ( ( tryyear == Data->MinDate.wYear ) && ( trymonth == Data->MinDate.wMonth ) && ( tryday == Data->MinDate.wDay ) )
		return FALSE;

	long int kjul = Julian( tryyear, trymonth, 1 );
	long int ljul = kjul;

	if ( !SeekUp( ljul, Data ) )
		return FALSE;

	int first = ljul - kjul + 1;

	if ( ( tryyear > Data->MinDate.wYear ) && ( trymonth == 1 ) && ( tryday == first ) ) {
		tryyear--;
		tryday = 1;
		if ( tryyear == Data->MinDate.wYear ) {
			trymonth = Data->MinDate.wMonth;
			tryday = Data->MinDate.wDay;
		}
	}
	else if ( tryyear > Data->MinDate.wYear ) {
		trymonth = 1;
		tryday = 1;
	}
	else {
		trymonth = Data->MinDate.wMonth;
		tryday = Data->MinDate.wDay;
	}

	long int julian = Julian( tryyear, trymonth, tryday );

	if ( !SeekUp( julian, Data ) )
		return FALSE;

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelAheadYear( HWND hwnd, _PopCalData *Data )
{
	unsigned int tryyear  = pickDate.wYear;
	unsigned int trymonth = pickDate.wMonth;
	unsigned int tryday   = pickDate.wDay;

	if ( ( tryyear == Data->MaxDate.wYear ) && ( trymonth == Data->MaxDate.wMonth ) && ( tryday == Data->MaxDate.wDay ) )
		return FALSE;

	if ( tryyear < Data->MaxDate.wYear ) {
		tryyear++;
		trymonth = 1;
		tryday = 1;
	}
	else {
		trymonth = Data->MaxDate.wMonth;
		tryday = Data->MaxDate.wDay;
	}

	long int julian = Julian( tryyear, trymonth, tryday );

	if ( ( tryyear == Data->MaxDate.wYear ) && ( trymonth == Data->MaxDate.wMonth ) ) {
		if ( !SeekDown( julian, Data ) )
			return FALSE;
	}
	else {
		if ( !SeekUp( julian, Data ) )
			return FALSE;
	}

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelWeekTop( HWND hwnd, _PopCalData *Data )
{
	long int julian = Julian( pickDate );
	long int minjul = Julian( Data->MinDate );
	long int maxjul = Julian( Data->MinDate );

	long int t = Julian( pickDate.wYear, pickDate.wMonth, 1 );
	if ( minjul < t )
		minjul = t;

	t = t + MonthLength[pickDate.wMonth][Leapness( pickDate.wYear )] - 1;
	if ( maxjul > t )
		maxjul = t;

	int col = ( julian - LeftColumnDow + 7 ) % 7;

	julian -= col;

	if ( julian < minjul )
		julian = minjul;

	if ( !SeekUp( julian, Data ) )
		return FALSE;

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelWeekEnd( HWND hwnd, _PopCalData *Data )
{
	long int julian = Julian( pickDate );
	long int minjul = Julian( Data->MinDate );
	long int maxjul = Julian( Data->MaxDate );

	long int t = Julian( pickDate.wYear, pickDate.wMonth, 1 );
	if ( minjul < t )
		minjul = t;

	t = t + MonthLength[pickDate.wMonth][Leapness( pickDate.wYear )] - 1;
	if ( maxjul > t )
		maxjul = t;

	int col = ( julian - LeftColumnDow + 7 ) % 7;

	julian += ( 6 - col );

	if ( julian > maxjul )
		julian = maxjul;

	if ( !SeekDown( julian, Data ) )
		return FALSE;

	MakeDate( julian, pickDate );

	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelMinimum( HWND hwnd, _PopCalData *Data )
{
	long int julian = Julian( Data->MinDate );
	if ( !SeekUp( julian, Data ) )
		return FALSE;

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelMaximum( HWND hwnd, _PopCalData *Data )
{
	long int julian = Julian( Data->MaxDate );
	if ( !SeekDown( julian, Data ) )
		return FALSE;

	MakeDate( julian, pickDate );
	pickView = pickDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL SelInit( HWND hwnd, _PopCalData *Data )
{
	pickDate = Data->InitDate;
	pickView = Data->InitDate;
	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL ChangeSelection( HWND hwnd, int boxnum, _PopCalData *Data )
{
	if ( BoxVal[boxnum] == 0 )
		return FALSE;

	if ( BoxVal[boxnum] & BoxFlagDisabled )
		return FALSE;

	pickDate.wYear  = pickView.wYear;
	pickDate.wMonth = pickView.wMonth;
	pickDate.wDay   = BoxVal[boxnum] & 31;

	DrawPopCalendar( hwnd, Data );
	DrawStatusLine( hwnd );
	return TRUE;
}



BOOL QuickExit( HWND hwnd, int boxnum, _PopCalData *Data )
{
	if ( ChangeSelection( hwnd, boxnum, Data ) )
		EndDialog( hwnd, IDOK );

	return FALSE;
}



BOOL ChangeLeftColumn (HWND hwnd, _PopCalData *Data )
{
	LeftColumnDow = ( LeftColumnDow + 1 ) % 7;
	
	FillDowLine( hwnd );
	DrawPopCalendar( hwnd, Data );
	return TRUE;
}



int EndCountDown( _PopCalData *Data )
{
	DoCountDown = FALSE;
	CountDown = 0;
	SetDlgItemText( ThisWindowHandle, IDOK, L"OK" );
	return 0;
}



LRESULT CALLBACK MyKeyboardProc( int code, WPARAM wParam, LPARAM lParam )
{
	_PopCalData *Data		= (_PopCalData*) GetWindowLongPtr( ThisWindowHandle, DWLP_USER );	//	Fix me!


	if ( code < 0 )
		return CallNextHookEx( KeyboardHookHandle, code, wParam, lParam );

	if ( KeystrokeEndsCountDown )
		EndCountDown( Data );

	switch ( wParam ) {

	case VK_LEFT:
		if ( !( lParam & 0x80000000 ) )
			PostMessage( ThisWindowHandle, WM_COMMAND, SELECTBACKDAY | 0x10000, 0 );
		return 1;

	case VK_RIGHT:
		if ( !( lParam & 0x80000000 ) )
			PostMessage( ThisWindowHandle, WM_COMMAND, SELECTAHEADDAY | 0x10000, 0 );
		return 1;

	case VK_UP:
		if ( !( lParam & 0x80000000 ) )
			PostMessage( ThisWindowHandle, WM_COMMAND, SELECTBACKWEEK | 0x10000, 0 );
		return 1;

	case VK_DOWN:
		if ( !( lParam & 0x80000000 ) )
			PostMessage( ThisWindowHandle, WM_COMMAND, SELECTAHEADWEEK | 0x10000, 0 );
		return 1;

	case VK_PRIOR:
		if ( !( lParam & 0x80000000 ) ) {
			if ( GetKeyState( VK_CONTROL ) & inKeyMask )
				PostMessage( ThisWindowHandle, WM_COMMAND, SELECTBACKYEAR | 0x10000, 0 );
			else
				PostMessage( ThisWindowHandle, WM_COMMAND, SELECTBACKMONTH | 0x10000, 0 );
		}
		return 1;

	case VK_NEXT:
		if ( !( lParam & 0x80000000 ) ) {
			if ( GetKeyState( VK_CONTROL ) & inKeyMask )
				PostMessage( ThisWindowHandle, WM_COMMAND, SELECTAHEADYEAR | 0x10000, 0 );
			else
				PostMessage( ThisWindowHandle, WM_COMMAND, SELECTAHEADMONTH | 0x10000, 0 );
		}
		return 1;

	case VK_HOME:
		if ( !( lParam & 0x80000000 ) ) {
			if ( GetKeyState( VK_CONTROL ) & inKeyMask )
				PostMessage( ThisWindowHandle, WM_COMMAND, SELECTMINIMUM | 0x10000, 0 );
			else
				PostMessage( ThisWindowHandle, WM_COMMAND, SELECTWEEKTOP | 0x10000, 0 );
		}
		return 1;

	case VK_END:
		if ( !( lParam & 0x80000000 ) ) {
			if ( GetKeyState( VK_CONTROL ) & inKeyMask )
				PostMessage( ThisWindowHandle, WM_COMMAND, SELECTMAXIMUM | 0x10000, 0 );
			else
				PostMessage( ThisWindowHandle, WM_COMMAND, SELECTWEEKEND | 0x10000, 0 );
		}
		return 1;

	case VK_BACK:
		if ( !( lParam & 0x80000000 ) )
			PostMessage( ThisWindowHandle, WM_COMMAND, SELECTINITIAL | 0x10000, 0 );
		return 1;

	case VK_PAUSE:
		if ( !( lParam & 0x80000000 ) ) {
			if ( GetKeyState( VK_SHIFT ) & inKeyMask )
				PostMessage( ThisWindowHandle, WM_COMMAND, OTHERBUTTON | 0x10000, 0 );
		}
		return 1;

	default:
		return CallNextHookEx( KeyboardHookHandle, code, wParam, lParam );
	}
}



VOID CALLBACK MyTimerCallback( HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime )
{
	wchar_t shortbuf[20];

	KeystrokeEndsCountDown = TRUE;

	if ( !DoCountDown )
		return;

	if ( CountDown == 0 )
		EndDialog( ThisWindowHandle, IDOK );

	CountDown--;
	wsprintf( shortbuf, L"OK - %u", CountDown );
	SetDlgItemText( ThisWindowHandle, IDOK, shortbuf );
}



int DeleteMyObjects( _PopCalData *Data )
{
	if (HeaderFont)			DeleteObject(HeaderFont);
	if (DateFont)			DeleteObject(DateFont);
	if (SelectedFont)		DeleteObject(SelectedFont);
	if (OutOfRangeFont)		DeleteObject(OutOfRangeFont);
	if (SmallFont)			DeleteObject(SmallFont);
	if (ButtonFont)			DeleteObject(ButtonFont);

	if (SelectedDateBoxBackground)	DeleteObject(SelectedDateBoxBackground);

	if (KeyboardHookHandle)	UnhookWindowsHookEx(KeyboardHookHandle);
	
	TerminateMyTimer();

	return 0;
}



BOOL InitializeDialog( HWND hwnd, _PopCalData *Data )
{
	HDC hdc;
	long lfHeight1, lfHeight2, lfHeight3, lfHeight4;

	ThisWindowHandle = hwnd;

	CalDateIcon = LoadIcon( MyDLLHandle, L"CALDATE_ICON" );
	if ( CalDateIcon )
		SendMessage( hwnd, WM_SETICON, ICON_BIG, (LPARAM) CalDateIcon );

	if ( arglen[0] )
		SetWindowText( hwnd, arg[0] );

	MyCenterDialog( hwnd );

	hdc = GetDC( NULL );

	//  Calculate heights for the various fonts used:
	lfHeight1 = -MulDiv( 16, GetDeviceCaps( hdc, LOGPIXELSY ), 72 );  // big
	lfHeight2 = -MulDiv( 12, GetDeviceCaps( hdc, LOGPIXELSY ), 72 );  // medium
	lfHeight3 = -MulDiv( 9, GetDeviceCaps( hdc, LOGPIXELSY ), 72 );   // small

	if ( ( ( ProgramFlags & FlagUseEnglish ) > 0 ) || ( ( GetUserDefaultLangID() & 0x00ff ) == 0x09 ) )
		lfHeight4 = -MulDiv( 16, GetDeviceCaps( hdc, LOGPIXELSY ), 72 );  // super big for English
	else
		lfHeight4 = -MulDiv( 12, GetDeviceCaps( hdc, LOGPIXELSY ), 72 );  // not so big for non-English

	ReleaseDC( NULL, hdc );

	HeaderFont = CreateFont( lfHeight4, 0, 0, 0, FW_BOLD, 0, 0, 0, 0, 0, 0, 0, 0, L"Arial" );
	if (!HeaderFont) EndDialog( hwnd, -101 );

	DateFont = CreateFont( lfHeight1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, L"Arial" );
	if (!DateFont) EndDialog( hwnd, -102 );

	SelectedFont = CreateFont( lfHeight1, 0, 0, 0, FW_BOLD, 0, 0, 0, 0, 0, 0, 0, 0, L"Arial" );
	if (!SelectedFont) EndDialog( hwnd, -103 );

	OutOfRangeFont = CreateFont( lfHeight1, 0, 0, 0, 0, TRUE, 0, 0, 0, 0, 0, 0, 0, L"Arial" );
	if (!OutOfRangeFont) EndDialog( hwnd, -104 );

	SmallFont = CreateFont( lfHeight3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, L"Arial" );
	if (!SmallFont) EndDialog( hwnd, -105 );

	ButtonFont = CreateFont( lfHeight2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, L"Arial" );
	if (!ButtonFont) EndDialog( hwnd, -106 );

	SendDlgItemMessage( hwnd, HEADERLINE, WM_SETFONT, (WPARAM) HeaderFont, TRUE );
	SendDlgItemMessage( hwnd, BACKBUTTON, WM_SETFONT, (WPARAM) ButtonFont, TRUE );
	SendDlgItemMessage( hwnd, FORWARDBUTTON, WM_SETFONT, (WPARAM) ButtonFont, TRUE );
	SendDlgItemMessage( hwnd, YEARBACKBUTTON, WM_SETFONT, (WPARAM) SmallFont, TRUE );
	SendDlgItemMessage( hwnd, YEARFORWARDBUTTON, WM_SETFONT, (WPARAM) SmallFont, TRUE );
	SendDlgItemMessage( hwnd, IDOK, WM_SETFONT, (WPARAM) ButtonFont, TRUE );
	SendDlgItemMessage( hwnd, IDCANCEL, WM_SETFONT, (WPARAM) ButtonFont, TRUE );
	SendDlgItemMessage( hwnd, OTHERBUTTON, WM_SETFONT, (WPARAM) ButtonFont, TRUE );
	SendDlgItemMessage( hwnd, STATUSLINE, WM_SETFONT, (WPARAM) SmallFont, TRUE );

	if ( RgbLum( GetSysColor( COLOR_BTNFACE ) ) >= 500 ) {
		SelectedDateBoxBackground = CreateSolidBrush( RGB( 0xb8, 0xb8, 0xff ) );
		SundayTextColor  = RGB( 0xb0, 0x00, 0x00 );
		HolidayTextColor = RGB( 0x00, 0x00, 0xc0 );
	}
	else {
		SelectedDateBoxBackground = CreateSolidBrush( RGB( 0x30, 0x30, 0x60 ) );
		SundayTextColor  = RGB( 0xff, 0x40, 0x40 );
		HolidayTextColor = RGB( 0x60, 0x80, 0xff );
	}

	DisabledColor = GetSysColor( COLOR_GRAYTEXT );

	KeyboardHookHandle = SetWindowsHookEx( WH_KEYBOARD, (HOOKPROC) MyKeyboardProc, NULL, GetCurrentThreadId() );

	if ( ProgramFlags & FlagNoCancel ) {
		EnableWindow( GetDlgItem( hwnd, IDCANCEL ), FALSE );
		ShowWindow( GetDlgItem( hwnd, IDCANCEL ), FALSE );
		// HMENU hMenu = GetSystemMenu( hwnd, FALSE );
		// RemoveMenu( hMenu, SC_CLOSE, MF_BYCOMMAND );
		// DrawMenuBar( hwnd );
	}

	DrawStatusLine( hwnd );

	if ( CountDown ) {
		KeystrokeEndsCountDown = FALSE;
		wchar_t shortbuf[20];
		wsprintf( shortbuf, L" OK - %u", CountDown );
		SetDlgItemText( hwnd, IDOK, shortbuf );
		DoCountDown = TRUE;
	}
	else
		DoCountDown = FALSE;

	TimerID = SetTimer( NULL, NULL, TickLength, (TIMERPROC) MyTimerCallback );

	return 0;
}



INT_PTR CALLBACK PopCalDlgProc( HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam )
{
	_PopCalData *Data		= (_PopCalData*) GetWindowLongPtr( hwnd, DWLP_USER );

	switch ( Message ) {

	case WM_INITDIALOG:
		SetWindowLongPtr( hwnd, DWLP_USER, lParam );
		InitializeDialog( hwnd, Data );
		FillDowLine( hwnd );
		DrawPopCalendar( hwnd, Data );
		return TRUE;

	case WM_COMMAND: {
		TerminateMyTimer();

		switch ( LOWORD ( wParam ) ) {

		case IDOK:
			EndDialog( hwnd, IDOK );
			break;

		case IDCANCEL:
			EndDialog( hwnd, IDCANCEL );
			break;

		case BACKBUTTON:
			BackMonth( hwnd, Data );
			break;

		case FORWARDBUTTON:
			AheadMonth( hwnd, Data );
			break;

		case YEARBACKBUTTON:
			BackYear( hwnd, Data );
			break;

		case YEARFORWARDBUTTON:
			AheadYear( hwnd, Data );
			break;

		case SELECTBACKDAY:
			SelBackDay( hwnd, Data );
			break;

		case SELECTAHEADDAY:
			SelAheadDay( hwnd, Data );
			break;

		case SELECTBACKWEEK:
			SelBackWeek( hwnd, Data );
			break;

		case SELECTAHEADWEEK:
			SelAheadWeek( hwnd, Data );
			break;

		case SELECTBACKMONTH:
			SelBackMonth( hwnd, Data );
			break;

		case SELECTAHEADMONTH:
			SelAheadMonth( hwnd, Data );
			break;

		case SELECTBACKYEAR:
			SelBackYear( hwnd, Data );
			break;

		case SELECTAHEADYEAR:
			SelAheadYear( hwnd, Data );
			break;

		case SELECTWEEKTOP:
			SelWeekTop( hwnd, Data );
			break;

		case SELECTWEEKEND:
			SelWeekEnd( hwnd, Data );
			break;

		case SELECTMINIMUM:
			SelMinimum( hwnd, Data );
			break;

		case SELECTMAXIMUM:
			SelMaximum( hwnd, Data );
			break;

		case SELECTINITIAL:
			SelInit( hwnd, Data );
			break;

		case OTHERBUTTON:
			ChangeLeftColumn( hwnd, Data );
			break;

		case STATUSLINE:
			if ( HIWORD( wParam ) == STN_CLICKED ) {
				pickView = pickDate;
				DrawPopCalendar( hwnd, Data );
			}
			break;

		default:
			if ( ( LOWORD( wParam ) >= DATEBOX0 ) && ( LOWORD( wParam ) <= ( DATEBOX0 + NumBoxes - 1 ) ) && ( HIWORD( wParam ) == STN_CLICKED ) )
				ChangeSelection( hwnd, LOWORD( wParam ) - DATEBOX0, Data );
			else if ( ( LOWORD( wParam ) >= DATEBOX0 ) && ( LOWORD( wParam ) <= ( DATEBOX0 + NumBoxes - 1 ) ) && ( HIWORD( wParam ) == STN_DBLCLK ) )
				QuickExit( hwnd, LOWORD( wParam ) - DATEBOX0, Data );
			}
		}

		case WM_CTLCOLORSTATIC: {
			HDC hdcStatic = (HDC) wParam;
			int ID = GetDlgCtrlID( ( HWND ) lParam);

			if ( ( ID >= DATEBOX0 ) && ( ID < ( DATEBOX0 + NumBoxes ) ) ) {
				int box = ID - DATEBOX0;
				if ( BoxVal[box] & BoxFlagSelected ) {
					SetBkMode( hdcStatic, TRANSPARENT );
					if ( BoxVal[box] & BoxFlagHoliday )
						SetTextColor( hdcStatic, HolidayTextColor );
					else if ( BoxVal[box] & BoxFlagSunday )
						SetTextColor( hdcStatic, SundayTextColor );
					return (LONG) SelectedDateBoxBackground;
				}
				else if ( BoxVal[box] & BoxFlagDisabled ) {
					SetTextColor( hdcStatic, DisabledColor );
					SetBkMode( hdcStatic, TRANSPARENT );
					return (LONG) (COLOR_BTNFACE + 1);
				}
				else if ( BoxVal[box] & BoxFlagHoliday ) {
					SetBkMode( hdcStatic, TRANSPARENT );
					SetTextColor( hdcStatic, HolidayTextColor );
					return (LONG) (COLOR_BTNFACE + 1);
				}
				else if ( BoxVal[box] & BoxFlagSunday ) {
					SetTextColor( hdcStatic, SundayTextColor );
					SetBkMode( hdcStatic, TRANSPARENT );
					return (LONG) (COLOR_BTNFACE + 1);
				}
			}
			else if ( ( ID == ( DOWBOX0 + ( ( 7 - LeftColumnDow ) % 7 ) ) ) ) {
				SetTextColor( hdcStatic, SundayTextColor );
				SetBkMode( hdcStatic, TRANSPARENT );
				return (LONG) (COLOR_BTNFACE + 1);
				}
			}
			return FALSE;

		case WM_DESTROY:
			DeleteMyObjects( Data );
			return 0;

		default:
			return FALSE;
	}

	return TRUE;
}



BOOL ResolveMaskedSelection( _PopCalData *Data )
{
	long int minjul = Julian( Data->MinDate );
	long int maxjul = Julian( Data->MinDate );
	long int julian = Julian( pickDate );
	long int upjul = julian;
	long int downjul = julian;

	if ( ( DowMask & 0x7f ) == 0x7f )
		return FALSE;

	if ( ( DowMask & ( 1 << ( julian % 7 ) ) ) == 0 )
		return TRUE;

	while ( DowMask & ( 1 << ( upjul % 7 ) ) )
		upjul++;

	while ( DowMask & ( 1 << ( downjul % 7 ) ) )
		downjul--;

	if ( ( downjul < minjul ) && ( upjul > maxjul ) )
		return FALSE;

	if ( downjul < minjul ) {
		MakeDate( upjul, pickDate );
		return TRUE;
	}

	if ( upjul > maxjul ) {
		MakeDate( downjul, pickDate );
		return TRUE;
	}

	if ( ( ProgramFlags & FlagPreferMask ) == FlagPreferEarlier ) {
		MakeDate( downjul, pickDate );
		return TRUE;
	}

	if ( ( ProgramFlags & FlagPreferMask ) == FlagPreferLater ) {
		MakeDate( upjul, pickDate );
		return TRUE;
	}

	if ( ( upjul - julian ) < ( julian - downjul ) ) {
		MakeDate( upjul, pickDate );
		return TRUE;
	}

	MakeDate( downjul, pickDate );
	return TRUE;
}



DLLExports INT WINAPI f_pickdate( LPTSTR String )
{
	if ( !String )
		return 1;


	_PopCalData Data;


	GetNow( FALSE );
	GetLocalDateFormat();
	pickDate = Now;


	memset( &Data, 0x00, sizeof( _PopCalData ) );


	Data.MinDate.wYear  = Epoch;
	Data.MinDate.wMonth = 1;
	Data.MinDate.wDay   = 1;
	Data.MaxDate.wYear  = Ultima;
	Data.MaxDate.wMonth = 12;
	Data.MaxDate.wDay   = 31;

	DowMask = 0;
	ProgramFlags = 0;

	int rv = 0, fmt = 0x100, errfmt = 0x100;
	HolidaysYear = 0;

	memset( PopCalHolidays, 0x00, sizeof( PopCalHolidays ) );


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[2] ) {		// Third parameter:  Date format
		fmt = ParseInt( arg[2] );

		if ( fmt < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid format \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[7] ) {		// Eighth parameter:  Disabled days-of-the-week
		DowMask = ParseDowMask( arg[7] );

		if ( ( DowMask < 0 ) || ( DowMask > 126 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid days-of-the-week mask \"%s\"\r\n", arg[7] );
			return -1;
		}
	}

	if ( arglen[5] ) {		// Sixth parameter:  Misc program flags
		ProgramFlags = ParseInt( arg[5] );

		if ( ProgramFlags < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid flags \"%s\"\r\n", arg[5] );
			return -1;
		}
	}

	if ( arglen[1] ) {		// Second parameter:  Initial (selected) date
		if ( ParseDate( arg[1], fmt, pickDate ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid initial date \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( arglen[3] ) {		// Fourth parameter:  Minimum date
		if ( IsSign( arg[3][0] ) && iswdigit( arg[3][1] ) ) {
			long int n = Julian( pickDate ) + wcstol( arg[3], NULL, 10 );

			if ( ( n < MinimumJulian ) || ( n > MaximumJulian ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Invalid minimum date \"%s\" (offset)\r\n", arg[3] );
				return -1;
			}

			MakeDate( n, Data.MinDate );
		}
		else if ( ParseDate( arg[3], fmt, Data.MinDate ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid minimum date \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	if ( arglen[4] ) {		// Fifth parameter:  Maximum date
		if ( IsSign( arg[4][0] ) && iswdigit( arg[4][1] ) ) {
			long int n = Julian( pickDate ) + wcstol( arg[4], NULL, 10 );

			if ( ( n < MinimumJulian ) || ( n > MaximumJulian ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Invalid maximum date \"%s\" (offset)\r\n", arg[4] );
				return -1;
			}

			MakeDate( n, Data.MaxDate );
		}
		else if ( ParseDate( arg[4], fmt, Data.MaxDate ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid maximum date \"%s\"\r\n", arg[4] );
			return -1;
		}
	}

	if ( arglen[6] ) {		// Seventh parameter:  Left column day-of-week
		int lc = ParseDow( arg[6] );

		if ( lc < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid left-column day value \"%s\"\r\n", arg[6] );
			return -1;
		}
		else
			LeftColumnDow = lc % 7;
	}

	if ( arglen[8] ) {		// Ninth parameter:  Timeout value
		CountDown = ParseInt( arg[8] );

		if ( ( CountDown < 0 ) || ( CountDown > 600 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid timeout \"%s\"\r\n", arg[8] );
			return -1;
		}
	}
	else
		CountDown = 0;

	if ( ProgramFlags & FlagRestrict ) {
		if ( Data.MinDate.wYear < 1980) {
			Data.MinDate.wYear  = 1980;
			Data.MinDate.wMonth = 1;
			Data.MinDate.wDay   = 1;
		}

		if ( Data.MaxDate.wYear > 2099 ) {
			Data.MaxDate.wYear  = 2099;
			Data.MaxDate.wMonth = 12;
			Data.MaxDate.wDay   = 31;
		}
	}

	if ( ProgramFlags & FlagNoWeekends )
		DowMask = 0x41;

	if ( ( fmt & 0xff ) <= 4 )
		errfmt = fmt;
	
	if ( Julian( Data.MinDate ) > Julian( Data.MaxDate ) ) {
		OutputFormattedDate( Data.MinDate, errfmt, arg[3] );		//	fix me!
		OutputFormattedDate( Data.MaxDate, errfmt, arg[4] );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid date range %s - %s\r\n", arg[3], arg[4] );
		return -1;
	}
	
	if ( ( arglen[1] == 0 ) && ( arglen[3] > 0 ) && ( Julian( pickDate ) < Julian( Data.MinDate ) ) )
		pickDate = Data.MinDate;
	else if ( ( arglen[1] == 0 ) && ( arglen[4] > 0 ) && ( Julian( pickDate ) > Julian( Data.MaxDate ) ) )
		pickDate = Data.MaxDate;

	if ( ( Julian( pickDate ) < Julian( Data.MinDate ) ) || ( Julian( pickDate ) > Julian( Data.MaxDate ) ) ) {
		OutputFormattedDate( pickDate, errfmt, arg[1] );		//	fix me!
		OutputFormattedDate( Data.MinDate, errfmt, arg[3] );
		OutputFormattedDate( Data.MaxDate, errfmt, arg[4] );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Date %s not in range %s - %s\r\n", arg[1], arg[3], arg[4] );
		return -1;
	}

	if ( !ResolveMaskedSelection( &Data ) ) {
		OutputFormattedDate( Data.MinDate, errfmt, arg[3] );		//	fix me!
		OutputFormattedDate( Data.MaxDate, errfmt, arg[4] );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No unmasked days in range %s - %s\r\n", arg[3], arg[4] );
		return -1;
	}

	if ( ( ProgramFlags & FlagNoHolidays ) == 0 )
		if ( QcalReadHolidaysFile() )
			QcalDefaultHolidays();


	pickView = pickDate;
	pickdatefmt = fmt;

	Data.InitDate	= pickDate;

	HoldSignals();

	rv = (int) DialogBoxParamW( MyDLLHandle, MAKEINTRESOURCE( DLG_POPCAL ), GetForegroundWindow(), PopCalDlgProc, (LPARAM) &Data );

	switch ( rv ) {

	case IDOK:
		// return selected date :
		OutputFormattedDate( pickDate, fmt, String );
		return 0;

	case IDCANCEL:
		// canceled by user; return info string :
		wcscpy_s( String, MAX_ARG_LEN, L"*CANCEL*" );
		return 0;

	default:
		// dialog failed?
		wsprintf( String, L"Dialog error %d", rv );
		return -1;
	}
}


#endif	//	POPCAL
